@@ -49,6 +49,8 @@ fwupd_status_to_string(FwupdStatus status)
 		return "waiting-for-auth";
 	if (status == FWUPD_STATUS_SHUTDOWN)
 		return "shutdown";
+	if (status == FWUPD_STATUS_WAITING_FOR_USER)
+		return "waiting-for-user";
 	return NULL;
 }
 
@@ -93,6 +95,8 @@ fwupd_status_from_string(const gchar *status)
 		return FWUPD_STATUS_WAITING_FOR_AUTH;
 	if (g_strcmp0(status, "shutdown") == 0)
 		return FWUPD_STATUS_SHUTDOWN;
+	if (g_strcmp0(status, "waiting-for-user") == 0)
+		return FWUPD_STATUS_WAITING_FOR_USER;
 	return FWUPD_STATUS_LAST;
 }
 
@@ -209,6 +213,12 @@ fwupd_device_flag_to_string(FwupdDeviceFlags device_flag)
 		return "signed-payload";
 	if (device_flag == FWUPD_DEVICE_FLAG_UNSIGNED_PAYLOAD)
 		return "unsigned-payload";
+	if (device_flag == FWUPD_DEVICE_FLAG_EMULATED)
+		return "emulated";
+	if (device_flag == FWUPD_DEVICE_FLAG_EMULATION_TAG)
+		return "emulation-tag";
+	if (device_flag == FWUPD_DEVICE_FLAG_ONLY_EXPLICIT_UPDATES)
+		return "only-explicit-updates";
 	if (device_flag == FWUPD_DEVICE_FLAG_UNKNOWN)
 		return "unknown";
 	return NULL;
@@ -328,6 +338,12 @@ fwupd_device_flag_from_string(const gchar *device_flag)
 		return FWUPD_DEVICE_FLAG_SIGNED_PAYLOAD;
 	if (g_strcmp0(device_flag, "unsigned-payload") == 0)
 		return FWUPD_DEVICE_FLAG_UNSIGNED_PAYLOAD;
+	if (g_strcmp0(device_flag, "emulated") == 0)
+		return FWUPD_DEVICE_FLAG_EMULATED;
+	if (g_strcmp0(device_flag, "emulation-tag") == 0)
+		return FWUPD_DEVICE_FLAG_EMULATION_TAG;
+	if (g_strcmp0(device_flag, "only-explicit-updates") == 0)
+		return FWUPD_DEVICE_FLAG_ONLY_EXPLICIT_UPDATES;
 	return FWUPD_DEVICE_FLAG_UNKNOWN;
 }
 
@@ -362,6 +378,14 @@ fwupd_device_problem_to_string(FwupdDeviceProblem device_problem)
 		return "is-emulated";
 	if (device_problem == FWUPD_DEVICE_PROBLEM_MISSING_LICENSE)
 		return "missing-license";
+	if (device_problem == FWUPD_DEVICE_PROBLEM_SYSTEM_INHIBIT)
+		return "system-inhibit";
+	if (device_problem == FWUPD_DEVICE_PROBLEM_UPDATE_IN_PROGRESS)
+		return "update-in-progress";
+	if (device_problem == FWUPD_DEVICE_PROBLEM_IN_USE)
+		return "in-use";
+	if (device_problem == FWUPD_DEVICE_PROBLEM_DISPLAY_REQUIRED)
+		return "display-required";
 	if (device_problem == FWUPD_DEVICE_PROBLEM_UNKNOWN)
 		return "unknown";
 	return NULL;
@@ -398,6 +422,14 @@ fwupd_device_problem_from_string(const gchar *device_problem)
 		return FWUPD_DEVICE_PROBLEM_IS_EMULATED;
 	if (g_strcmp0(device_problem, "missing-license") == 0)
 		return FWUPD_DEVICE_PROBLEM_MISSING_LICENSE;
+	if (g_strcmp0(device_problem, "system-inhibit") == 0)
+		return FWUPD_DEVICE_PROBLEM_SYSTEM_INHIBIT;
+	if (g_strcmp0(device_problem, "update-in-progress") == 0)
+		return FWUPD_DEVICE_PROBLEM_UPDATE_IN_PROGRESS;
+	if (g_strcmp0(device_problem, "in-use") == 0)
+		return FWUPD_DEVICE_PROBLEM_IN_USE;
+	if (g_strcmp0(device_problem, "display-required") == 0)
+		return FWUPD_DEVICE_PROBLEM_DISPLAY_REQUIRED;
 	return FWUPD_DEVICE_PROBLEM_UNKNOWN;
 }
 
@@ -432,6 +464,8 @@ fwupd_plugin_flag_to_string(FwupdPluginFlags plugin_flag)
 		return "efivar-not-mounted";
 	if (plugin_flag == FWUPD_PLUGIN_FLAG_ESP_NOT_FOUND)
 		return "esp-not-found";
+	if (plugin_flag == FWUPD_PLUGIN_FLAG_ESP_NOT_VALID)
+		return "esp-not-valid";
 	if (plugin_flag == FWUPD_PLUGIN_FLAG_LEGACY_BIOS)
 		return "legacy-bios";
 	if (plugin_flag == FWUPD_PLUGIN_FLAG_FAILED_OPEN)
@@ -448,6 +482,12 @@ fwupd_plugin_flag_to_string(FwupdPluginFlags plugin_flag)
 		return "secure-config";
 	if (plugin_flag == FWUPD_PLUGIN_FLAG_MODULAR)
 		return "modular";
+	if (plugin_flag == FWUPD_PLUGIN_FLAG_MEASURE_SYSTEM_INTEGRITY)
+		return "measure-system-integrity";
+	if (plugin_flag == FWUPD_PLUGIN_FLAG_READY)
+		return "ready";
+	if (plugin_flag == FWUPD_PLUGIN_FLAG_TEST_ONLY)
+		return "test-only";
 	return NULL;
 }
 
@@ -482,6 +522,8 @@ fwupd_plugin_flag_from_string(const gchar *plugin_flag)
 		return FWUPD_PLUGIN_FLAG_EFIVAR_NOT_MOUNTED;
 	if (g_strcmp0(plugin_flag, "esp-not-found") == 0)
 		return FWUPD_PLUGIN_FLAG_ESP_NOT_FOUND;
+	if (g_strcmp0(plugin_flag, "esp-not-valid") == 0)
+		return FWUPD_PLUGIN_FLAG_ESP_NOT_VALID;
 	if (g_strcmp0(plugin_flag, "legacy-bios") == 0)
 		return FWUPD_PLUGIN_FLAG_LEGACY_BIOS;
 	if (g_strcmp0(plugin_flag, "failed-open") == 0)
@@ -496,6 +538,12 @@ fwupd_plugin_flag_from_string(const gchar *plugin_flag)
 		return FWUPD_PLUGIN_FLAG_SECURE_CONFIG;
 	if (g_strcmp0(plugin_flag, "modular") == 0)
 		return FWUPD_PLUGIN_FLAG_MODULAR;
+	if (g_strcmp0(plugin_flag, "measure-system-integrity") == 0)
+		return FWUPD_PLUGIN_FLAG_MEASURE_SYSTEM_INTEGRITY;
+	if (g_strcmp0(plugin_flag, "ready") == 0)
+		return FWUPD_PLUGIN_FLAG_READY;
+	if (g_strcmp0(plugin_flag, "test-only") == 0)
+		return FWUPD_PLUGIN_FLAG_TEST_ONLY;
 	return FWUPD_DEVICE_FLAG_UNKNOWN;
 }
 
@@ -632,6 +680,8 @@ fwupd_feature_flag_to_string(FwupdFeatureFlags feature_flag)
 		return "show-problems";
 	if (feature_flag == FWUPD_FEATURE_FLAG_ALLOW_AUTHENTICATION)
 		return "allow-authentication";
+	if (feature_flag == FWUPD_FEATURE_FLAG_REQUESTS_NON_GENERIC)
+		return "requests-non-generic";
 	return NULL;
 }
 
@@ -668,6 +718,8 @@ fwupd_feature_flag_from_string(const gchar *feature_flag)
 		return FWUPD_FEATURE_FLAG_SHOW_PROBLEMS;
 	if (g_strcmp0(feature_flag, "allow-authentication") == 0)
 		return FWUPD_FEATURE_FLAG_ALLOW_AUTHENTICATION;
+	if (g_strcmp0(feature_flag, "requests-non-generic") == 0)
+		return FWUPD_FEATURE_FLAG_REQUESTS_NON_GENERIC;
 	return FWUPD_FEATURE_FLAG_LAST;
 }
 
@@ -750,6 +802,8 @@ fwupd_release_flag_to_string(FwupdReleaseFlags release_flag)
 		return "is-alternate-branch";
 	if (release_flag == FWUPD_RELEASE_FLAG_IS_COMMUNITY)
 		return "is-community";
+	if (release_flag == FWUPD_RELEASE_FLAG_TRUSTED_REPORT)
+		return "trusted-report";
 	return NULL;
 }
 
@@ -782,6 +836,8 @@ fwupd_release_flag_from_string(const gchar *release_flag)
 		return FWUPD_RELEASE_FLAG_IS_ALTERNATE_BRANCH;
 	if (g_strcmp0(release_flag, "is-community") == 0)
 		return FWUPD_RELEASE_FLAG_IS_COMMUNITY;
+	if (g_strcmp0(release_flag, "trusted-report") == 0)
+		return FWUPD_RELEASE_FLAG_TRUSTED_REPORT;
 	return FWUPD_RELEASE_FLAG_NONE;
 }
 
@@ -870,6 +926,8 @@ fwupd_version_format_from_string(const gchar *str)
 		return FWUPD_VERSION_FORMAT_DELL_BIOS;
 	if (g_strcmp0(str, "hex") == 0)
 		return FWUPD_VERSION_FORMAT_HEX;
+	if (g_strcmp0(str, "dell-bios-msb") == 0)
+		return FWUPD_VERSION_FORMAT_DELL_BIOS_MSB;
 	return FWUPD_VERSION_FORMAT_UNKNOWN;
 }
 
@@ -910,5 +968,7 @@ fwupd_version_format_to_string(FwupdVersionFormat kind)
 		return "dell-bios";
 	if (kind == FWUPD_VERSION_FORMAT_HEX)
 		return "hex";
+	if (kind == FWUPD_VERSION_FORMAT_DELL_BIOS_MSB)
+		return "dell-bios-msb";
 	return NULL;
 }
