@@ -9,8 +9,12 @@
 #include "config.h"
 
 #include "fu-bios-settings-private.h"
+#include "fu-common-private.h"
+#include "fu-config-private.h"
 #include "fu-context-private.h"
-#include "fu-hwids.h"
+#include "fu-fdt-firmware.h"
+#include "fu-hwids-private.h"
+#include "fu-path.h"
 #include "fu-smbios-private.h"
 #include "fu-volume-private.h"
 
@@ -24,29 +28,36 @@
 typedef struct {
 	FuContextFlags flags;
 	FuHwids *hwids;
+	FuConfig *config;
 	FuSmbios *smbios;
+	FuSmbiosChassisKind chassis_kind;
 	FuQuirks *quirks;
 	GHashTable *runtime_versions;
 	GHashTable *compile_versions;
-	GPtrArray *udev_subsystems;
+	GHashTable *udev_subsystems; /* utf8:GPtrArray */
 	GPtrArray *esp_volumes;
-	GHashTable *firmware_gtypes;
+	GHashTable *firmware_gtypes; /* utf8:GType */
 	GHashTable *hwid_flags; /* str: */
-	FuBatteryState battery_state;
+	FuPowerState power_state;
 	FuLidState lid_state;
+	FuDisplayState display_state;
 	guint battery_level;
 	guint battery_threshold;
 	FuBiosSettings *host_bios_settings;
+	FuFirmware *fdt; /* optional */
+	gchar *esp_location;
 } FuContextPrivate;
 
 enum { SIGNAL_SECURITY_CHANGED, SIGNAL_LAST };
 
 enum {
 	PROP_0,
-	PROP_BATTERY_STATE,
+	PROP_POWER_STATE,
 	PROP_LID_STATE,
+	PROP_DISPLAY_STATE,
 	PROP_BATTERY_LEVEL,
 	PROP_BATTERY_THRESHOLD,
+	PROP_FLAGS,
 	PROP_LAST
 };
 
@@ -56,11 +67,134 @@ G_DEFINE_TYPE_WITH_PRIVATE(FuContext, fu_context, G_TYPE_OBJECT)
 
 #define GET_PRIVATE(o) (fu_context_get_instance_private(o))
 
+static GFile *
+fu_context_get_fdt_file(GError **error)
+{
+	g_autofree gchar *fdtfn_local = NULL;
+	g_autofree gchar *fdtfn_sys = NULL;
+	g_autofree gchar *localstatedir_pkg = fu_path_from_kind(FU_PATH_KIND_LOCALSTATEDIR_PKG);
+	g_autofree gchar *sysfsdir = NULL;
+
+	/* look for override first, fall back to system value */
+	fdtfn_local = g_build_filename(localstatedir_pkg, "system.dtb", NULL);
+	if (g_file_test(fdtfn_local, G_FILE_TEST_EXISTS))
+		return g_file_new_for_path(fdtfn_local);
+
+	/* actual hardware value */
+	sysfsdir = fu_path_from_kind(FU_PATH_KIND_SYSFSDIR_FW);
+	fdtfn_sys = g_build_filename(sysfsdir, "fdt", NULL);
+	if (g_file_test(fdtfn_sys, G_FILE_TEST_EXISTS))
+		return g_file_new_for_path(fdtfn_sys);
+
+	/* failed */
+	g_set_error(error,
+		    FWUPD_ERROR,
+		    FWUPD_ERROR_NOT_SUPPORTED,
+		    "cannot find %s or override %s",
+		    fdtfn_sys,
+		    fdtfn_local);
+	return NULL;
+}
+
+/**
+ * fu_context_get_fdt:
+ * @self: a #FuContext
+ * @error: (nullable): optional return location for an error
+ *
+ * Gets and parses the system FDT, aka. the Flat Device Tree.
+ *
+ * The results are cached internally to the context, and subsequent calls to this function
+ * returns the pre-parsed object.
+ *
+ * Returns: (transfer full): a #FuFdtFirmware, or %NULL
+ *
+ * Since: 1.8.10
+ **/
+FuFirmware *
+fu_context_get_fdt(FuContext *self, GError **error)
+{
+	FuContextPrivate *priv = GET_PRIVATE(self);
+
+	g_return_val_if_fail(FU_IS_CONTEXT(self), NULL);
+	g_return_val_if_fail(error == NULL || *error == NULL, NULL);
+
+	/* load if not already parsed */
+	if (priv->fdt == NULL) {
+		g_autoptr(FuFirmware) fdt_tmp = fu_fdt_firmware_new();
+		g_autoptr(GFile) file = fu_context_get_fdt_file(error);
+		if (file == NULL)
+			return NULL;
+		if (!fu_firmware_parse_file(fdt_tmp, file, FWUPD_INSTALL_FLAG_NO_SEARCH, error)) {
+			g_prefix_error(error, "failed to parse FDT: ");
+			return NULL;
+		}
+		priv->fdt = g_steal_pointer(&fdt_tmp);
+	}
+
+	/* success */
+	return g_object_ref(priv->fdt);
+}
+
+/**
+ * fu_context_get_smbios:
+ * @self: a #FuContext
+ *
+ * Gets the SMBIOS store.
+ *
+ * Returns: (transfer none): a #FuSmbios
+ *
+ * Since: 1.8.10
+ **/
+FuSmbios *
+fu_context_get_smbios(FuContext *self)
+{
+	FuContextPrivate *priv = GET_PRIVATE(self);
+	g_return_val_if_fail(FU_IS_CONTEXT(self), NULL);
+	return priv->smbios;
+}
+
+/**
+ * fu_context_get_hwids:
+ * @self: a #FuContext
+ *
+ * Gets the HWIDs store.
+ *
+ * Returns: (transfer none): a #FuHwids
+ *
+ * Since: 1.8.10
+ **/
+FuHwids *
+fu_context_get_hwids(FuContext *self)
+{
+	FuContextPrivate *priv = GET_PRIVATE(self);
+	g_return_val_if_fail(FU_IS_CONTEXT(self), NULL);
+	return priv->hwids;
+}
+
+/**
+ * fu_context_get_config:
+ * @self: a #FuContext
+ *
+ * Gets the system config.
+ *
+ * Returns: (transfer none): a #FuHwids
+ *
+ * Since: 1.9.1
+ **/
+FuConfig *
+fu_context_get_config(FuContext *self)
+{
+	FuContextPrivate *priv = GET_PRIVATE(self);
+	g_return_val_if_fail(FU_IS_CONTEXT(self), NULL);
+	return priv->config;
+}
+
 /**
  * fu_context_get_smbios_string:
  * @self: a #FuContext
  * @structure_type: a SMBIOS structure type, e.g. %FU_SMBIOS_STRUCTURE_TYPE_BIOS
  * @offset: a SMBIOS offset
+ * @error: (nullable): optional return location for an error
  *
  * Gets a hardware SMBIOS string.
  *
@@ -72,30 +206,43 @@ G_DEFINE_TYPE_WITH_PRIVATE(FuContext, fu_context, G_TYPE_OBJECT)
  * Since: 1.6.0
  **/
 const gchar *
-fu_context_get_smbios_string(FuContext *self, guint8 structure_type, guint8 offset)
+fu_context_get_smbios_string(FuContext *self, guint8 structure_type, guint8 offset, GError **error)
 {
 	FuContextPrivate *priv = GET_PRIVATE(self);
 	g_return_val_if_fail(FU_IS_CONTEXT(self), NULL);
-	return fu_smbios_get_string(priv->smbios, structure_type, offset, NULL);
+	if (!fu_context_has_flag(self, FU_CONTEXT_FLAG_LOADED_HWINFO)) {
+		g_critical("cannot use SMBIOS before calling ->load_hwinfo()");
+		return NULL;
+	}
+	return fu_smbios_get_string(priv->smbios, structure_type, offset, error);
 }
 
 /**
  * fu_context_get_smbios_data:
  * @self: a #FuContext
  * @structure_type: a SMBIOS structure type, e.g. %FU_SMBIOS_STRUCTURE_TYPE_BIOS
+ * @error: (nullable): optional return location for an error
  *
- * Gets a hardware SMBIOS data.
+ * Gets all hardware SMBIOS data for a specific type.
  *
- * Returns: (transfer full): a #GBytes, or %NULL
+ * Returns: (transfer container) (element-type GBytes): a #GBytes, or %NULL if not found
  *
- * Since: 1.6.0
+ * Since: 1.9.8
  **/
-GBytes *
-fu_context_get_smbios_data(FuContext *self, guint8 structure_type)
+GPtrArray *
+fu_context_get_smbios_data(FuContext *self, guint8 structure_type, GError **error)
 {
 	FuContextPrivate *priv = GET_PRIVATE(self);
+
 	g_return_val_if_fail(FU_IS_CONTEXT(self), NULL);
-	return fu_smbios_get_data(priv->smbios, structure_type, NULL);
+
+	/* must be valid and non-zero length */
+	if (!fu_context_has_flag(self, FU_CONTEXT_FLAG_LOADED_HWINFO)) {
+		g_critical("cannot use SMBIOS before calling ->load_hwinfo()");
+		g_set_error_literal(error, G_IO_ERROR, G_IO_ERROR_NOT_INITIALIZED, "no data");
+		return NULL;
+	}
+	return fu_smbios_get_data(priv->smbios, structure_type, error);
 }
 
 /**
@@ -103,6 +250,7 @@ fu_context_get_smbios_data(FuContext *self, guint8 structure_type)
  * @self: a #FuContext
  * @type: a structure type, e.g. %FU_SMBIOS_STRUCTURE_TYPE_BIOS
  * @offset: a structure offset
+ * @error: (nullable): optional return location for an error
  *
  * Reads an integer value from the SMBIOS string table of a specific structure.
  *
@@ -114,11 +262,15 @@ fu_context_get_smbios_data(FuContext *self, guint8 structure_type)
  * Since: 1.6.0
  **/
 guint
-fu_context_get_smbios_integer(FuContext *self, guint8 type, guint8 offset)
+fu_context_get_smbios_integer(FuContext *self, guint8 type, guint8 offset, GError **error)
 {
 	FuContextPrivate *priv = GET_PRIVATE(self);
 	g_return_val_if_fail(FU_IS_CONTEXT(self), G_MAXUINT);
-	return fu_smbios_get_integer(priv->smbios, type, offset, NULL);
+	if (!fu_context_has_flag(self, FU_CONTEXT_FLAG_LOADED_HWINFO)) {
+		g_critical("cannot use SMBIOS before calling ->load_hwinfo()");
+		return G_MAXUINT;
+	}
+	return fu_smbios_get_integer(priv->smbios, type, offset, error);
 }
 
 /**
@@ -196,6 +348,41 @@ fu_context_get_bios_setting_pending_reboot(FuContext *self)
 	return ret;
 }
 
+/**
+ * fu_context_get_chassis_kind:
+ * @self: a #FuContext
+ *
+ * Gets the chassis kind, if known.
+ *
+ * Returns: a #FuSmbiosChassisKind, e.g. %FU_SMBIOS_CHASSIS_KIND_LAPTOP
+ *
+ * Since: 1.8.10
+ **/
+FuSmbiosChassisKind
+fu_context_get_chassis_kind(FuContext *self)
+{
+	FuContextPrivate *priv = GET_PRIVATE(self);
+	g_return_val_if_fail(FU_IS_CONTEXT(self), FALSE);
+	return priv->chassis_kind;
+}
+
+/**
+ * fu_context_set_chassis_kind:
+ * @self: a #FuContext
+ * @chassis_kind: a #FuSmbiosChassisKind, e.g. %FU_SMBIOS_CHASSIS_KIND_TABLET
+ *
+ * Sets the chassis kind.
+ *
+ * Since: 1.8.10
+ **/
+void
+fu_context_set_chassis_kind(FuContext *self, FuSmbiosChassisKind chassis_kind)
+{
+	FuContextPrivate *priv = GET_PRIVATE(self);
+	g_return_if_fail(FU_IS_CONTEXT(self));
+	priv->chassis_kind = chassis_kind;
+}
+
 /**
  * fu_context_has_hwid_guid:
  * @self: a #FuContext
@@ -212,6 +399,10 @@ fu_context_has_hwid_guid(FuContext *self, const gchar *guid)
 {
 	FuContextPrivate *priv = GET_PRIVATE(self);
 	g_return_val_if_fail(FU_IS_CONTEXT(self), FALSE);
+	if (!fu_context_has_flag(self, FU_CONTEXT_FLAG_LOADED_HWINFO)) {
+		g_critical("cannot use HWIDs before calling ->load_hwinfo()");
+		return FALSE;
+	}
 	return fu_hwids_has_guid(priv->hwids, guid);
 }
 
@@ -231,6 +422,10 @@ fu_context_get_hwid_guids(FuContext *self)
 {
 	FuContextPrivate *priv = GET_PRIVATE(self);
 	g_return_val_if_fail(FU_IS_CONTEXT(self), NULL);
+	if (!fu_context_has_flag(self, FU_CONTEXT_FLAG_LOADED_HWINFO)) {
+		g_critical("cannot use HWIDs before calling ->load_hwinfo()");
+		return NULL;
+	}
 	return fu_hwids_get_guids(priv->hwids);
 }
 
@@ -252,6 +447,10 @@ fu_context_get_hwid_value(FuContext *self, const gchar *key)
 	FuContextPrivate *priv = GET_PRIVATE(self);
 	g_return_val_if_fail(FU_IS_CONTEXT(self), NULL);
 	g_return_val_if_fail(key != NULL, NULL);
+	if (!fu_context_has_flag(self, FU_CONTEXT_FLAG_LOADED_HWINFO)) {
+		g_critical("cannot use HWIDs before calling ->load_hwinfo()");
+		return NULL;
+	}
 	return fu_hwids_get_value(priv->hwids, key);
 }
 
@@ -272,8 +471,16 @@ gchar *
 fu_context_get_hwid_replace_value(FuContext *self, const gchar *keys, GError **error)
 {
 	FuContextPrivate *priv = GET_PRIVATE(self);
+
 	g_return_val_if_fail(FU_IS_CONTEXT(self), NULL);
 	g_return_val_if_fail(keys != NULL, NULL);
+	g_return_val_if_fail(error == NULL || *error == NULL, NULL);
+
+	if (!fu_context_has_flag(self, FU_CONTEXT_FLAG_LOADED_HWINFO)) {
+		g_critical("cannot use HWIDs before calling ->load_hwinfo()");
+		g_set_error_literal(error, G_IO_ERROR, G_IO_ERROR_NOT_INITIALIZED, "no data");
+		return NULL;
+	}
 	return fu_hwids_get_replace_values(priv->hwids, keys, error);
 }
 
@@ -302,21 +509,45 @@ fu_context_add_runtime_version(FuContext *self, const gchar *component_id, const
 }
 
 /**
- * fu_context_set_runtime_versions:
+ * fu_context_get_runtime_version:
  * @self: a #FuContext
- * @runtime_versions: (element-type utf8 utf8): dictionary of versions
+ * @component_id: an AppStream component id, e.g. `org.gnome.Software`
  *
- * Sets the runtime versions for a plugin.
+ * Sets a runtime version of a specific dependency.
  *
- * Since: 1.6.0
+ * Returns: a version string, e.g. `1.2.3`, or %NULL
+ *
+ * Since: 1.9.10
  **/
-void
-fu_context_set_runtime_versions(FuContext *self, GHashTable *runtime_versions)
+const gchar *
+fu_context_get_runtime_version(FuContext *self, const gchar *component_id)
 {
 	FuContextPrivate *priv = GET_PRIVATE(self);
-	g_return_if_fail(FU_IS_CONTEXT(self));
-	g_return_if_fail(runtime_versions != NULL);
-	priv->runtime_versions = g_hash_table_ref(runtime_versions);
+
+	g_return_val_if_fail(FU_IS_CONTEXT(self), NULL);
+	g_return_val_if_fail(component_id != NULL, NULL);
+
+	if (priv->runtime_versions == NULL)
+		return NULL;
+	return g_hash_table_lookup(priv->runtime_versions, component_id);
+}
+
+/**
+ * fu_context_get_runtime_versions:
+ * @self: a #FuContext
+ *
+ * Gets the runtime versions for the context.
+ *
+ * Returns: (transfer none) (element-type utf8 utf8): dictionary of versions
+ *
+ * Since: 1.9.10
+ **/
+GHashTable *
+fu_context_get_runtime_versions(FuContext *self)
+{
+	FuContextPrivate *priv = GET_PRIVATE(self);
+	g_return_val_if_fail(FU_IS_CONTEXT(self), NULL);
+	return priv->runtime_versions;
 }
 
 /**
@@ -344,27 +575,36 @@ fu_context_add_compile_version(FuContext *self, const gchar *component_id, const
 }
 
 /**
- * fu_context_set_compile_versions:
+ * fu_context_get_compile_versions:
  * @self: a #FuContext
- * @compile_versions: (element-type utf8 utf8): dictionary of versions
  *
- * Sets the compile time versions for a plugin.
+ * Gets the compile time versions for the context.
  *
- * Since: 1.6.0
+ * Returns: (transfer none) (element-type utf8 utf8): dictionary of versions
+ *
+ * Since: 1.9.10
  **/
-void
-fu_context_set_compile_versions(FuContext *self, GHashTable *compile_versions)
+GHashTable *
+fu_context_get_compile_versions(FuContext *self)
 {
 	FuContextPrivate *priv = GET_PRIVATE(self);
-	g_return_if_fail(FU_IS_CONTEXT(self));
-	g_return_if_fail(compile_versions != NULL);
-	priv->compile_versions = g_hash_table_ref(compile_versions);
+	g_return_val_if_fail(FU_IS_CONTEXT(self), NULL);
+	return priv->compile_versions;
+}
+
+static gint
+fu_context_udev_plugin_names_sort_cb(gconstpointer a, gconstpointer b)
+{
+	const gchar *str_a = *((const gchar **)a);
+	const gchar *str_b = *((const gchar **)b);
+	return g_strcmp0(str_a, str_b);
 }
 
 /**
  * fu_context_add_udev_subsystem:
  * @self: a #FuContext
  * @subsystem: a subsystem name, e.g. `pciport`
+ * @plugin_name: (nullable): a plugin name, e.g. `iommu`
  *
  * Registers the udev subsystem to be watched by the daemon.
  *
@@ -373,20 +613,75 @@ fu_context_set_compile_versions(FuContext *self, GHashTable *compile_versions)
  * Since: 1.6.0
  **/
 void
-fu_context_add_udev_subsystem(FuContext *self, const gchar *subsystem)
+fu_context_add_udev_subsystem(FuContext *self, const gchar *subsystem, const gchar *plugin_name)
 {
 	FuContextPrivate *priv = GET_PRIVATE(self);
+	GPtrArray *plugin_names;
 
 	g_return_if_fail(FU_IS_CONTEXT(self));
 	g_return_if_fail(subsystem != NULL);
 
-	for (guint i = 0; i < priv->udev_subsystems->len; i++) {
-		const gchar *subsystem_tmp = g_ptr_array_index(priv->udev_subsystems, i);
-		if (g_strcmp0(subsystem_tmp, subsystem) == 0)
-			return;
+	/* already exists */
+	plugin_names = g_hash_table_lookup(priv->udev_subsystems, subsystem);
+	if (plugin_names != NULL) {
+		if (plugin_name != NULL) {
+			for (guint i = 0; i < plugin_names->len; i++) {
+				const gchar *tmp = g_ptr_array_index(plugin_names, i);
+				if (g_strcmp0(tmp, plugin_name) == 0)
+					return;
+			}
+			g_ptr_array_add(plugin_names, g_strdup(plugin_name));
+			g_ptr_array_sort(plugin_names, fu_context_udev_plugin_names_sort_cb);
+		}
+		return;
 	}
-	g_debug("added udev subsystem watch of %s", subsystem);
-	g_ptr_array_add(priv->udev_subsystems, g_strdup(subsystem));
+
+	/* add */
+	plugin_names = g_ptr_array_new_with_free_func(g_free);
+	if (plugin_name != NULL)
+		g_ptr_array_add(plugin_names, g_strdup(plugin_name));
+	g_hash_table_insert(priv->udev_subsystems,
+			    g_strdup(subsystem),
+			    g_steal_pointer(&plugin_names));
+	if (plugin_name != NULL)
+		g_info("added udev subsystem watch of %s for plugin %s", subsystem, plugin_name);
+	else
+		g_info("added udev subsystem watch of %s", subsystem);
+}
+
+/**
+ * fu_context_get_plugin_names_for_udev_subsystem:
+ * @self: a #FuContext
+ * @subsystem: a subsystem name, e.g. `pciport`
+ * @error: (nullable): optional return location for an error
+ *
+ * Gets the plugins which registered for a specific subsystem.
+ *
+ * Returns: (transfer container) (element-type utf8): List of plugin names
+ *
+ * Since: 1.9.3
+ **/
+GPtrArray *
+fu_context_get_plugin_names_for_udev_subsystem(FuContext *self,
+					       const gchar *subsystem,
+					       GError **error)
+{
+	FuContextPrivate *priv = GET_PRIVATE(self);
+	GPtrArray *plugin_names;
+
+	g_return_val_if_fail(FU_IS_CONTEXT(self), NULL);
+	g_return_val_if_fail(subsystem != NULL, NULL);
+
+	plugin_names = g_hash_table_lookup(priv->udev_subsystems, subsystem);
+	if (plugin_names == NULL) {
+		g_set_error(error,
+			    G_IO_ERROR,
+			    G_IO_ERROR_NOT_FOUND,
+			    "no plugins registered for %s",
+			    subsystem);
+		return NULL;
+	}
+	return g_ptr_array_ref(plugin_names);
 }
 
 /**
@@ -395,7 +690,7 @@ fu_context_add_udev_subsystem(FuContext *self, const gchar *subsystem)
  *
  * Gets the udev subsystems required by all plugins.
  *
- * Returns: (transfer none) (element-type utf8): List of subsystems
+ * Returns: (transfer container) (element-type utf8): List of subsystems
  *
  * Since: 1.6.0
  **/
@@ -403,8 +698,16 @@ GPtrArray *
 fu_context_get_udev_subsystems(FuContext *self)
 {
 	FuContextPrivate *priv = GET_PRIVATE(self);
+	g_autoptr(GList) keys = g_hash_table_get_keys(priv->udev_subsystems);
+	g_autoptr(GPtrArray) subsystems = g_ptr_array_new_with_free_func(g_free);
+
 	g_return_val_if_fail(FU_IS_CONTEXT(self), NULL);
-	return priv->udev_subsystems;
+
+	for (GList *l = keys; l != NULL; l = l->next) {
+		const gchar *subsystem = (const gchar *)l->data;
+		g_ptr_array_add(subsystems, g_strdup(subsystem));
+	}
+	return g_steal_pointer(&subsystems);
 }
 
 /**
@@ -465,7 +768,7 @@ fu_context_gtypes_sort_cb(gconstpointer a, gconstpointer b)
  *
  * Returns all the firmware #GType IDs.
  *
- * Returns: (transfer none) (element-type utf8): List of subsystems
+ * Returns: (transfer container) (element-type utf8): firmware IDs
  *
  * Since: 1.6.0
  **/
@@ -486,6 +789,32 @@ fu_context_get_firmware_gtype_ids(FuContext *self)
 	return firmware_gtypes;
 }
 
+/**
+ * fu_context_get_firmware_gtypes:
+ * @self: a #FuContext
+ *
+ * Returns all the firmware #GType's.
+ *
+ * Returns: (transfer none) (element-type GType): Firmware types
+ *
+ * Since: 1.9.1
+ **/
+GArray *
+fu_context_get_firmware_gtypes(FuContext *self)
+{
+	FuContextPrivate *priv = GET_PRIVATE(self);
+	GArray *firmware_gtypes = g_array_new(FALSE, FALSE, sizeof(GType));
+	g_autoptr(GList) values = g_hash_table_get_values(priv->firmware_gtypes);
+
+	g_return_val_if_fail(FU_IS_CONTEXT(self), NULL);
+
+	for (GList *l = values; l != NULL; l = l->next) {
+		GType gtype = GPOINTER_TO_SIZE(l->data);
+		g_array_append_val(firmware_gtypes, gtype);
+	}
+	return firmware_gtypes;
+}
+
 /**
  * fu_context_add_quirk_key:
  * @self: a #FuContext
@@ -534,11 +863,28 @@ fu_context_lookup_quirk_by_id(FuContext *self, const gchar *guid, const gchar *k
 	return fu_quirks_lookup_by_id(priv->quirks, guid, key);
 }
 
+typedef struct {
+	FuContext *self; /* noref */
+	FuContextLookupIter iter_cb;
+	gpointer user_data;
+} FuContextQuirkLookupHelper;
+
+static void
+fu_context_lookup_quirk_by_id_iter_cb(FuQuirks *self,
+				      const gchar *key,
+				      const gchar *value,
+				      gpointer user_data)
+{
+	FuContextQuirkLookupHelper *helper = (FuContextQuirkLookupHelper *)user_data;
+	helper->iter_cb(helper->self, key, value, helper->user_data);
+}
+
 /**
  * fu_context_lookup_quirk_by_id_iter:
  * @self: a #FuContext
  * @guid: GUID to lookup
- * @iter_cb: (scope async): a function to call for each result
+ * @key: (nullable): an ID to match the entry, e.g. `Name`, or %NULL for all keys
+ * @iter_cb: (scope call) (closure user_data): a function to call for each result
  * @user_data: user data passed to @iter_cb
  *
  * Looks up all entries in the hardware database using a GUID value.
@@ -550,14 +896,24 @@ fu_context_lookup_quirk_by_id(FuContext *self, const gchar *guid, const gchar *k
 gboolean
 fu_context_lookup_quirk_by_id_iter(FuContext *self,
 				   const gchar *guid,
+				   const gchar *key,
 				   FuContextLookupIter iter_cb,
 				   gpointer user_data)
 {
 	FuContextPrivate *priv = GET_PRIVATE(self);
+	FuContextQuirkLookupHelper helper = {
+	    .self = self,
+	    .iter_cb = iter_cb,
+	    .user_data = user_data,
+	};
 	g_return_val_if_fail(FU_IS_CONTEXT(self), FALSE);
 	g_return_val_if_fail(guid != NULL, FALSE);
 	g_return_val_if_fail(iter_cb != NULL, FALSE);
-	return fu_quirks_lookup_by_id_iter(priv->quirks, guid, (FuQuirksIter)iter_cb, user_data);
+	return fu_quirks_lookup_by_id_iter(priv->quirks,
+					   guid,
+					   key,
+					   fu_context_lookup_quirk_by_id_iter_cb,
+					   &helper);
 }
 
 /**
@@ -575,52 +931,103 @@ fu_context_security_changed(FuContext *self)
 	g_signal_emit(self, signals[SIGNAL_SECURITY_CHANGED], 0);
 }
 
+typedef gboolean (*FuContextHwidsSetupFunc)(FuContext *self, FuHwids *hwids, GError **error);
+
+static void
+fu_context_hwid_quirk_cb(FuContext *self, const gchar *key, const gchar *value, gpointer user_data)
+{
+	FuContextPrivate *priv = GET_PRIVATE(self);
+	if (value != NULL) {
+		g_auto(GStrv) values = g_strsplit(value, ",", -1);
+		for (guint j = 0; values[j] != NULL; j++)
+			g_hash_table_add(priv->hwid_flags, g_strdup(values[j]));
+	}
+}
+
 /**
  * fu_context_load_hwinfo:
  * @self: a #FuContext
+ * @progress: a #FuProgress
+ * @flags: a #FuContextHwidFlags, e.g. %FU_CONTEXT_HWID_FLAG_LOAD_SMBIOS
  * @error: (nullable): optional return location for an error
  *
  * Loads all hardware information parts of the context.
  *
  * Returns: %TRUE for success
  *
- * Since: 1.6.0
+ * Since: 1.8.10
  **/
 gboolean
-fu_context_load_hwinfo(FuContext *self, GError **error)
+fu_context_load_hwinfo(FuContext *self,
+		       FuProgress *progress,
+		       FuContextHwidFlags flags,
+		       GError **error)
 {
 	FuContextPrivate *priv = GET_PRIVATE(self);
 	GPtrArray *guids;
-	g_autoptr(GError) error_smbios = NULL;
 	g_autoptr(GError) error_hwids = NULL;
 	g_autoptr(GError) error_bios_settings = NULL;
+	struct {
+		const gchar *name;
+		FuContextHwidFlags flag;
+		FuContextHwidsSetupFunc func;
+	} hwids_setup_map[] = {{"config", FU_CONTEXT_HWID_FLAG_LOAD_CONFIG, fu_hwids_config_setup},
+			       {"smbios", FU_CONTEXT_HWID_FLAG_LOAD_SMBIOS, fu_hwids_smbios_setup},
+			       {"fdt", FU_CONTEXT_HWID_FLAG_LOAD_FDT, fu_hwids_fdt_setup},
+			       {"kenv", FU_CONTEXT_HWID_FLAG_LOAD_KENV, fu_hwids_kenv_setup},
+			       {"dmi", FU_CONTEXT_HWID_FLAG_LOAD_DMI, fu_hwids_dmi_setup},
+			       {"darwin", FU_CONTEXT_HWID_FLAG_LOAD_DARWIN, fu_hwids_darwin_setup},
+			       {NULL, FU_CONTEXT_HWID_FLAG_NONE, NULL}};
 
 	g_return_val_if_fail(FU_IS_CONTEXT(self), FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
 
-	if (!fu_smbios_setup(priv->smbios, &error_smbios))
-		g_warning("Failed to load SMBIOS: %s", error_smbios->message);
-	if (!fu_hwids_setup(priv->hwids, priv->smbios, &error_hwids))
+	/* progress */
+	fu_progress_set_id(progress, G_STRLOC);
+	fu_progress_add_step(progress, FWUPD_STATUS_LOADING, 1, "hwids-setup-funcs");
+	fu_progress_add_step(progress, FWUPD_STATUS_LOADING, 1, "hwids-setup");
+	fu_progress_add_step(progress, FWUPD_STATUS_LOADING, 3, "set-flags");
+	fu_progress_add_step(progress, FWUPD_STATUS_LOADING, 95, "reload-bios-settings");
+
+	/* required always */
+	if (!fu_config_load(priv->config, error))
+		return FALSE;
+
+	/* run all the HWID setup funcs */
+	for (guint i = 0; hwids_setup_map[i].name != NULL; i++) {
+		if ((flags & hwids_setup_map[i].flag) > 0) {
+			g_autoptr(GError) error_local = NULL;
+			if (!hwids_setup_map[i].func(self, priv->hwids, &error_local)) {
+				g_info("failed to load %s: %s",
+				       hwids_setup_map[i].name,
+				       error_local->message);
+				continue;
+			}
+		}
+	}
+	fu_context_add_flag(self, FU_CONTEXT_FLAG_LOADED_HWINFO);
+	fu_progress_step_done(progress);
+
+	if (!fu_hwids_setup(priv->hwids, &error_hwids))
 		g_warning("Failed to load HWIDs: %s", error_hwids->message);
+	fu_progress_step_done(progress);
 
 	/* set the hwid flags */
 	guids = fu_context_get_hwid_guids(self);
 	for (guint i = 0; i < guids->len; i++) {
 		const gchar *guid = g_ptr_array_index(guids, i);
-		const gchar *value;
-
-		/* does prefixed quirk exist */
-		value = fu_context_lookup_quirk_by_id(self, guid, FU_QUIRKS_FLAGS);
-		if (value != NULL) {
-			g_auto(GStrv) values = g_strsplit(value, ",", -1);
-			for (guint j = 0; values[j] != NULL; j++)
-				g_hash_table_add(priv->hwid_flags, g_strdup(values[j]));
-		}
+		fu_context_lookup_quirk_by_id_iter(self,
+						   guid,
+						   FU_QUIRKS_FLAGS,
+						   fu_context_hwid_quirk_cb,
+						   NULL);
 	}
+	fu_progress_step_done(progress);
 
-	fu_context_add_udev_subsystem(self, "firmware-attributes");
+	fu_context_add_udev_subsystem(self, "firmware-attributes", NULL);
 	if (!fu_context_reload_bios_settings(self, &error_bios_settings))
 		g_debug("%s", error_bios_settings->message);
+	fu_progress_step_done(progress);
 
 	/* always */
 	return TRUE;
@@ -676,42 +1083,42 @@ fu_context_load_quirks(FuContext *self, FuQuirksLoadFlags flags, GError **error)
 }
 
 /**
- * fu_context_get_battery_state:
+ * fu_context_get_power_state:
  * @self: a #FuContext
  *
  * Gets if the system is on battery power, e.g. UPS or laptop battery.
  *
- * Returns: a battery state, e.g. %FU_BATTERY_STATE_DISCHARGING
+ * Returns: a power state, e.g. %FU_POWER_STATE_BATTERY_DISCHARGING
  *
- * Since: 1.6.0
+ * Since: 1.8.11
  **/
-FuBatteryState
-fu_context_get_battery_state(FuContext *self)
+FuPowerState
+fu_context_get_power_state(FuContext *self)
 {
 	FuContextPrivate *priv = GET_PRIVATE(self);
 	g_return_val_if_fail(FU_IS_CONTEXT(self), FALSE);
-	return priv->battery_state;
+	return priv->power_state;
 }
 
 /**
- * fu_context_set_battery_state:
+ * fu_context_set_power_state:
  * @self: a #FuContext
- * @battery_state: a battery state, e.g. %FU_BATTERY_STATE_DISCHARGING
+ * @power_state: a power state, e.g. %FU_POWER_STATE_BATTERY_DISCHARGING
  *
  * Sets if the system is on battery power, e.g. UPS or laptop battery.
  *
- * Since: 1.6.0
+ * Since: 1.8.11
  **/
 void
-fu_context_set_battery_state(FuContext *self, FuBatteryState battery_state)
+fu_context_set_power_state(FuContext *self, FuPowerState power_state)
 {
 	FuContextPrivate *priv = GET_PRIVATE(self);
 	g_return_if_fail(FU_IS_CONTEXT(self));
-	if (priv->battery_state == battery_state)
+	if (priv->power_state == power_state)
 		return;
-	priv->battery_state = battery_state;
-	g_debug("battery state now %s", fu_battery_state_to_string(battery_state));
-	g_object_notify(G_OBJECT(self), "battery-state");
+	priv->power_state = power_state;
+	g_info("power state now %s", fu_power_state_to_string(power_state));
+	g_object_notify(G_OBJECT(self), "power-state");
 }
 
 /**
@@ -720,7 +1127,7 @@ fu_context_set_battery_state(FuContext *self, FuBatteryState battery_state)
  *
  * Gets the laptop lid state, if applicable.
  *
- * Returns: a battery state, e.g. %FU_LID_STATE_CLOSED
+ * Returns: a lid state, e.g. %FU_LID_STATE_CLOSED
  *
  * Since: 1.7.4
  **/
@@ -735,7 +1142,7 @@ fu_context_get_lid_state(FuContext *self)
 /**
  * fu_context_set_lid_state:
  * @self: a #FuContext
- * @lid_state: a battery state, e.g. %FU_LID_STATE_CLOSED
+ * @lid_state: a lid state, e.g. %FU_LID_STATE_CLOSED
  *
  * Sets the laptop lid state, if applicable.
  *
@@ -749,10 +1156,49 @@ fu_context_set_lid_state(FuContext *self, FuLidState lid_state)
 	if (priv->lid_state == lid_state)
 		return;
 	priv->lid_state = lid_state;
-	g_debug("lid state now %s", fu_lid_state_to_string(lid_state));
+	g_info("lid state now %s", fu_lid_state_to_string(lid_state));
 	g_object_notify(G_OBJECT(self), "lid-state");
 }
 
+/**
+ * fu_context_get_display_state:
+ * @self: a #FuContext
+ *
+ * Gets the display state, if applicable.
+ *
+ * Returns: a display state, e.g. %FU_DISPLAY_STATE_CONNECTED
+ *
+ * Since: 1.9.6
+ **/
+FuDisplayState
+fu_context_get_display_state(FuContext *self)
+{
+	FuContextPrivate *priv = GET_PRIVATE(self);
+	g_return_val_if_fail(FU_IS_CONTEXT(self), FALSE);
+	return priv->display_state;
+}
+
+/**
+ * fu_context_set_display_state:
+ * @self: a #FuContext
+ * @display_state: a display state, e.g. %FU_DISPLAY_STATE_CONNECTED
+ *
+ * Sets the display state, if applicable.
+ *
+ * Since: 1.9.6
+ **/
+void
+fu_context_set_display_state(FuContext *self, FuDisplayState display_state)
+{
+	FuContextPrivate *priv = GET_PRIVATE(self);
+	g_return_if_fail(FU_IS_CONTEXT(self));
+	if (priv->display_state == display_state)
+		return;
+	priv->display_state = display_state;
+	g_info("display-state now %s", fu_display_state_to_string(display_state));
+	g_object_notify(G_OBJECT(self), "display-state");
+}
+
 /**
  * fu_context_get_battery_level:
  * @self: a #FuContext
@@ -789,7 +1235,7 @@ fu_context_set_battery_level(FuContext *self, guint battery_level)
 	if (priv->battery_level == battery_level)
 		return;
 	priv->battery_level = battery_level;
-	g_debug("battery level now %u", battery_level);
+	g_info("battery level now %u", battery_level);
 	g_object_notify(G_OBJECT(self), "battery-level");
 }
 
@@ -829,7 +1275,7 @@ fu_context_set_battery_threshold(FuContext *self, guint battery_threshold)
 	if (priv->battery_threshold == battery_threshold)
 		return;
 	priv->battery_threshold = battery_threshold;
-	g_debug("battery threshold now %u", battery_threshold);
+	g_info("battery threshold now %u", battery_threshold);
 	g_object_notify(G_OBJECT(self), "battery-threshold");
 }
 
@@ -847,7 +1293,30 @@ fu_context_add_flag(FuContext *context, FuContextFlags flag)
 {
 	FuContextPrivate *priv = GET_PRIVATE(context);
 	g_return_if_fail(FU_IS_CONTEXT(context));
+	if (priv->flags & flag)
+		return;
 	priv->flags |= flag;
+	g_object_notify(G_OBJECT(context), "flags");
+}
+
+/**
+ * fu_context_remove_flag:
+ * @context: a #FuContext
+ * @flag: the context flag
+ *
+ * Removes a specific flag from the context.
+ *
+ * Since: 1.8.10
+ **/
+void
+fu_context_remove_flag(FuContext *context, FuContextFlags flag)
+{
+	FuContextPrivate *priv = GET_PRIVATE(context);
+	g_return_if_fail(FU_IS_CONTEXT(context));
+	if ((priv->flags & flag) == 0)
+		return;
+	priv->flags &= ~flag;
+	g_object_notify(G_OBJECT(context), "flags");
 }
 
 /**
@@ -899,6 +1368,39 @@ fu_context_add_esp_volume(FuContext *self, FuVolume *volume)
 	g_ptr_array_add(priv->esp_volumes, g_object_ref(volume));
 }
 
+/**
+ * fu_context_set_esp_location:
+ * @self: A #FuContext object.
+ * @location: The path to the preferred ESP.
+ *
+ * Sets the user's desired ESP (EFI System Partition) location for the given #FuContext.
+ */
+void
+fu_context_set_esp_location(FuContext *self, const gchar *location)
+{
+	FuContextPrivate *priv = GET_PRIVATE(self);
+	g_return_if_fail(FU_IS_CONTEXT(self));
+	g_return_if_fail(location != NULL);
+	g_free(priv->esp_location);
+	priv->esp_location = g_strdup(location);
+}
+
+/**
+ * fu_context_get_esp_location:
+ * @self: The FuContext object.
+ *
+ * Retrieves the user's desired ESP (EFI System Partition) location for the given #FuContext
+ *
+ * Return: The preferred ESP location as a string.
+ */
+const gchar *
+fu_context_get_esp_location(FuContext *self)
+{
+	FuContextPrivate *priv = GET_PRIVATE(self);
+	g_return_val_if_fail(FU_IS_CONTEXT(self), NULL);
+	return priv->esp_location;
+}
+
 /**
  * fu_context_get_esp_volumes:
  * @self: a #FuContext
@@ -944,6 +1446,9 @@ fu_context_get_esp_volumes(FuContext *self, GError **error)
 	} else {
 		for (guint i = 0; i < volumes_esp->len; i++) {
 			FuVolume *vol = g_ptr_array_index(volumes_esp, i);
+			g_autofree gchar *type = fu_volume_get_id_type(vol);
+			if (g_strcmp0(type, "vfat") != 0)
+				continue;
 			fu_context_add_esp_volume(self, vol);
 		}
 	}
@@ -966,7 +1471,26 @@ fu_context_get_esp_volumes(FuContext *self, GError **error)
 
 	/* nothing found */
 	if (priv->esp_volumes->len == 0) {
-		g_set_error_literal(error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND, "No ESP or BDP found");
+		g_autofree gchar *base = NULL;
+		g_autofree gchar *path = NULL;
+		g_autoptr(GPtrArray) devices = NULL;
+
+		/* udisks2 is working */
+		devices = fu_common_get_block_devices(NULL);
+		if (devices != NULL) {
+			g_set_error_literal(error,
+					    G_IO_ERROR,
+					    G_IO_ERROR_NOT_FOUND,
+					    "No ESP or BDP found");
+			return NULL;
+		}
+		base = fu_path_from_kind(FU_PATH_KIND_SYSCONFDIR_PKG);
+		path = g_build_filename(base, "fwupd.conf", NULL);
+		g_set_error(error,
+			    G_IO_ERROR,
+			    G_IO_ERROR_NOT_FOUND,
+			    "No valid 'EspLocation' specified in %s",
+			    path);
 		return NULL;
 	}
 
@@ -980,18 +1504,24 @@ fu_context_get_property(GObject *object, guint prop_id, GValue *value, GParamSpe
 	FuContext *self = FU_CONTEXT(object);
 	FuContextPrivate *priv = GET_PRIVATE(self);
 	switch (prop_id) {
-	case PROP_BATTERY_STATE:
-		g_value_set_uint(value, priv->battery_state);
+	case PROP_POWER_STATE:
+		g_value_set_uint(value, priv->power_state);
 		break;
 	case PROP_LID_STATE:
 		g_value_set_uint(value, priv->lid_state);
 		break;
+	case PROP_DISPLAY_STATE:
+		g_value_set_uint(value, priv->display_state);
+		break;
 	case PROP_BATTERY_LEVEL:
 		g_value_set_uint(value, priv->battery_level);
 		break;
 	case PROP_BATTERY_THRESHOLD:
 		g_value_set_uint(value, priv->battery_threshold);
 		break;
+	case PROP_FLAGS:
+		g_value_set_uint64(value, priv->flags);
+		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
 		break;
@@ -1002,19 +1532,26 @@ static void
 fu_context_set_property(GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
 {
 	FuContext *self = FU_CONTEXT(object);
+	FuContextPrivate *priv = GET_PRIVATE(self);
 	switch (prop_id) {
-	case PROP_BATTERY_STATE:
-		fu_context_set_battery_state(self, g_value_get_uint(value));
+	case PROP_POWER_STATE:
+		fu_context_set_power_state(self, g_value_get_uint(value));
 		break;
 	case PROP_LID_STATE:
 		fu_context_set_lid_state(self, g_value_get_uint(value));
 		break;
+	case PROP_DISPLAY_STATE:
+		fu_context_set_display_state(self, g_value_get_uint(value));
+		break;
 	case PROP_BATTERY_LEVEL:
 		fu_context_set_battery_level(self, g_value_get_uint(value));
 		break;
 	case PROP_BATTERY_THRESHOLD:
 		fu_context_set_battery_threshold(self, g_value_get_uint(value));
 		break;
+	case PROP_FLAGS:
+		priv->flags = g_value_get_uint64(value);
+		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
 		break;
@@ -1027,17 +1564,19 @@ fu_context_finalize(GObject *object)
 	FuContext *self = FU_CONTEXT(object);
 	FuContextPrivate *priv = GET_PRIVATE(self);
 
-	if (priv->runtime_versions != NULL)
-		g_hash_table_unref(priv->runtime_versions);
-	if (priv->compile_versions != NULL)
-		g_hash_table_unref(priv->compile_versions);
+	if (priv->fdt != NULL)
+		g_object_unref(priv->fdt);
+	g_free(priv->esp_location);
+	g_hash_table_unref(priv->runtime_versions);
+	g_hash_table_unref(priv->compile_versions);
 	g_object_unref(priv->hwids);
+	g_object_unref(priv->config);
 	g_hash_table_unref(priv->hwid_flags);
 	g_object_unref(priv->quirks);
 	g_object_unref(priv->smbios);
 	g_object_unref(priv->host_bios_settings);
 	g_hash_table_unref(priv->firmware_gtypes);
-	g_ptr_array_unref(priv->udev_subsystems);
+	g_hash_table_unref(priv->udev_subsystems);
 	g_ptr_array_unref(priv->esp_volumes);
 
 	G_OBJECT_CLASS(fu_context_parent_class)->finalize(object);
@@ -1053,20 +1592,20 @@ fu_context_class_init(FuContextClass *klass)
 	object_class->set_property = fu_context_set_property;
 
 	/**
-	 * FuContext:battery-state:
+	 * FuContext:power-state:
 	 *
-	 * The system battery state.
+	 * The system power state.
 	 *
-	 * Since: 1.6.0
+	 * Since: 1.8.11
 	 */
-	pspec = g_param_spec_uint("battery-state",
+	pspec = g_param_spec_uint("power-state",
 				  NULL,
 				  NULL,
-				  FU_BATTERY_STATE_UNKNOWN,
-				  FU_BATTERY_STATE_LAST,
-				  FU_BATTERY_STATE_UNKNOWN,
+				  FU_POWER_STATE_UNKNOWN,
+				  FU_POWER_STATE_LAST,
+				  FU_POWER_STATE_UNKNOWN,
 				  G_PARAM_READWRITE | G_PARAM_STATIC_NAME);
-	g_object_class_install_property(object_class, PROP_BATTERY_STATE, pspec);
+	g_object_class_install_property(object_class, PROP_POWER_STATE, pspec);
 
 	/**
 	 * FuContext:lid-state:
@@ -1084,6 +1623,22 @@ fu_context_class_init(FuContextClass *klass)
 				  G_PARAM_READWRITE | G_PARAM_STATIC_NAME);
 	g_object_class_install_property(object_class, PROP_LID_STATE, pspec);
 
+	/**
+	 * FuContext:display-state:
+	 *
+	 * The display state.
+	 *
+	 * Since: 1.9.6
+	 */
+	pspec = g_param_spec_uint("display-state",
+				  NULL,
+				  NULL,
+				  FU_DISPLAY_STATE_UNKNOWN,
+				  FU_DISPLAY_STATE_LAST,
+				  FU_DISPLAY_STATE_UNKNOWN,
+				  G_PARAM_READWRITE | G_PARAM_STATIC_NAME);
+	g_object_class_install_property(object_class, PROP_DISPLAY_STATE, pspec);
+
 	/**
 	 * FuContext:battery-level:
 	 *
@@ -1116,6 +1671,22 @@ fu_context_class_init(FuContextClass *klass)
 				  G_PARAM_READWRITE | G_PARAM_STATIC_NAME);
 	g_object_class_install_property(object_class, PROP_BATTERY_THRESHOLD, pspec);
 
+	/**
+	 * FuContext:flags:
+	 *
+	 * The context flags.
+	 *
+	 * Since: 1.8.10
+	 */
+	pspec = g_param_spec_uint64("flags",
+				    NULL,
+				    NULL,
+				    FU_CONTEXT_FLAG_NONE,
+				    G_MAXUINT64,
+				    FU_CONTEXT_FLAG_NONE,
+				    G_PARAM_READWRITE | G_PARAM_STATIC_NAME);
+	g_object_class_install_property(object_class, PROP_FLAGS, pspec);
+
 	/**
 	 * FuContext::security-changed:
 	 * @self: the #FuContext instance that emitted the signal
@@ -1143,16 +1714,23 @@ static void
 fu_context_init(FuContext *self)
 {
 	FuContextPrivate *priv = GET_PRIVATE(self);
+	priv->chassis_kind = FU_SMBIOS_CHASSIS_KIND_UNKNOWN;
 	priv->battery_level = FWUPD_BATTERY_LEVEL_INVALID;
 	priv->battery_threshold = FWUPD_BATTERY_LEVEL_INVALID;
 	priv->smbios = fu_smbios_new();
 	priv->hwids = fu_hwids_new();
+	priv->config = fu_config_new();
 	priv->hwid_flags = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);
-	priv->udev_subsystems = g_ptr_array_new_with_free_func(g_free);
+	priv->udev_subsystems = g_hash_table_new_full(g_str_hash,
+						      g_str_equal,
+						      g_free,
+						      (GDestroyNotify)g_ptr_array_unref);
 	priv->firmware_gtypes = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);
 	priv->quirks = fu_quirks_new();
 	priv->host_bios_settings = fu_bios_settings_new();
 	priv->esp_volumes = g_ptr_array_new_with_free_func((GDestroyNotify)g_object_unref);
+	priv->runtime_versions = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
+	priv->compile_versions = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
 }
 
 /**
