@@ -11,16 +11,15 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <fwupd.h>
-#include <glib/gstdio.h>
 #include <gmodule.h>
 #include <string.h>
 #include <unistd.h>
 
 #include "fu-bytes.h"
+#include "fu-config-private.h"
 #include "fu-context-private.h"
 #include "fu-device-private.h"
 #include "fu-kernel.h"
-#include "fu-mutex.h"
 #include "fu-path.h"
 #include "fu-plugin-private.h"
 #include "fu-security-attr.h"
@@ -47,9 +46,9 @@ typedef struct {
 	GHashTable *runtime_versions;
 	GHashTable *compile_versions;
 	FuContext *ctx;
-	GArray *device_gtypes; /* (nullable): of #GType */
-	GHashTable *cache;     /* (nullable): platform_id:GObject */
-	GRWLock cache_mutex;
+	GArray *device_gtypes;	     /* (nullable): of #GType */
+	GType device_gtype_default;
+	GHashTable *cache;	     /* (nullable): platform_id:GObject */
 	GHashTable *report_metadata; /* (nullable): key:value */
 	GFileMonitor *config_monitor;
 	FuPluginData *data;
@@ -63,7 +62,6 @@ enum {
 	SIGNAL_DEVICE_REMOVED,
 	SIGNAL_DEVICE_REGISTER,
 	SIGNAL_RULES_CHANGED,
-	SIGNAL_CONFIG_CHANGED,
 	SIGNAL_CHECK_SUPPORTED,
 	SIGNAL_LAST
 };
@@ -86,9 +84,6 @@ typedef gboolean (*FuPluginFlaggedDeviceFunc)(FuPlugin *self,
 					      GError **error);
 typedef gboolean (*FuPluginDeviceArrayFunc)(FuPlugin *self, GPtrArray *devices, GError **error);
 
-#define FU_PLUGIN_FILE_MODE_NONSECURE 0644
-#define FU_PLUGIN_FILE_MODE_SECURE    0640
-
 /**
  * fu_plugin_is_open:
  * @self: a #FuPlugin
@@ -135,7 +130,20 @@ fu_plugin_get_name(FuPlugin *self)
 void
 fu_plugin_set_name(FuPlugin *self, const gchar *name)
 {
+	FuPluginPrivate *priv = GET_PRIVATE(self);
+
 	g_return_if_fail(FU_IS_PLUGIN(self));
+	g_return_if_fail(!priv->done_init);
+
+	if (g_strcmp0(name, fwupd_plugin_get_name(FWUPD_PLUGIN(self))) == 0) {
+		g_critical("plugin name set to original value: %s", name);
+		return;
+	}
+	if (fwupd_plugin_get_name(FWUPD_PLUGIN(self)) != NULL) {
+		g_debug("overwriting plugin name %s -> %s",
+			fwupd_plugin_get_name(FWUPD_PLUGIN(self)),
+			name);
+	}
 	fwupd_plugin_set_name(FWUPD_PLUGIN(self), name);
 }
 
@@ -163,10 +171,8 @@ gpointer
 fu_plugin_cache_lookup(FuPlugin *self, const gchar *id)
 {
 	FuPluginPrivate *priv = GET_PRIVATE(self);
-	g_autoptr(GRWLockReaderLocker) locker = g_rw_lock_reader_locker_new(&priv->cache_mutex);
 	g_return_val_if_fail(FU_IS_PLUGIN(self), NULL);
 	g_return_val_if_fail(id != NULL, NULL);
-	g_return_val_if_fail(locker != NULL, NULL);
 	if (priv->cache == NULL)
 		return NULL;
 	return g_hash_table_lookup(priv->cache, id);
@@ -186,11 +192,9 @@ void
 fu_plugin_cache_add(FuPlugin *self, const gchar *id, gpointer dev)
 {
 	FuPluginPrivate *priv = GET_PRIVATE(self);
-	g_autoptr(GRWLockWriterLocker) locker = g_rw_lock_writer_locker_new(&priv->cache_mutex);
 	g_return_if_fail(FU_IS_PLUGIN(self));
 	g_return_if_fail(id != NULL);
 	g_return_if_fail(G_IS_OBJECT(dev));
-	g_return_if_fail(locker != NULL);
 	if (priv->cache == NULL) {
 		priv->cache = g_hash_table_new_full(g_str_hash,
 						    g_str_equal,
@@ -213,10 +217,8 @@ void
 fu_plugin_cache_remove(FuPlugin *self, const gchar *id)
 {
 	FuPluginPrivate *priv = GET_PRIVATE(self);
-	g_autoptr(GRWLockWriterLocker) locker = g_rw_lock_writer_locker_new(&priv->cache_mutex);
 	g_return_if_fail(FU_IS_PLUGIN(self));
 	g_return_if_fail(id != NULL);
-	g_return_if_fail(locker != NULL);
 	if (priv->cache == NULL)
 		return;
 	g_hash_table_remove(priv->cache, id);
@@ -288,14 +290,6 @@ fu_plugin_guess_name_from_fn(const gchar *filename)
 	return name;
 }
 
-static gchar *
-fu_plugin_get_config_filename(FuPlugin *self)
-{
-	g_autofree gchar *conf_dir = fu_path_from_kind(FU_PATH_KIND_SYSCONFDIR_PKG);
-	g_autofree gchar *conf_file = g_strdup_printf("%s.conf", fu_plugin_get_name(self));
-	return g_build_filename(conf_dir, conf_file, NULL);
-}
-
 /**
  * fu_plugin_open:
  * @self: a #FuPlugin
@@ -356,23 +350,6 @@ fu_plugin_open(FuPlugin *self, const gchar *filename, GError **error)
 		fu_plugin_set_name(self, str);
 	}
 
-	/* ensure the configure file is set to the correct permission */
-	if (fu_plugin_has_flag(self, FWUPD_PLUGIN_FLAG_SECURE_CONFIG)) {
-		g_autofree gchar *conf_path = fu_plugin_get_config_filename(self);
-		if (g_file_test(conf_path, G_FILE_TEST_EXISTS)) {
-			gint rc = g_chmod(conf_path, FU_PLUGIN_FILE_MODE_SECURE);
-			if (rc != 0) {
-				g_set_error(error,
-					    G_IO_ERROR,
-					    G_IO_ERROR_FAILED,
-					    "failed to change permission of %s",
-					    filename);
-				fu_plugin_add_flag(self, FWUPD_PLUGIN_FLAG_FAILED_OPEN);
-				return FALSE;
-			}
-		}
-	}
-
 	/* optional */
 	if (vfuncs->load != NULL) {
 		FuContext *ctx = fu_plugin_get_context(self);
@@ -432,6 +409,12 @@ fu_plugin_add_string(FuPlugin *self, guint idt, GString *str)
 		fu_string_append_ku(str, idt + 1, "Order", priv->order);
 	if (priv->priority != 0)
 		fu_string_append_ku(str, idt + 1, "Priority", priv->priority);
+	if (priv->device_gtype_default != G_TYPE_INVALID) {
+		fu_string_append(str,
+				 idt,
+				 "DeviceGTypeDefault",
+				 g_type_name(priv->device_gtype_default));
+	}
 
 	/* optional */
 	if (vfuncs->to_string != NULL)
@@ -507,19 +490,6 @@ fu_plugin_device_child_removed_cb(FuDevice *device, FuDevice *child, FuPlugin *s
 	fu_plugin_device_remove(self, child);
 }
 
-static void
-fu_plugin_config_monitor_changed_cb(GFileMonitor *monitor,
-				    GFile *file,
-				    GFile *other_file,
-				    GFileMonitorEvent event_type,
-				    gpointer user_data)
-{
-	FuPlugin *self = FU_PLUGIN(user_data);
-	g_autofree gchar *fn = g_file_get_path(file);
-	g_debug("%s changed, sending signal", fn);
-	g_signal_emit(self, signals[SIGNAL_CONFIG_CHANGED], 0);
-}
-
 /**
  * fu_plugin_device_add:
  * @self: a #FuPlugin
@@ -565,7 +535,8 @@ fu_plugin_device_add(FuPlugin *self, FuDevice *device)
 	}
 
 	g_debug("emit added from %s: %s", fu_plugin_get_name(self), fu_device_get_id(device));
-	fu_device_set_created(device, (guint64)g_get_real_time() / G_USEC_PER_SEC);
+	if (fu_device_get_created(device) == 0)
+		fu_device_set_created(device, (guint64)g_get_real_time() / G_USEC_PER_SEC);
 	fu_device_set_plugin(device, fu_plugin_get_name(self));
 	g_signal_emit(self, signals[SIGNAL_DEVICE_ADDED], 0, device);
 
@@ -894,11 +865,8 @@ fu_plugin_device_read_firmware(FuPlugin *self,
 gboolean
 fu_plugin_runner_startup(FuPlugin *self, FuProgress *progress, GError **error)
 {
-	FuPluginPrivate *priv = GET_PRIVATE(self);
 	FuPluginVfuncs *vfuncs = fu_plugin_get_vfuncs(self);
-	g_autofree gchar *config_filename = fu_plugin_get_config_filename(self);
 	g_autoptr(GError) error_local = NULL;
-	g_autoptr(GFile) file = g_file_new_for_path(config_filename);
 
 	g_return_val_if_fail(FU_IS_PLUGIN(self), FALSE);
 
@@ -913,12 +881,61 @@ fu_plugin_runner_startup(FuPlugin *self, FuProgress *progress, GError **error)
 		return TRUE;
 
 	/* optional */
-	if (vfuncs->startup == NULL)
+	if (vfuncs->startup != NULL) {
+		g_debug("startup(%s)", fu_plugin_get_name(self));
+		if (!vfuncs->startup(self, progress, &error_local)) {
+			if (error_local == NULL) {
+				g_critical("unset plugin error in startup(%s)",
+					   fu_plugin_get_name(self));
+				g_set_error_literal(&error_local,
+						    FWUPD_ERROR,
+						    FWUPD_ERROR_INTERNAL,
+						    "unspecified error");
+			}
+			g_propagate_prefixed_error(error,
+						   g_steal_pointer(&error_local),
+						   "failed to startup using %s: ",
+						   fu_plugin_get_name(self));
+			return FALSE;
+		}
+	}
+
+	/* success */
+	return TRUE;
+}
+
+/**
+ * fu_plugin_runner_ready:
+ * @self: a #FuPlugin
+ * @error: (nullable): optional return location for an error
+ *
+ * Runs the ready routine for the plugin
+ *
+ * Returns: #TRUE for success, #FALSE for failure
+ *
+ * Since: 1.9.6
+ **/
+gboolean
+fu_plugin_runner_ready(FuPlugin *self, FuProgress *progress, GError **error)
+{
+	FuPluginVfuncs *vfuncs = fu_plugin_get_vfuncs(self);
+	g_autoptr(GError) error_local = NULL;
+
+	g_return_val_if_fail(FU_IS_PLUGIN(self), FALSE);
+
+	/* progress */
+	fu_progress_set_name(progress, fu_plugin_get_name(self));
+	if (fu_plugin_has_flag(self, FWUPD_PLUGIN_FLAG_DISABLED))
 		return TRUE;
-	g_debug("startup(%s)", fu_plugin_get_name(self));
-	if (!vfuncs->startup(self, progress, &error_local)) {
+	fu_plugin_add_flag(self, FWUPD_PLUGIN_FLAG_READY);
+	if (vfuncs->ready == NULL)
+		return TRUE;
+
+	/* optional */
+	g_debug("ready(%s)", fu_plugin_get_name(self));
+	if (!vfuncs->ready(self, progress, &error_local)) {
 		if (error_local == NULL) {
-			g_critical("unset plugin error in startup(%s)", fu_plugin_get_name(self));
+			g_critical("unset plugin error in ready(%s)", fu_plugin_get_name(self));
 			g_set_error_literal(&error_local,
 					    FWUPD_ERROR,
 					    FWUPD_ERROR_INTERNAL,
@@ -926,20 +943,11 @@ fu_plugin_runner_startup(FuPlugin *self, FuProgress *progress, GError **error)
 		}
 		g_propagate_prefixed_error(error,
 					   g_steal_pointer(&error_local),
-					   "failed to startup using %s: ",
+					   "failed to ready using %s: ",
 					   fu_plugin_get_name(self));
 		return FALSE;
 	}
 
-	/* create a monitor on the config file */
-	priv->config_monitor = g_file_monitor_file(file, G_FILE_MONITOR_NONE, NULL, error);
-	if (priv->config_monitor == NULL)
-		return FALSE;
-	g_signal_connect(G_FILE_MONITOR(priv->config_monitor),
-			 "changed",
-			 G_CALLBACK(fu_plugin_config_monitor_changed_cb),
-			 self);
-
 	/* success */
 	return TRUE;
 }
@@ -1257,7 +1265,6 @@ fu_plugin_runner_prepare(FuPlugin *self,
 			 GError **error)
 {
 	FuPluginVfuncs *vfuncs = fu_plugin_get_vfuncs(self);
-	fu_device_add_backend_tag(device, "prepare");
 	return fu_plugin_runner_flagged_device_generic(self,
 						       device,
 						       progress,
@@ -1289,7 +1296,6 @@ fu_plugin_runner_cleanup(FuPlugin *self,
 			 GError **error)
 {
 	FuPluginVfuncs *vfuncs = fu_plugin_get_vfuncs(self);
-	fu_device_add_backend_tag(device, "cleanup");
 	return fu_plugin_runner_flagged_device_generic(self,
 						       device,
 						       progress,
@@ -1316,7 +1322,6 @@ gboolean
 fu_plugin_runner_attach(FuPlugin *self, FuDevice *device, FuProgress *progress, GError **error)
 {
 	FuPluginVfuncs *vfuncs = fu_plugin_get_vfuncs(self);
-	fu_device_add_backend_tag(device, "attach");
 	return fu_plugin_runner_device_generic_progress(
 	    self,
 	    device,
@@ -1343,7 +1348,6 @@ gboolean
 fu_plugin_runner_detach(FuPlugin *self, FuDevice *device, FuProgress *progress, GError **error)
 {
 	FuPluginVfuncs *vfuncs = fu_plugin_get_vfuncs(self);
-	fu_device_add_backend_tag(device, "detach");
 	return fu_plugin_runner_device_generic_progress(
 	    self,
 	    device,
@@ -1379,10 +1383,35 @@ fu_plugin_runner_reload(FuPlugin *self, FuDevice *device, GError **error)
 	locker = fu_device_locker_new(proxy, error);
 	if (locker == NULL)
 		return FALSE;
-	fu_device_add_backend_tag(device, "reload");
 	return fu_device_reload(device, error);
 }
 
+/**
+ * fu_plugin_runner_reboot_cleanup:
+ * @self: a #FuPlugin
+ * @device: a device
+ * @error: (nullable): optional return location for an error
+ *
+ * Performs cleanup actions after the reboot has been performed.
+ *
+ * Returns: #TRUE for success, #FALSE for failure
+ *
+ * Since: 1.9.7
+ **/
+gboolean
+fu_plugin_runner_reboot_cleanup(FuPlugin *self, FuDevice *device, GError **error)
+{
+	FuPluginVfuncs *vfuncs = fu_plugin_get_vfuncs(self);
+
+	/* optional */
+	if (fu_plugin_has_flag(self, FWUPD_PLUGIN_FLAG_DISABLED))
+		return TRUE;
+	if (vfuncs->reboot_cleanup == NULL)
+		return TRUE;
+	g_debug("reboot_cleanup(%s)", fu_plugin_get_name(self));
+	return vfuncs->reboot_cleanup(self, device, error);
+}
+
 /**
  * fu_plugin_runner_add_security_attrs:
  * @self: a #FuPlugin
@@ -1416,6 +1445,8 @@ fu_plugin_runner_add_security_attrs(FuPlugin *self, FuSecurityAttrs *attrs)
  *
  * Plugins can use this method only in fu_plugin_init()
  *
+ * See also: fu_plugin_set_device_gtype_default()
+ *
  * Since: 1.6.0
  **/
 void
@@ -1423,15 +1454,75 @@ fu_plugin_add_device_gtype(FuPlugin *self, GType device_gtype)
 {
 	FuPluginPrivate *priv = GET_PRIVATE(self);
 
+	g_return_if_fail(FU_IS_PLUGIN(self));
+
 	/* create as required */
 	if (priv->device_gtypes == NULL)
 		priv->device_gtypes = g_array_new(FALSE, FALSE, sizeof(GType));
 
+	/* check for duplicates */
+	for (guint i = 0; i < priv->device_gtypes->len; i++) {
+		GType device_gtype_tmp = g_array_index(priv->device_gtypes, GType, i);
+		if (device_gtype == device_gtype_tmp)
+			return;
+	}
+
 	/* ensure (to allow quirks to use it) then add */
 	g_type_ensure(device_gtype);
 	g_array_append_val(priv->device_gtypes, device_gtype);
 }
 
+/**
+ * fu_plugin_get_device_gtype_default:
+ * @self: a #FuPlugin
+ *
+ * Gets the default device #GType.
+ *
+ * If there is only one possible #GType added from fu_plugin_add_device_gtype() it will also be
+ * returned here.
+ *
+ * Returns: a #GType, or %G_TYPE_INVALID on error
+ *
+ * Since: 1.9.14
+ **/
+GType
+fu_plugin_get_device_gtype_default(FuPlugin *self)
+{
+	FuPluginPrivate *priv = GET_PRIVATE(self);
+
+	g_return_val_if_fail(FU_IS_PLUGIN(self), G_TYPE_INVALID);
+
+	if (priv->device_gtype_default != G_TYPE_INVALID)
+		return priv->device_gtype_default;
+	if (priv->device_gtypes->len == 1)
+		return g_array_index(priv->device_gtypes, GType, 0);
+	return G_TYPE_INVALID;
+}
+
+/**
+ * fu_plugin_set_device_gtype_default:
+ * @self: a #FuPlugin
+ * @device_gtype: a #GType, e.g. `FU_TYPE_DEVICE`
+ *
+ * Sets the default device #GType.
+ *
+ * This will also add the device #GType using fu_plugin_add_device_gtype().
+ *
+ * Plugins can use this method only in fu_plugin_init()
+ *
+ * Since: 1.9.14
+ **/
+void
+fu_plugin_set_device_gtype_default(FuPlugin *self, GType device_gtype)
+{
+	FuPluginPrivate *priv = GET_PRIVATE(self);
+
+	g_return_if_fail(FU_IS_PLUGIN(self));
+
+	fu_plugin_add_device_gtype(self, device_gtype);
+	priv->device_gtype_default = device_gtype;
+}
+
 static gchar *
 fu_common_string_uncamelcase(const gchar *str)
 {
@@ -1472,6 +1563,41 @@ fu_plugin_check_amdgpu_dpaux(FuPlugin *self, GError **error)
 	return TRUE;
 }
 
+/**
+ * fu_plugin_add_device_udev_subsystem:
+ * @self: a #FuPlugin
+ * @subsystem: a subsystem name, e.g. `pciport`
+ *
+ * Add this plugin as a possible handler of devices with this udev subsystem.
+ * Use fu_plugin_add_udev_subsystem() if you just want to ensure the subsystem is watched.
+ *
+ * Plugins can use this method only in fu_plugin_init()
+ *
+ * Since: 1.9.3
+ **/
+void
+fu_plugin_add_device_udev_subsystem(FuPlugin *self, const gchar *subsystem)
+{
+	FuPluginPrivate *priv = GET_PRIVATE(self);
+
+	g_return_if_fail(FU_IS_PLUGIN(self));
+	g_return_if_fail(subsystem != NULL);
+
+	/* see https://github.com/fwupd/fwupd/issues/1121 for more details */
+	if (g_strcmp0(subsystem, "drm_dp_aux_dev") == 0) {
+		g_autoptr(GError) error = NULL;
+		if (!fu_plugin_check_amdgpu_dpaux(self, &error)) {
+			g_warning("failed to add subsystem: %s", error->message);
+			fu_plugin_add_flag(self, FWUPD_PLUGIN_FLAG_DISABLED);
+			fu_plugin_add_flag(self, FWUPD_PLUGIN_FLAG_KERNEL_TOO_OLD);
+			return;
+		}
+	}
+
+	/* proxy */
+	fu_context_add_udev_subsystem(priv->ctx, subsystem, fu_plugin_get_name(self));
+}
+
 /**
  * fu_plugin_add_udev_subsystem:
  * @self: a #FuPlugin
@@ -1488,19 +1614,11 @@ fu_plugin_add_udev_subsystem(FuPlugin *self, const gchar *subsystem)
 {
 	FuPluginPrivate *priv = GET_PRIVATE(self);
 
-	/* see https://github.com/fwupd/fwupd/issues/1121 for more details */
-	if (g_strcmp0(subsystem, "drm_dp_aux_dev") == 0) {
-		g_autoptr(GError) error = NULL;
-		if (!fu_plugin_check_amdgpu_dpaux(self, &error)) {
-			g_warning("failed to add subsystem: %s", error->message);
-			fu_plugin_add_flag(self, FWUPD_PLUGIN_FLAG_DISABLED);
-			fu_plugin_add_flag(self, FWUPD_PLUGIN_FLAG_KERNEL_TOO_OLD);
-			return;
-		}
-	}
+	g_return_if_fail(FU_IS_PLUGIN(self));
+	g_return_if_fail(subsystem != NULL);
 
 	/* proxy */
-	fu_context_add_udev_subsystem(priv->ctx, subsystem);
+	fu_context_add_udev_subsystem(priv->ctx, subsystem, NULL);
 }
 
 /**
@@ -1527,7 +1645,7 @@ fu_plugin_add_firmware_gtype(FuPlugin *self, const gchar *id, GType gtype)
 		g_autoptr(GString) str = g_string_new(g_type_name(gtype));
 		if (g_str_has_prefix(str->str, "Fu"))
 			g_string_erase(str, 0, 2);
-		fu_string_replace(str, "Firmware", "");
+		g_string_replace(str, "Firmware", "", 1);
 		id_safe = fu_common_string_uncamelcase(str->str);
 	}
 	fu_context_add_firmware_gtype(priv->ctx, id_safe, gtype);
@@ -1555,6 +1673,7 @@ fu_plugin_backend_device_added(FuPlugin *self,
 	FuDevice *proxy;
 	FuPluginPrivate *priv = GET_PRIVATE(self);
 	GType device_gtype = fu_device_get_specialized_gtype(FU_DEVICE(device));
+	GType proxy_gtype = fu_device_get_proxy_gtype(FU_DEVICE(device));
 	g_autoptr(FuDevice) dev = NULL;
 	g_autoptr(FuDeviceLocker) locker = NULL;
 
@@ -1566,20 +1685,45 @@ fu_plugin_backend_device_added(FuPlugin *self,
 	fu_progress_add_step(progress, FWUPD_STATUS_LOADING, 48, "add");
 
 	/* fall back to plugin default */
+	if (device_gtype == G_TYPE_INVALID)
+		device_gtype = fu_plugin_get_device_gtype_default(self);
 	if (device_gtype == G_TYPE_INVALID) {
 		if (priv->device_gtypes->len > 1) {
-			g_set_error_literal(error,
-					    FWUPD_ERROR,
-					    FWUPD_ERROR_INTERNAL,
-					    "too many GTypes to choose a default");
+			g_autoptr(GString) str = g_string_new(NULL);
+			for (guint i = 0; i < priv->device_gtypes->len; i++) {
+				device_gtype = g_array_index(priv->device_gtypes, GType, i);
+				if (str->len > 0)
+					g_string_append(str, ",");
+				g_string_append(str, g_type_name(device_gtype));
+			}
+			g_set_error(error,
+				    FWUPD_ERROR,
+				    FWUPD_ERROR_INTERNAL,
+				    "too many GTypes to choose a default, got: %s",
+				    str->str);
 			return FALSE;
 		}
-		device_gtype = g_array_index(priv->device_gtypes, GType, 0);
+		g_set_error_literal(error,
+				    FWUPD_ERROR,
+				    FWUPD_ERROR_INTERNAL,
+				    "no possible device GTypes");
+		return FALSE;
 	}
 
 	/* create new device and incorporate existing properties */
 	dev = g_object_new(device_gtype, "context", priv->ctx, NULL);
 	fu_device_incorporate(dev, FU_DEVICE(device));
+
+	/* any proxy device to create too? */
+	if (proxy_gtype != G_TYPE_INVALID) {
+		g_autoptr(FuDevice) proxy_tmp =
+		    g_object_new(proxy_gtype, "context", priv->ctx, NULL);
+		fu_device_incorporate(proxy_tmp, device);
+		fu_device_add_internal_flag(dev, FU_DEVICE_INTERNAL_FLAG_REFCOUNTED_PROXY);
+		fu_device_set_proxy(dev, proxy_tmp);
+	}
+
+	/* notify plugins */
 	if (!fu_plugin_runner_device_created(self, dev, error))
 		return FALSE;
 	fu_progress_step_done(progress);
@@ -1591,20 +1735,22 @@ fu_plugin_backend_device_added(FuPlugin *self,
 			return FALSE;
 		fu_device_convert_instance_ids(dev);
 		if (!fu_plugin_check_supported_device(self, dev)) {
-			g_autofree gchar *guids = fu_device_get_guids_as_str(dev);
-			g_debug("%s has no updates, so ignoring device", guids);
+			GPtrArray *guids = fu_device_get_guids(dev);
+			g_autofree gchar *guids_str = fu_strjoin(",", guids);
+			g_info("%s has no updates, so ignoring device", guids_str);
 			fu_progress_finished(progress);
 			return TRUE;
 		}
 	}
 
 	/* open */
-	proxy = fu_device_get_proxy(device);
+	proxy = fu_device_get_proxy(dev);
 	if (proxy != NULL) {
 		g_autoptr(FuDeviceLocker) locker_proxy = NULL;
 		locker_proxy = fu_device_locker_new(proxy, error);
 		if (locker_proxy == NULL)
 			return FALSE;
+		fu_device_incorporate(dev, proxy);
 	}
 	locker = fu_device_locker_new(dev, error);
 	if (locker == NULL)
@@ -1663,7 +1809,7 @@ fu_plugin_runner_backend_device_added(FuPlugin *self,
 		return FALSE;
 	}
 	g_debug("backend_device_added(%s)", fu_plugin_get_name(self));
-	if (!vfuncs->backend_device_added(self, device, &error_local)) {
+	if (!vfuncs->backend_device_added(self, device, progress, &error_local)) {
 		if (error_local == NULL) {
 			g_critical("unset plugin error in backend_device_added(%s)",
 				   fu_plugin_get_name(self));
@@ -1974,7 +2120,6 @@ fu_plugin_runner_activate(FuPlugin *self, FuDevice *device, FuProgress *progress
 	}
 
 	/* run vfunc */
-	fu_device_add_backend_tag(device, "activate");
 	if (!fu_plugin_runner_device_generic_progress(
 		self,
 		device,
@@ -2024,7 +2169,6 @@ fu_plugin_runner_unlock(FuPlugin *self, FuDevice *device, GError **error)
 	}
 
 	/* run vfunc */
-	fu_device_add_backend_tag(device, "unlock");
 	if (!fu_plugin_runner_device_generic(self,
 					     device,
 					     "fu_plugin_unlock",
@@ -2074,7 +2218,6 @@ fu_plugin_runner_write_firmware(FuPlugin *self,
 		g_debug("plugin not enabled, skipping");
 		return TRUE;
 	}
-	fu_device_add_backend_tag(device, "write-firmware");
 
 	/* optional */
 	if (vfuncs->write_firmware == NULL) {
@@ -2211,6 +2354,66 @@ fu_plugin_runner_get_results(FuPlugin *self, FuDevice *device, GError **error)
 	return TRUE;
 }
 
+/**
+ * fu_plugin_runner_fix_host_security_attr:
+ * @self: a #FuPlugin
+ * @attr: (nullable): a security attribute
+ * @error: (nullable): optional return location for an error
+ *
+ * Fix the specific security attribute.
+ *
+ * Returns: #TRUE for success, #FALSE for failure
+ *
+ * Since: 1.9.6
+ **/
+gboolean
+fu_plugin_runner_fix_host_security_attr(FuPlugin *self, FwupdSecurityAttr *attr, GError **error)
+{
+	FuPluginVfuncs *vfuncs = fu_plugin_get_vfuncs(self);
+
+	g_return_val_if_fail(FU_IS_PLUGIN(self), FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
+	if (vfuncs->fix_host_security_attr == NULL) {
+		g_set_error_literal(error,
+				    G_IO_ERROR,
+				    G_IO_ERROR_NOT_SUPPORTED,
+				    "fix is not supported");
+		return FALSE;
+	}
+	return vfuncs->fix_host_security_attr(self, attr, error);
+}
+
+/**
+ * fu_plugin_runner_undo_host_security_attr:
+ * @self: a #FuPlugin
+ * @attr: (nullable): a security attribute
+ * @error: (nullable): optional return location for an error
+ *
+ * Fix the security issue of given security attribute.
+ *
+ * Returns: #TRUE for success, #FALSE for failure
+ *
+ * Since: 1.9.6
+ **/
+gboolean
+fu_plugin_runner_undo_host_security_attr(FuPlugin *self, FwupdSecurityAttr *attr, GError **error)
+{
+	FuPluginVfuncs *vfuncs = fu_plugin_get_vfuncs(self);
+
+	g_return_val_if_fail(FU_IS_PLUGIN(self), FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
+	if (vfuncs->undo_host_security_attr == NULL) {
+		g_set_error_literal(error,
+				    G_IO_ERROR,
+				    G_IO_ERROR_NOT_SUPPORTED,
+				    "undo is not supported");
+		return FALSE;
+	}
+	return vfuncs->undo_host_security_attr(self, attr, error);
+}
+
 /**
  * fu_plugin_get_order:
  * @self: a #FuPlugin
@@ -2397,21 +2600,42 @@ fu_plugin_get_report_metadata(FuPlugin *self)
  * @self: a #FuPlugin
  * @key: a settings key
  *
- * Return the value of a key if it's been configured
+ * Return the value of a key, falling back to the default value.
  *
  * Since: 1.0.6
  **/
 gchar *
 fu_plugin_get_config_value(FuPlugin *self, const gchar *key)
 {
-	g_autofree gchar *conf_path = fu_plugin_get_config_filename(self);
-	g_autoptr(GKeyFile) keyfile = NULL;
-	if (!g_file_test(conf_path, G_FILE_TEST_IS_REGULAR))
+	FuPluginPrivate *priv = fu_plugin_get_instance_private(self);
+	FuConfig *config = fu_context_get_config(priv->ctx);
+	if (config == NULL) {
+		g_critical("cannot get config value with no loaded context!");
 		return NULL;
-	keyfile = g_key_file_new();
-	if (!g_key_file_load_from_file(keyfile, conf_path, G_KEY_FILE_NONE, NULL))
-		return NULL;
-	return g_key_file_get_string(keyfile, fu_plugin_get_name(self), key, NULL);
+	}
+	return fu_config_get_value(config, fu_plugin_get_name(self), key);
+}
+
+/**
+ * fu_plugin_set_config_default:
+ * @self: a #FuPlugin
+ * @key: a settings key
+ * @value: the default value of the key if not found
+ *
+ * Sets the config default value.
+ *
+ * Since: 2.0.0
+ **/
+void
+fu_plugin_set_config_default(FuPlugin *self, const gchar *key, const gchar *value)
+{
+	FuPluginPrivate *priv = fu_plugin_get_instance_private(self);
+	FuConfig *config = fu_context_get_config(priv->ctx);
+	if (config == NULL) {
+		g_critical("cannot set config default with no loaded context!");
+		return;
+	}
+	fu_config_set_default(config, fu_plugin_get_name(self), key, value);
 }
 
 /**
@@ -2457,7 +2681,7 @@ g_file_set_contents_full(const gchar *filename,
 	if (length < 0)
 		length = strlen(contents);
 	fd = g_open(filename, O_CREAT, mode);
-	if (fd <= 0) {
+	if (fd < 0) {
 		g_set_error(error,
 			    G_IO_ERROR,
 			    G_IO_ERROR_FAILED,
@@ -2479,38 +2703,6 @@ g_file_set_contents_full(const gchar *filename,
 }
 #endif
 
-static gboolean
-fu_plugin_set_config_value_internal(FuPlugin *self,
-				    const gchar *key,
-				    const gchar *value,
-				    guint32 mode,
-				    GError **error)
-{
-	g_autofree gchar *conf_path = fu_plugin_get_config_filename(self);
-	g_autofree gchar *data = NULL;
-	g_autoptr(GKeyFile) keyfile = g_key_file_new();
-
-	g_return_val_if_fail(FU_IS_PLUGIN(self), FALSE);
-	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
-
-	if (!g_file_test(conf_path, G_FILE_TEST_EXISTS)) {
-		g_set_error(error, FWUPD_ERROR, FWUPD_ERROR_NOT_FOUND, "%s is missing", conf_path);
-		return FALSE;
-	}
-	if (!g_key_file_load_from_file(keyfile, conf_path, G_KEY_FILE_KEEP_COMMENTS, error))
-		return FALSE;
-	g_key_file_set_string(keyfile, fu_plugin_get_name(self), key, value);
-	data = g_key_file_to_data(keyfile, NULL, error);
-	if (data == NULL)
-		return FALSE;
-	return g_file_set_contents_full(conf_path,
-					data,
-					-1,
-					G_FILE_SET_CONTENTS_CONSISTENT,
-					mode,
-					error);
-}
-
 /**
  * fu_plugin_set_config_value:
  * @self: a #FuPlugin
@@ -2527,23 +2719,47 @@ fu_plugin_set_config_value_internal(FuPlugin *self,
 gboolean
 fu_plugin_set_config_value(FuPlugin *self, const gchar *key, const gchar *value, GError **error)
 {
+	FuPluginPrivate *priv = fu_plugin_get_instance_private(self);
+	FuConfig *config = fu_context_get_config(priv->ctx);
 	g_return_val_if_fail(FU_IS_PLUGIN(self), FALSE);
 	g_return_val_if_fail(key != NULL, FALSE);
 	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
-
-	/* when removing fu_plugin_set_secure_config_value() just remove this instead */
-	if (fu_plugin_has_flag(self, FWUPD_PLUGIN_FLAG_SECURE_CONFIG)) {
-		return fu_plugin_set_config_value_internal(self,
-							   key,
-							   value,
-							   FU_PLUGIN_FILE_MODE_SECURE,
-							   error);
+	if (config == NULL) {
+		g_set_error_literal(error,
+				    G_IO_ERROR,
+				    G_IO_ERROR_FAILED,
+				    "cannot get config value with no loaded context");
+		return FALSE;
 	}
-	return fu_plugin_set_config_value_internal(self,
-						   key,
-						   value,
-						   FU_PLUGIN_FILE_MODE_NONSECURE,
-						   error);
+	return fu_config_set_value(config, fu_plugin_get_name(self), key, value, error);
+}
+
+/**
+ * fu_plugin_reset_config_values:
+ * @self: a #FuPlugin
+ * @error: (nullable): optional return location for an error
+ *
+ * Reset all the plugin keys back to the default values.
+ *
+ * Returns: %TRUE for success
+ *
+ * Since: 1.9.15
+ **/
+gboolean
+fu_plugin_reset_config_values(FuPlugin *self, GError **error)
+{
+	FuPluginPrivate *priv = fu_plugin_get_instance_private(self);
+	FuConfig *config = fu_context_get_config(priv->ctx);
+	g_return_val_if_fail(FU_IS_PLUGIN(self), FALSE);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+	if (config == NULL) {
+		g_set_error_literal(error,
+				    G_IO_ERROR,
+				    G_IO_ERROR_FAILED,
+				    "cannot reset config values with no loaded context");
+		return FALSE;
+	}
+	return fu_config_reset_defaults(config, fu_plugin_get_name(self), error);
 }
 
 /**
@@ -2560,10 +2776,13 @@ fu_plugin_set_config_value(FuPlugin *self, const gchar *key, const gchar *value,
 gboolean
 fu_plugin_get_config_value_boolean(FuPlugin *self, const gchar *key)
 {
-	g_autofree gchar *tmp = fu_plugin_get_config_value(self, key);
-	if (tmp == NULL)
+	FuPluginPrivate *priv = fu_plugin_get_instance_private(self);
+	FuConfig *config = fu_context_get_config(priv->ctx);
+	if (config == NULL) {
+		g_critical("cannot get config value with no loaded context!");
 		return FALSE;
-	return g_ascii_strcasecmp(tmp, "true") == 0;
+	}
+	return fu_config_get_value_bool(config, fu_plugin_get_name(self), key);
 }
 
 /**
@@ -2766,25 +2985,6 @@ fu_plugin_class_init(FuPluginClass *klass)
 						     g_cclosure_marshal_VOID__VOID,
 						     G_TYPE_NONE,
 						     0);
-	/**
-	 * FuPlugin::config-changed:
-	 * @self: the #FuPlugin instance that emitted the signal
-	 *
-	 * The ::config-changed signal is emitted when one or more config files have changed which
-	 * may affect how the daemon should be run.
-	 *
-	 * Since: 1.7.0
-	 **/
-	signals[SIGNAL_CONFIG_CHANGED] =
-	    g_signal_new("config-changed",
-			 G_TYPE_FROM_CLASS(object_class),
-			 G_SIGNAL_RUN_LAST,
-			 G_STRUCT_OFFSET(FuPluginClass, _config_changed),
-			 NULL,
-			 NULL,
-			 g_cclosure_marshal_VOID__VOID,
-			 G_TYPE_NONE,
-			 0);
 
 	/**
 	 * FuPlugin:context:
@@ -2805,7 +3005,7 @@ static void
 fu_plugin_init(FuPlugin *self)
 {
 	FuPluginPrivate *priv = GET_PRIVATE(self);
-	g_rw_lock_init(&priv->cache_mutex);
+	priv->device_gtype_default = G_TYPE_INVALID;
 }
 
 static void
@@ -2815,8 +3015,6 @@ fu_plugin_finalize(GObject *object)
 	FuPluginPrivate *priv = GET_PRIVATE(self);
 	FuPluginVfuncs *vfuncs = fu_plugin_get_vfuncs(self);
 
-	g_rw_lock_clear(&priv->cache_mutex);
-
 	/* optional */
 	if (priv->done_init && vfuncs->finalize != NULL) {
 		g_debug("finalize(%s)", fu_plugin_get_name(self));
