@@ -12,9 +12,11 @@
 #include "fu-bytes.h"
 #include "fu-common.h"
 #include "fu-coswid-firmware.h"
+#include "fu-lzma-common.h"
 #include "fu-mem.h"
 #include "fu-string.h"
 #include "fu-uswid-firmware.h"
+#include "fu-uswid-struct.h"
 
 /**
  * FuUswidFirmware:
@@ -26,42 +28,13 @@
 
 typedef struct {
 	guint8 hdrver;
-	gboolean compressed;
+	FuUswidPayloadCompression compression;
 } FuUswidFirmwarePrivate;
 
 G_DEFINE_TYPE_WITH_PRIVATE(FuUswidFirmware, fu_uswid_firmware, FU_TYPE_FIRMWARE)
 #define GET_PRIVATE(o) (fu_uswid_firmware_get_instance_private(o))
 
-#define USWID_HEADER_VERSION_V1 1
-#define USWID_HEADER_SIZE_V1	23
-#define USWID_HEADER_SIZE_V2	24
-
-#define USWID_HEADER_FLAG_COMPRESSED 0x01
-
-typedef struct __attribute__((packed)) {
-	guint8 magic[16];
-	guint8 hdrver;
-	guint16 hdrsz;
-	guint32 payloadsz;
-	guint8 flags;
-} FuUswidFirmwareHdr;
-
-const guint8 USWID_HEADER_MAGIC[] = {0x53,
-				     0x42,
-				     0x4F,
-				     0x4D,
-				     0xD6,
-				     0xBA,
-				     0x2E,
-				     0xAC,
-				     0xA3,
-				     0xE6,
-				     0x7A,
-				     0x52,
-				     0xAA,
-				     0xEE,
-				     0x3B,
-				     0xAF};
+#define FU_USWID_FIRMARE_MINIMUM_HDRVER 1
 
 static void
 fu_uswid_firmware_export(FuFirmware *firmware, FuFirmwareExportFlags flags, XbBuilderNode *bn)
@@ -69,35 +42,18 @@ fu_uswid_firmware_export(FuFirmware *firmware, FuFirmwareExportFlags flags, XbBu
 	FuUswidFirmware *self = FU_USWID_FIRMWARE(firmware);
 	FuUswidFirmwarePrivate *priv = GET_PRIVATE(self);
 	fu_xmlb_builder_insert_kx(bn, "hdrver", priv->hdrver);
-	fu_xmlb_builder_insert_kb(bn, "compressed", priv->compressed);
+	if (priv->compression != FU_USWID_PAYLOAD_COMPRESSION_NONE) {
+		fu_xmlb_builder_insert_kv(
+		    bn,
+		    "compression",
+		    fu_uswid_payload_compression_to_string(priv->compression));
+	}
 }
 
 static gboolean
 fu_uswid_firmware_check_magic(FuFirmware *firmware, GBytes *fw, gsize offset, GError **error)
 {
-	guint8 magic[16] = {0x0};
-
-	if (!fu_memcpy_safe(magic,
-			    sizeof(magic),
-			    0, /* dst */
-			    g_bytes_get_data(fw, NULL),
-			    g_bytes_get_size(fw),
-			    offset,
-			    sizeof(magic),
-			    error)) {
-		g_prefix_error(error, "failed to read magic: ");
-		return FALSE;
-	}
-	if (memcmp(magic, USWID_HEADER_MAGIC, sizeof(magic)) != 0) {
-		g_set_error_literal(error,
-				    FWUPD_ERROR,
-				    FWUPD_ERROR_INVALID_FILE,
-				    "invalid magic for file");
-		return FALSE;
-	}
-
-	/* success */
-	return TRUE;
+	return fu_struct_uswid_validate_bytes(fw, offset, error);
 }
 
 static gboolean
@@ -109,21 +65,19 @@ fu_uswid_firmware_parse(FuFirmware *firmware,
 {
 	FuUswidFirmware *self = FU_USWID_FIRMWARE(firmware);
 	FuUswidFirmwarePrivate *priv = GET_PRIVATE(self);
-	guint16 hdrsz = 0;
-	guint32 payloadsz = 0;
-	guint8 uswid_flags = 0;
-	gsize bufsz;
-	const guint8 *buf = g_bytes_get_data(fw, &bufsz);
+	guint16 hdrsz;
+	guint32 payloadsz;
+	g_autoptr(GByteArray) st = NULL;
 	g_autoptr(GBytes) payload = NULL;
 
-	/* hdrver */
-	if (!fu_memread_uint8_safe(buf,
-				   bufsz,
-				   offset + G_STRUCT_OFFSET(FuUswidFirmwareHdr, hdrver),
-				   &priv->hdrver,
-				   error))
+	/* unpack */
+	st = fu_struct_uswid_parse_bytes(fw, offset, error);
+	if (st == NULL)
 		return FALSE;
-	if (priv->hdrver < USWID_HEADER_VERSION_V1) {
+
+	/* hdrver */
+	priv->hdrver = fu_struct_uswid_get_hdrver(st);
+	if (priv->hdrver < FU_USWID_FIRMARE_MINIMUM_HDRVER) {
 		g_set_error_literal(error,
 				    FWUPD_ERROR,
 				    FWUPD_ERROR_NOT_SUPPORTED,
@@ -131,30 +85,9 @@ fu_uswid_firmware_parse(FuFirmware *firmware,
 		return FALSE;
 	}
 
-	/* hdrsz */
-	if (!fu_memread_uint16_safe(buf,
-				    bufsz,
-				    offset + G_STRUCT_OFFSET(FuUswidFirmwareHdr, hdrsz),
-				    &hdrsz,
-				    G_LITTLE_ENDIAN,
-				    error))
-		return FALSE;
-	if (hdrsz < USWID_HEADER_SIZE_V1) {
-		g_set_error_literal(error,
-				    FWUPD_ERROR,
-				    FWUPD_ERROR_NOT_SUPPORTED,
-				    "header size is invalid");
-		return FALSE;
-	}
-
-	/* payloadsz */
-	if (!fu_memread_uint32_safe(buf,
-				    bufsz,
-				    offset + G_STRUCT_OFFSET(FuUswidFirmwareHdr, payloadsz),
-				    &payloadsz,
-				    G_LITTLE_ENDIAN,
-				    error))
-		return FALSE;
+	/* hdrsz+payloadsz */
+	hdrsz = fu_struct_uswid_get_hdrsz(st);
+	payloadsz = fu_struct_uswid_get_payloadsz(st);
 	if (payloadsz == 0x0) {
 		g_set_error_literal(error,
 				    FWUPD_ERROR,
@@ -165,18 +98,22 @@ fu_uswid_firmware_parse(FuFirmware *firmware,
 	fu_firmware_set_size(firmware, hdrsz + payloadsz);
 
 	/* flags */
-	if (priv->hdrver >= 0x02) {
-		if (!fu_memread_uint8_safe(buf,
-					   bufsz,
-					   offset + G_STRUCT_OFFSET(FuUswidFirmwareHdr, flags),
-					   &uswid_flags,
-					   error))
-			return FALSE;
-		priv->compressed = (uswid_flags & USWID_HEADER_FLAG_COMPRESSED) > 0;
+	if (priv->hdrver >= 0x03) {
+		if (fu_struct_uswid_get_flags(st) & FU_USWID_HEADER_FLAG_COMPRESSED) {
+			priv->compression = fu_struct_uswid_get_compression(st);
+		} else {
+			priv->compression = FU_USWID_PAYLOAD_COMPRESSION_NONE;
+		}
+	} else if (priv->hdrver >= 0x02) {
+		priv->compression = fu_struct_uswid_get_flags(st) & FU_USWID_HEADER_FLAG_COMPRESSED
+					? FU_USWID_PAYLOAD_COMPRESSION_ZLIB
+					: FU_USWID_PAYLOAD_COMPRESSION_NONE;
+	} else {
+		priv->compression = FU_USWID_PAYLOAD_COMPRESSION_NONE;
 	}
 
 	/* zlib stream */
-	if (priv->compressed) {
+	if (priv->compression == FU_USWID_PAYLOAD_COMPRESSION_ZLIB) {
 		g_autoptr(GBytes) payload_tmp = NULL;
 		g_autoptr(GConverter) conv = NULL;
 		g_autoptr(GInputStream) istream1 = NULL;
@@ -192,10 +129,26 @@ fu_uswid_firmware_parse(FuFirmware *firmware,
 		if (payload == NULL)
 			return FALSE;
 		payloadsz = g_bytes_get_size(payload);
-	} else {
+	} else if (priv->compression == FU_USWID_PAYLOAD_COMPRESSION_LZMA) {
+		g_autoptr(GBytes) payload_tmp = NULL;
+		payload_tmp = fu_bytes_new_offset(fw, offset + hdrsz, payloadsz, error);
+		if (payload_tmp == NULL)
+			return FALSE;
+		payload = fu_lzma_decompress_bytes(payload_tmp, error);
+		if (payload == NULL)
+			return FALSE;
+		payloadsz = g_bytes_get_size(payload);
+	} else if (priv->compression == FU_USWID_PAYLOAD_COMPRESSION_NONE) {
 		payload = fu_bytes_new_offset(fw, offset + hdrsz, payloadsz, error);
 		if (payload == NULL)
 			return FALSE;
+	} else {
+		g_set_error(error,
+			    FWUPD_ERROR,
+			    FWUPD_ERROR_NOT_SUPPORTED,
+			    "compression format 0x%x is not supported",
+			    priv->compression);
+		return FALSE;
 	}
 
 	/* payload */
@@ -212,7 +165,8 @@ fu_uswid_firmware_parse(FuFirmware *firmware,
 				       flags | FWUPD_INSTALL_FLAG_NO_SEARCH,
 				       error))
 			return FALSE;
-		fu_firmware_add_image(firmware, firmware_coswid);
+		if (!fu_firmware_add_image_full(firmware, firmware_coswid, error))
+			return FALSE;
 		if (fu_firmware_get_size(firmware_coswid) == 0) {
 			g_set_error_literal(error,
 					    FWUPD_ERROR,
@@ -227,21 +181,12 @@ fu_uswid_firmware_parse(FuFirmware *firmware,
 	return TRUE;
 }
 
-static guint8
-fu_uswid_firmware_calculate_hdrsz(FuUswidFirmware *self)
-{
-	FuUswidFirmwarePrivate *priv = GET_PRIVATE(self);
-	if (priv->hdrver == 0x1)
-		return USWID_HEADER_SIZE_V1;
-	return USWID_HEADER_SIZE_V2;
-}
-
-static GBytes *
+static GByteArray *
 fu_uswid_firmware_write(FuFirmware *firmware, GError **error)
 {
 	FuUswidFirmware *self = FU_USWID_FIRMWARE(firmware);
 	FuUswidFirmwarePrivate *priv = GET_PRIVATE(self);
-	g_autoptr(GByteArray) buf = g_byte_array_new();
+	g_autoptr(GByteArray) buf = fu_struct_uswid_new();
 	g_autoptr(GByteArray) payload = g_byte_array_new();
 	g_autoptr(GBytes) payload_blob = NULL;
 	g_autoptr(GPtrArray) images = fu_firmware_get_images(firmware);
@@ -256,7 +201,7 @@ fu_uswid_firmware_write(FuFirmware *firmware, GError **error)
 	}
 
 	/* zlibify */
-	if (priv->compressed) {
+	if (priv->compression == FU_USWID_PAYLOAD_COMPRESSION_ZLIB) {
 		g_autoptr(GConverter) conv = NULL;
 		g_autoptr(GInputStream) istream1 = NULL;
 		g_autoptr(GInputStream) istream2 = NULL;
@@ -267,25 +212,47 @@ fu_uswid_firmware_write(FuFirmware *firmware, GError **error)
 		payload_blob = fu_bytes_get_contents_stream(istream2, G_MAXSIZE, error);
 		if (payload_blob == NULL)
 			return NULL;
+	} else if (priv->compression == FU_USWID_PAYLOAD_COMPRESSION_LZMA) {
+		g_autoptr(GBytes) payload_tmp = g_bytes_new(payload->data, payload->len);
+		payload_blob = fu_lzma_compress_bytes(payload_tmp, error);
+		if (payload_blob == NULL)
+			return NULL;
 	} else {
-		payload_blob = g_byte_array_free_to_bytes(g_steal_pointer(&payload));
+		payload_blob = g_bytes_new(payload->data, payload->len);
 	}
 
-	/* header then CBOR blob */
-	g_byte_array_append(buf, USWID_HEADER_MAGIC, sizeof(USWID_HEADER_MAGIC));
-	fu_byte_array_append_uint8(buf, priv->hdrver);
-	fu_byte_array_append_uint16(buf, fu_uswid_firmware_calculate_hdrsz(self), G_LITTLE_ENDIAN);
-	fu_byte_array_append_uint32(buf, g_bytes_get_size(payload_blob), G_LITTLE_ENDIAN);
-	if (priv->hdrver >= 2) {
+	/* pack */
+	fu_struct_uswid_set_hdrver(buf, priv->hdrver);
+	fu_struct_uswid_set_payloadsz(buf, g_bytes_get_size(payload_blob));
+	if (priv->hdrver >= 3) {
 		guint8 flags = 0;
-		if (priv->compressed)
-			flags |= USWID_HEADER_FLAG_COMPRESSED;
-		fu_byte_array_append_uint8(buf, flags);
+		if (priv->compression != FU_USWID_PAYLOAD_COMPRESSION_NONE)
+			flags |= FU_USWID_HEADER_FLAG_COMPRESSED;
+		fu_struct_uswid_set_flags(buf, flags);
+		fu_struct_uswid_set_compression(buf, priv->compression);
+	} else if (priv->hdrver >= 2) {
+		guint8 flags = 0;
+		if (priv->compression != FU_USWID_PAYLOAD_COMPRESSION_NONE) {
+			if (priv->compression != FU_USWID_PAYLOAD_COMPRESSION_ZLIB) {
+				g_set_error_literal(error,
+						    FWUPD_ERROR,
+						    FWUPD_ERROR_NOT_SUPPORTED,
+						    "hdrver 0x02 only supports zlib compression");
+				return NULL;
+			}
+			flags |= FU_USWID_HEADER_FLAG_COMPRESSED;
+		}
+		fu_struct_uswid_set_flags(buf, flags);
+		g_byte_array_set_size(buf, buf->len - 1);
+		fu_struct_uswid_set_hdrsz(buf, buf->len);
+	} else {
+		g_byte_array_set_size(buf, buf->len - 2);
+		fu_struct_uswid_set_hdrsz(buf, buf->len);
 	}
 	fu_byte_array_append_bytes(buf, payload_blob);
 
 	/* success */
-	return g_byte_array_free_to_bytes(g_steal_pointer(&buf));
+	return g_steal_pointer(&buf);
 }
 
 static gboolean
@@ -302,10 +269,19 @@ fu_uswid_firmware_build(FuFirmware *firmware, XbNode *n, GError **error)
 		priv->hdrver = tmp;
 
 	/* simple properties */
-	str = xb_node_query_text(n, "compressed", NULL);
+	str = xb_node_query_text(n, "compression", NULL);
 	if (str != NULL) {
-		if (!fu_strtobool(str, &priv->compressed, error))
+		priv->compression = fu_uswid_payload_compression_from_string(str);
+		if (priv->compression == FU_USWID_PAYLOAD_COMPRESSION_NONE) {
+			g_set_error(error,
+				    G_IO_ERROR,
+				    G_IO_ERROR_INVALID_DATA,
+				    "invalid compression type %s",
+				    str);
 			return FALSE;
+		}
+	} else {
+		priv->compression = FU_USWID_PAYLOAD_COMPRESSION_NONE;
 	}
 
 	/* success */
@@ -316,9 +292,11 @@ static void
 fu_uswid_firmware_init(FuUswidFirmware *self)
 {
 	FuUswidFirmwarePrivate *priv = GET_PRIVATE(self);
-	priv->hdrver = USWID_HEADER_VERSION_V1;
-	priv->compressed = FALSE;
+	priv->hdrver = FU_USWID_FIRMARE_MINIMUM_HDRVER;
+	priv->compression = FU_USWID_PAYLOAD_COMPRESSION_NONE;
 	fu_firmware_add_flag(FU_FIRMWARE(self), FU_FIRMWARE_FLAG_HAS_STORED_SIZE);
+	fu_firmware_add_flag(FU_FIRMWARE(self), FU_FIRMWARE_FLAG_ALWAYS_SEARCH);
+	fu_firmware_set_images_max(FU_FIRMWARE(self), 2000);
 }
 
 static void
