@@ -6,8 +6,6 @@
 
 #include "config.h"
 
-#include <fwupdplugin.h>
-
 #include <string.h>
 
 #include "fu-cros-ec-common.h"
@@ -81,15 +79,7 @@ fu_cros_ec_usb_device_get_configuration(FuCrosEcUsbDevice *self, GError **error)
 	guint8 index;
 	g_autofree gchar *configuration = NULL;
 
-#if G_USB_CHECK_VERSION(0, 3, 5)
 	index = g_usb_device_get_configuration_index(usb_device);
-#else
-	g_set_error_literal(error,
-			    FWUPD_ERROR,
-			    FWUPD_ERROR_NOT_SUPPORTED,
-			    "this version of GUsb is not supported");
-	return FALSE;
-#endif
 	configuration = g_usb_device_get_string_descriptor(usb_device, index, error);
 	if (configuration == NULL)
 		return FALSE;
@@ -109,7 +99,6 @@ fu_cros_ec_usb_device_get_configuration(FuCrosEcUsbDevice *self, GError **error)
 static gboolean
 fu_cros_ec_usb_device_find_interface(FuUsbDevice *device, GError **error)
 {
-#if G_USB_CHECK_VERSION(0, 3, 3)
 	GUsbDevice *usb_device = fu_usb_device_get_dev(device);
 	FuCrosEcUsbDevice *self = FU_CROS_EC_USB_DEVICE(device);
 	g_autoptr(GPtrArray) intfs = NULL;
@@ -125,9 +114,7 @@ fu_cros_ec_usb_device_find_interface(FuUsbDevice *device, GError **error)
 		    g_usb_interface_get_subclass(intf) == USB_SUBCLASS_GOOGLE_UPDATE &&
 		    g_usb_interface_get_protocol(intf) == USB_PROTOCOL_GOOGLE_UPDATE) {
 			GUsbEndpoint *ep;
-			g_autoptr(GPtrArray) endpoints = NULL;
-
-			endpoints = g_usb_interface_get_endpoints(intf);
+			g_autoptr(GPtrArray) endpoints = g_usb_interface_get_endpoints(intf);
 			if (NULL == endpoints || 0 == endpoints->len)
 				continue;
 			ep = g_ptr_array_index(endpoints, 0);
@@ -140,13 +127,6 @@ fu_cros_ec_usb_device_find_interface(FuUsbDevice *device, GError **error)
 	}
 	g_set_error_literal(error, FWUPD_ERROR, FWUPD_ERROR_NOT_FOUND, "no update interface found");
 	return FALSE;
-#else
-	g_set_error_literal(error,
-			    FWUPD_ERROR,
-			    FWUPD_ERROR_NOT_SUPPORTED,
-			    "this version of GUsb is not supported");
-	return FALSE;
-#endif
 }
 
 static gboolean
@@ -348,10 +328,9 @@ fu_cros_ec_usb_device_start_request(FuDevice *device, gpointer user_data, GError
 {
 	FuCrosEcUsbDevice *self = FU_CROS_EC_USB_DEVICE(device);
 	guint8 *start_resp = (guint8 *)user_data;
-	struct update_frame_header ufh;
+	struct update_frame_header ufh = {0x0};
 	gsize rxed_size = 0;
 
-	memset(&ufh, 0, sizeof(ufh));
 	ufh.block_size = GUINT32_TO_BE(sizeof(ufh));
 	if (!fu_cros_ec_usb_device_do_xfer(self,
 					   (const guint8 *)&ufh,
@@ -382,7 +361,7 @@ fu_cros_ec_usb_device_setup(FuDevice *device, GError **error)
 {
 	FuCrosEcUsbDevice *self = FU_CROS_EC_USB_DEVICE(device);
 	guint32 error_code;
-	START_RESP start_resp;
+	START_RESP start_resp = {0x0};
 	g_auto(GStrv) config_split = NULL;
 
 	/* FuUsbDevice->setup */
@@ -481,14 +460,31 @@ fu_cros_ec_usb_device_setup(FuDevice *device, GError **error)
 		fu_device_set_version(FU_DEVICE(device), self->active_version.triplet);
 		fu_device_set_version_bootloader(FU_DEVICE(device), self->version.triplet);
 	}
-	fu_device_add_instance_id(FU_DEVICE(device), self->version.boardname);
+
+	/* one extra instance ID */
+	fu_device_add_instance_str(FU_DEVICE(device), "BOARDNAME", self->version.boardname);
+	if (!fu_device_build_instance_id(FU_DEVICE(device),
+					 error,
+					 "USB",
+					 "VID",
+					 "PID",
+					 "BOARDNAME",
+					 NULL))
+		return FALSE;
 
 	/* success */
 	return TRUE;
 }
 
 static gboolean
-fu_cros_ec_usb_device_transfer_block(FuDevice *device, gpointer user_data, GError **error)
+fu_cros_ec_usb_device_reload(FuDevice *device, GError **error)
+{
+	fu_device_remove_flag(device, FWUPD_DEVICE_FLAG_ANOTHER_WRITE_REQUIRED);
+	return fu_cros_ec_usb_device_setup(device, error);
+}
+
+static gboolean
+fu_cros_ec_usb_device_transfer_block_cb(FuDevice *device, gpointer user_data, GError **error)
 {
 	FuCrosEcUsbDevice *self = FU_CROS_EC_USB_DEVICE(device);
 	FuCrosEcUsbBlockHelper *helper = (FuCrosEcUsbBlockHelper *)user_data;
@@ -520,6 +516,9 @@ fu_cros_ec_usb_device_transfer_block(FuDevice *device, gpointer user_data, GErro
 		return FALSE;
 	}
 
+	/* we're in a retry handler */
+	fu_progress_reset(helper->progress);
+
 	/* send the block, chunk by chunk */
 	chunks = fu_chunk_array_new(fu_chunk_get_data(helper->block),
 				    fu_chunk_get_data_sz(helper->block),
@@ -618,7 +617,7 @@ fu_cros_ec_usb_device_transfer_section(FuDevice *device,
 	}
 
 	/* smart update: trim trailing bytes */
-	while (data_len != 0 && (data_ptr[data_len - 1] == 0xff))
+	while (data_len > 1 && (data_ptr[data_len - 1] == 0xff))
 		data_len--;
 	g_debug("trimmed %" G_GSIZE_FORMAT " trailing bytes", section->size - data_len);
 	g_debug("sending 0x%x bytes to 0x%x", (guint)data_len, section->offset);
@@ -637,7 +636,7 @@ fu_cros_ec_usb_device_transfer_section(FuDevice *device,
 		    .progress = fu_progress_get_child(progress),
 		};
 		if (!fu_device_retry(device,
-				     fu_cros_ec_usb_device_transfer_block,
+				     fu_cros_ec_usb_device_transfer_block_cb,
 				     MAX_BLOCK_XFER_RETRIES,
 				     &helper,
 				     error)) {
@@ -703,9 +702,9 @@ fu_cros_ec_usb_device_send_subcommand(FuDevice *device,
 static gboolean
 fu_cros_ec_usb_device_reset_to_ro(FuDevice *device, GError **error)
 {
-	guint8 response;
+	guint8 response = 0x0;
 	guint16 subcommand = UPDATE_EXTRA_CMD_IMMEDIATE_RESET;
-	guint8 command_body[2]; /* Max command body size. */
+	guint8 command_body[2] = {0x0}; /* max command body size */
 	gsize command_body_size = 0;
 	gsize response_size = 1;
 	g_autoptr(GError) error_local = NULL;
@@ -730,9 +729,9 @@ fu_cros_ec_usb_device_reset_to_ro(FuDevice *device, GError **error)
 static gboolean
 fu_cros_ec_usb_device_jump_to_rw(FuDevice *device)
 {
-	guint8 response;
+	guint8 response = 0x0;
 	guint16 subcommand = UPDATE_EXTRA_CMD_JUMP_TO_RW;
-	guint8 command_body[2]; /* Max command body size. */
+	guint8 command_body[2] = {0x0}; /* max command body size */
 	gsize command_body_size = 0;
 	gsize response_size = 1;
 
@@ -768,9 +767,9 @@ static gboolean
 fu_cros_ec_usb_device_stay_in_ro(FuDevice *device, GError **error)
 {
 	gsize response_size = 1;
-	guint8 response;
+	guint8 response = 0x0;
 	guint16 subcommand = UPDATE_EXTRA_CMD_STAY_IN_RO;
-	guint8 command_body[2]; /* Max command body size. */
+	guint8 command_body[2] = {0x0}; /* max command body size */
 	gsize command_body_size = 0;
 
 	if (!fu_cros_ec_usb_device_send_subcommand(device,
@@ -801,7 +800,7 @@ fu_cros_ec_usb_device_write_firmware(FuDevice *device,
 	fu_device_remove_private_flag(device, FU_CROS_EC_USB_DEVICE_FLAG_SPECIAL);
 
 	if (fu_device_has_private_flag(device, FU_CROS_EC_USB_DEVICE_FLAG_REBOOTING_TO_RO)) {
-		START_RESP start_resp;
+		START_RESP start_resp = {0x0};
 
 		fu_device_remove_private_flag(device, FU_CROS_EC_USB_DEVICE_FLAG_REBOOTING_TO_RO);
 		if (!fu_cros_ec_usb_device_stay_in_ro(device, error)) {
@@ -932,7 +931,6 @@ fu_cros_ec_usb_device_attach(FuDevice *device, FuProgress *progress, GError **er
 {
 	FuCrosEcUsbDevice *self = FU_CROS_EC_USB_DEVICE(device);
 
-	fu_device_set_remove_delay(device, CROS_EC_REMOVE_DELAY_RE_ENUMERATE);
 	if (self->in_bootloader &&
 	    fu_device_has_private_flag(device, FU_CROS_EC_USB_DEVICE_FLAG_SPECIAL)) {
 		/*
@@ -976,10 +974,11 @@ fu_cros_ec_usb_device_detach(FuDevice *device, FuProgress *progress, GError **er
 		g_debug("skipping immediate reboot in case of already in bootloader");
 		/* in RO so skip reboot */
 		return TRUE;
-	} else if (self->targ.common.flash_protection != 0x0) {
+	}
+
+	if (self->targ.common.flash_protection != 0x0) {
 		/* in RW, and RO region is write protected, so jump to RO */
 		fu_device_add_private_flag(device, FU_CROS_EC_USB_DEVICE_FLAG_RO_WRITTEN);
-		fu_device_set_remove_delay(device, CROS_EC_REMOVE_DELAY_RE_ENUMERATE);
 		if (!fu_cros_ec_usb_device_reset_to_ro(device, error))
 			return FALSE;
 		fu_device_add_flag(device, FWUPD_DEVICE_FLAG_WAIT_FOR_REPLUG);
@@ -995,7 +994,9 @@ fu_cros_ec_usb_device_init(FuCrosEcUsbDevice *self)
 	fu_device_add_protocol(FU_DEVICE(self), "com.google.usb.crosec");
 	fu_device_add_flag(FU_DEVICE(self), FWUPD_DEVICE_FLAG_UPDATABLE);
 	fu_device_add_internal_flag(FU_DEVICE(self), FU_DEVICE_INTERNAL_FLAG_REPLUG_MATCH_GUID);
+	fu_device_set_acquiesce_delay(FU_DEVICE(self), 7500); /* ms */
 	fu_device_set_version_format(FU_DEVICE(self), FWUPD_VERSION_FORMAT_TRIPLET);
+	fu_device_set_remove_delay(FU_DEVICE(self), CROS_EC_REMOVE_DELAY_RE_ENUMERATE);
 	fu_device_add_flag(FU_DEVICE(self), FWUPD_DEVICE_FLAG_DUAL_IMAGE);
 	fu_device_add_flag(FU_DEVICE(self), FWUPD_DEVICE_FLAG_SIGNED_PAYLOAD);
 	fu_device_register_private_flag(FU_DEVICE(self),
@@ -1035,11 +1036,10 @@ static void
 fu_cros_ec_usb_device_set_progress(FuDevice *self, FuProgress *progress)
 {
 	fu_progress_set_id(progress, G_STRLOC);
-	fu_progress_add_flag(progress, FU_PROGRESS_FLAG_GUESSED);
-	fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_RESTART, 2, "detach");
-	fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 94, "write");
-	fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_RESTART, 2, "attach");
-	fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_BUSY, 2, "reload");
+	fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_RESTART, 1, "detach");
+	fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_WRITE, 76, "write");
+	fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_RESTART, 1, "attach");
+	fu_progress_add_step(progress, FWUPD_STATUS_DEVICE_BUSY, 22, "reload");
 }
 
 static void
@@ -1054,4 +1054,5 @@ fu_cros_ec_usb_device_class_init(FuCrosEcUsbDeviceClass *klass)
 	klass_device->write_firmware = fu_cros_ec_usb_device_write_firmware;
 	klass_device->probe = fu_cros_ec_usb_device_probe;
 	klass_device->set_progress = fu_cros_ec_usb_device_set_progress;
+	klass_device->reload = fu_cros_ec_usb_device_reload;
 }
