@@ -6,22 +6,16 @@
 
 #include "config.h"
 
-#include <fwupdplugin.h>
-
 #include <errno.h>
 #include <fcntl.h>
 #include <glib/gstdio.h>
 #include <string.h>
 
+#include "fu-firehose-updater.h"
 #include "fu-mbim-qdu-updater.h"
 #include "fu-mm-device.h"
 #include "fu-mm-utils.h"
 #include "fu-qmi-pdc-updater.h"
-
-#if MM_CHECK_VERSION(1, 17, 2)
-#include "fu-firehose-updater.h"
-#endif
-
 #include "fu-sahara-loader.h"
 
 /* Amount of time for the modem to boot in fastboot mode. */
@@ -32,20 +26,13 @@
  * modem boots without SIM card inserted (and therefore the initialization
  * may be very slow) and also where carrier config switching is explicitly
  * required (e.g. if switching from the default (DF) to generic (GC).*/
-#define FU_MM_DEVICE_REMOVE_DELAY_REPROBE 120000 /* ms */
+#define FU_MM_DEVICE_REMOVE_DELAY_REPROBE 210000 /* ms */
 
 #define FU_MM_DEVICE_AT_RETRIES 3
 #define FU_MM_DEVICE_AT_DELAY	3000 /* ms */
 
 /* Amount of time for the modem to get firmware version */
-#define MAX_WAIT_TIME_SECS 150 /* s */
-
-/**
- * FU_MM_DEVICE_FLAG_DETACH_AT_FASTBOOT_HAS_NO_RESPONSE
- *
- * If no AT response is expected when entering fastboot mode.
- */
-#define FU_MM_DEVICE_FLAG_DETACH_AT_FASTBOOT_HAS_NO_RESPONSE (1 << 0)
+#define MAX_WAIT_TIME_SECS 240 /* s */
 
 struct _FuMmDevice {
 	FuDevice parent_instance;
@@ -81,17 +68,15 @@ struct _FuMmDevice {
 
 	/* mbim-qdu update logic */
 	gchar *port_mbim;
-#if MBIM_CHECK_VERSION(1, 25, 3)
 	FuMbimQduUpdater *mbim_qdu_updater;
-#endif /* MBIM_CHECK_VERSION(1,25,3) */
 
 	/* firehose update handling */
 	gchar *port_qcdm;
 	gchar *port_edl;
+	gchar *firehose_prog_file;
 	FuSaharaLoader *sahara_loader;
-#if MM_CHECK_VERSION(1, 17, 2)
 	FuFirehoseUpdater *firehose_updater;
-#endif
+
 	/* for sahara */
 	FuUsbDevice *usb_device;
 
@@ -134,46 +119,14 @@ fu_mm_device_get_update_methods(FuMmDevice *device)
 	return device->update_methods;
 }
 
-const gchar *
-fu_mm_device_get_detach_fastboot_at(FuMmDevice *device)
-{
-	g_return_val_if_fail(FU_IS_MM_DEVICE(device), NULL);
-	return device->detach_fastboot_at;
-}
-
-gint
-fu_mm_device_get_port_at_ifnum(FuMmDevice *device)
-{
-	g_return_val_if_fail(FU_IS_MM_DEVICE(device), -1);
-	return device->port_at_ifnum;
-}
-
-gint
-fu_mm_device_get_port_qmi_ifnum(FuMmDevice *device)
-{
-	g_return_val_if_fail(FU_IS_MM_DEVICE(device), -1);
-	return device->port_qmi_ifnum;
-}
-
-gint
-fu_mm_device_get_port_mbim_ifnum(FuMmDevice *device)
-{
-	g_return_val_if_fail(FU_IS_MM_DEVICE(device), -1);
-	return device->port_mbim_ifnum;
-}
-
 static gboolean
 validate_firmware_update_method(MMModemFirmwareUpdateMethod methods, GError **error)
 {
 	static const MMModemFirmwareUpdateMethod supported_combinations[] = {
 		MM_MODEM_FIRMWARE_UPDATE_METHOD_FASTBOOT,
 		MM_MODEM_FIRMWARE_UPDATE_METHOD_QMI_PDC | MM_MODEM_FIRMWARE_UPDATE_METHOD_FASTBOOT,
-#if MM_CHECK_VERSION(1, 17, 1)
 		MM_MODEM_FIRMWARE_UPDATE_METHOD_MBIM_QDU,
-#endif /* MM_CHECK_VERSION(1,17,1) */
-#if MM_CHECK_VERSION(1, 17, 2)
 		MM_MODEM_FIRMWARE_UPDATE_METHOD_FIREHOSE,
-#endif
 #if MM_CHECK_VERSION(1, 19, 1)
 		MM_MODEM_FIRMWARE_UPDATE_METHOD_FIREHOSE | MM_MODEM_FIRMWARE_UPDATE_METHOD_SAHARA,
 #endif
@@ -183,7 +136,7 @@ validate_firmware_update_method(MMModemFirmwareUpdateMethod methods, GError **er
 	methods_str = mm_modem_firmware_update_method_build_string_from_mask(methods);
 	for (guint i = 0; i < G_N_ELEMENTS(supported_combinations); i++) {
 		if (supported_combinations[i] == methods) {
-			g_debug("valid firmware update combination: %s", methods_str);
+			g_info("valid firmware update combination: %s", methods_str);
 			return TRUE;
 		}
 	}
@@ -196,6 +149,30 @@ validate_firmware_update_method(MMModemFirmwareUpdateMethod methods, GError **er
 	return FALSE;
 }
 
+static void
+fu_mm_device_add_instance_id(FuDevice *dev, const gchar *device_id)
+{
+	if (g_pattern_match_simple("???\\VID_????", device_id)) {
+		fu_device_add_instance_id_full(dev, device_id, FU_DEVICE_INSTANCE_FLAG_QUIRKS);
+		return;
+	}
+	if (g_pattern_match_simple("???\\VID_????&PID_????", device_id)) {
+		fu_device_add_instance_id_full(dev, device_id, FU_DEVICE_INSTANCE_FLAG_QUIRKS);
+		return;
+	}
+	if (g_pattern_match_simple("???\\VID_????&PID_????&REV_????", device_id)) {
+		if (fu_device_has_internal_flag(dev, FU_DEVICE_INTERNAL_FLAG_ADD_INSTANCE_ID_REV))
+			fu_device_add_instance_id(dev, device_id);
+		return;
+	}
+	if (g_pattern_match_simple("???\\VID_????&PID_????&REV_????&CARRIER_*", device_id)) {
+		if (!fu_device_has_private_flag(dev, FU_MM_DEVICE_FLAG_USE_BRANCH))
+			fu_device_add_instance_id(dev, device_id);
+		return;
+	}
+	g_warning("failed to add instance ID %s", device_id);
+}
+
 static gboolean
 fu_mm_device_probe_default(FuDevice *device, GError **error)
 {
@@ -294,7 +271,6 @@ fu_mm_device_probe_default(FuDevice *device, GError **error)
 		if (fu_device_get_protocols(device)->len == 0)
 			fu_device_add_protocol(device, "com.qualcomm.qmi_pdc");
 	}
-#if MM_CHECK_VERSION(1, 17, 1)
 	if (self->update_methods & MM_MODEM_FIRMWARE_UPDATE_METHOD_MBIM_QDU) {
 		for (guint i = 0; i < n_ports; i++) {
 			if (ports[i].type == MM_MODEM_PORT_TYPE_MBIM) {
@@ -304,11 +280,11 @@ fu_mm_device_probe_default(FuDevice *device, GError **error)
 		}
 		fu_device_add_protocol(device, "com.qualcomm.mbim_qdu");
 	}
-#endif /* MM_CHECK_VERSION(1,17,1) */
-#if MM_CHECK_VERSION(1, 17, 2)
 	if (self->update_methods & MM_MODEM_FIRMWARE_UPDATE_METHOD_FIREHOSE) {
 		for (guint i = 0; i < n_ports; i++) {
-			if (ports[i].type == MM_MODEM_PORT_TYPE_QCDM)
+			if ((ports[i].type == MM_MODEM_PORT_TYPE_QCDM) ||
+			    (ports[i].type == MM_MODEM_PORT_TYPE_IGNORED &&
+			     g_strstr_len(ports[i].name, -1, "qcdm") != NULL))
 				self->port_qcdm = g_strdup_printf("/dev/%s", ports[i].name);
 			else if (ports[i].type == MM_MODEM_PORT_TYPE_MBIM)
 				self->port_mbim = g_strdup_printf("/dev/%s", ports[i].name);
@@ -318,7 +294,6 @@ fu_mm_device_probe_default(FuDevice *device, GError **error)
 		}
 		fu_device_add_protocol(device, "com.qualcomm.firehose");
 	}
-#endif
 
 	mm_modem_port_info_array_free(ports, n_ports);
 
@@ -342,7 +317,6 @@ fu_mm_device_probe_default(FuDevice *device, GError **error)
 		return FALSE;
 	}
 
-#if MM_CHECK_VERSION(1, 17, 1)
 	/* a mbim port is required for mbim-qdu */
 	if ((self->update_methods & MM_MODEM_FIRMWARE_UPDATE_METHOD_MBIM_QDU) &&
 	    (self->port_mbim == NULL)) {
@@ -353,8 +327,6 @@ fu_mm_device_probe_default(FuDevice *device, GError **error)
 		return FALSE;
 	}
 
-#endif /* MM_CHECK_VERSION(1,17,1) */
-#if MM_CHECK_VERSION(1, 17, 2)
 	/* a qcdm or mbim port is required for firehose */
 	if ((self->update_methods & MM_MODEM_FIRMWARE_UPDATE_METHOD_FIREHOSE) &&
 	    (self->port_qcdm == NULL && self->port_mbim == NULL)) {
@@ -364,7 +336,6 @@ fu_mm_device_probe_default(FuDevice *device, GError **error)
 				    "failed to find QCDM port");
 		return FALSE;
 	}
-#endif
 
 	if (self->port_at != NULL) {
 		fu_mm_utils_get_port_info(self->port_at,
@@ -483,9 +454,17 @@ fu_mm_device_probe_default(FuDevice *device, GError **error)
 		fu_device_set_vendor(device, mm_modem_get_manufacturer(modem));
 	if (mm_modem_get_model(modem) != NULL)
 		fu_device_set_name(device, mm_modem_get_model(modem));
+
+	/* only for modems that opt-in */
+	if (fu_device_has_private_flag(device, FU_MM_DEVICE_FLAG_USE_BRANCH))
+		fu_device_set_branch(device, mm_modem_get_carrier_configuration(modem));
+
 	fu_device_set_version(device, version);
+
+	/* filter these */
 	for (guint i = 0; device_ids[i] != NULL; i++)
-		fu_device_add_instance_id(device, device_ids[i]);
+		fu_mm_device_add_instance_id(device, device_ids[i]);
+
 	vendors = fu_device_get_vendor_ids(device);
 	if (vendors == NULL || vendors->len == 0) {
 		g_autofree gchar *path = NULL;
@@ -559,14 +538,11 @@ fu_mm_device_probe(FuDevice *device, GError **error)
 {
 	FuMmDevice *self = FU_MM_DEVICE(device);
 
-	if (self->omodem) {
+	if (self->omodem)
 		return fu_mm_device_probe_default(device, error);
-	} else {
-		return fu_mm_device_probe_udev(device, error);
-	}
+	return fu_mm_device_probe_udev(device, error);
 }
 
-#if MM_CHECK_VERSION(1, 17, 2)
 static gboolean
 fu_mm_device_io_open_qcdm(FuMmDevice *self, GError **error)
 {
@@ -596,8 +572,7 @@ fu_mm_device_qcdm_cmd(FuMmDevice *self, const guint8 *cmd, gsize cmd_len, GError
 
 	/* command */
 	qcdm_req = g_bytes_new(cmd, cmd_len);
-	if (g_getenv("FWUPD_MODEM_MANAGER_VERBOSE") != NULL)
-		fu_dump_bytes(G_LOG_DOMAIN, "writing", qcdm_req);
+	fu_dump_bytes(G_LOG_DOMAIN, "writing", qcdm_req);
 	if (!fu_io_channel_write_bytes(self->io_channel,
 				       qcdm_req,
 				       1500,
@@ -617,8 +592,7 @@ fu_mm_device_qcdm_cmd(FuMmDevice *self, const guint8 *cmd, gsize cmd_len, GError
 		g_prefix_error(error, "failed to read qcdm response: ");
 		return FALSE;
 	}
-	if (g_getenv("FWUPD_MODEM_MANAGER_VERBOSE") != NULL)
-		fu_dump_bytes(G_LOG_DOMAIN, "read", qcdm_res);
+	fu_dump_bytes(G_LOG_DOMAIN, "read", qcdm_res);
 
 	/* command == response */
 	if (g_bytes_compare(qcdm_res, qcdm_req) != 0) {
@@ -631,7 +605,6 @@ fu_mm_device_qcdm_cmd(FuMmDevice *self, const guint8 *cmd, gsize cmd_len, GError
 
 	return TRUE;
 }
-#endif /* MM_CHECK_VERSION(1,17,2) */
 
 typedef struct {
 	const gchar *cmd;
@@ -651,8 +624,7 @@ fu_mm_device_at_cmd_cb(FuDevice *device, gpointer user_data, GError **error)
 
 	/* command */
 	at_req = g_bytes_new(cmd_cr, strlen(cmd_cr));
-	if (g_getenv("FWUPD_MODEM_MANAGER_VERBOSE") != NULL)
-		fu_dump_bytes(G_LOG_DOMAIN, "writing", at_req);
+	fu_dump_bytes(G_LOG_DOMAIN, "writing", at_req);
 	if (!fu_io_channel_write_bytes(self->io_channel,
 				       at_req,
 				       1500,
@@ -664,7 +636,7 @@ fu_mm_device_at_cmd_cb(FuDevice *device, gpointer user_data, GError **error)
 
 	/* AT command has no response, return TRUE */
 	if (!helper->has_response) {
-		g_debug("No response expected for AT command: '%s', assuming succeed", helper->cmd);
+		g_debug("no response expected for AT command: '%s', assuming succeed", helper->cmd);
 		return TRUE;
 	}
 
@@ -678,9 +650,27 @@ fu_mm_device_at_cmd_cb(FuDevice *device, gpointer user_data, GError **error)
 		g_prefix_error(error, "failed to read response for %s: ", helper->cmd);
 		return FALSE;
 	}
-	if (g_getenv("FWUPD_MODEM_MANAGER_VERBOSE") != NULL)
-		fu_dump_bytes(G_LOG_DOMAIN, "read", at_res);
+	fu_dump_bytes(G_LOG_DOMAIN, "read", at_res);
 	buf = g_bytes_get_data(at_res, &bufsz);
+
+	/*
+	 * the first time the modem returns may be the command itself with one \n missing.
+	 * this is because the modem AT has enabled echo
+	 */
+	if (g_strrstr(buf, helper->cmd) != NULL && bufsz == strlen(helper->cmd) + 1) {
+		g_bytes_unref(at_res);
+		at_res = fu_io_channel_read_bytes(self->io_channel,
+						  -1,
+						  1500,
+						  FU_IO_CHANNEL_FLAG_SINGLE_SHOT,
+						  error);
+		if (at_res == NULL) {
+			g_prefix_error(error, "failed to read response for %s: ", helper->cmd);
+			return FALSE;
+		}
+		buf = g_bytes_get_data(at_res, &bufsz);
+	}
+
 	if (bufsz < 6) {
 		g_set_error(error,
 			    FWUPD_ERROR,
@@ -718,7 +708,7 @@ fu_mm_device_at_cmd_cb(FuDevice *device, gpointer user_data, GError **error)
 			if (g_strcmp0(parts[j], "") != 0 && g_strcmp0(parts[j], "OK") != 0) {
 				/* Set branch */
 				fu_device_set_branch(FU_DEVICE(self), parts[j]);
-				g_debug("Firmware branch reported as '%s'", parts[j]);
+				g_info("firmware branch reported as '%s'", parts[j]);
 				break;
 			}
 		}
@@ -1082,7 +1072,6 @@ fu_mm_device_write_firmware_qmi_pdc(FuDevice *device,
 	return TRUE;
 }
 
-#if MM_CHECK_VERSION(1, 17, 1) && MBIM_CHECK_VERSION(1, 25, 3)
 typedef struct {
 	FuDevice *device;
 	GMainLoop *mainloop;
@@ -1189,24 +1178,37 @@ fu_mm_device_writeln(const gchar *fn, const gchar *buf, GError **error)
 				       error);
 }
 
+static gboolean
+fu_mm_device_set_autosuspend_delay(FuMmDevice *self, guint timeout_ms, GError **error)
+{
+	g_autofree gchar *device_sysfs_path = NULL;
+	g_autofree gchar *autosuspend_delay_filename = NULL;
+	g_autofree gchar *buf = g_strdup_printf("%u", timeout_ms);
+
+	/* autosuspend delay updated for a proper firmware update */
+	fu_mm_utils_get_port_info(self->port_mbim, NULL, &device_sysfs_path, NULL, NULL);
+	autosuspend_delay_filename =
+	    g_build_filename(device_sysfs_path, "/power/autosuspend_delay_ms", NULL);
+	return fu_mm_device_writeln(autosuspend_delay_filename, buf, error);
+}
+
 static gboolean
 fu_mm_device_write_firmware_mbim_qdu(FuDevice *device,
 				     GBytes *fw,
 				     FuProgress *progress,
 				     GError **error)
 {
-	GBytes *data;
 	XbNode *part = NULL;
 	const gchar *filename = NULL;
 	const gchar *csum;
 	FuMmDevice *self = FU_MM_DEVICE(device);
-	g_autofree gchar *device_sysfs_path = NULL;
-	g_autofree gchar *autosuspend_delay_filename = NULL;
 	g_autofree gchar *csum_actual = NULL;
 	g_autofree gchar *version = NULL;
 	g_autoptr(FuArchive) archive = NULL;
 	g_autoptr(FuDeviceLocker) locker = NULL;
 	g_autoptr(GArray) digest = NULL;
+	g_autoptr(GBytes) data_xml = NULL;
+	g_autoptr(GBytes) data_part = NULL;
 	g_autoptr(XbBuilder) builder = xb_builder_new();
 	g_autoptr(XbBuilderSource) source = xb_builder_source_new();
 	g_autoptr(XbSilo) silo = NULL;
@@ -1224,10 +1226,10 @@ fu_mm_device_write_firmware_mbim_qdu(FuDevice *device,
 		return FALSE;
 
 	/* load the manifest of operations */
-	data = fu_archive_lookup_by_fn(archive, "flashfile.xml", error);
-	if (data == NULL)
+	data_xml = fu_archive_lookup_by_fn(archive, "flashfile.xml", error);
+	if (data_xml == NULL)
 		return FALSE;
-	if (!xb_builder_source_load_bytes(source, data, XB_BUILDER_SOURCE_FLAG_NONE, error))
+	if (!xb_builder_source_load_bytes(source, data_xml, XB_BUILDER_SOURCE_FLAG_NONE, error))
 		return FALSE;
 	xb_builder_import_source(builder, source);
 	silo = xb_builder_compile(builder, XB_BUILDER_COMPILE_FLAG_NONE, NULL, error);
@@ -1239,10 +1241,10 @@ fu_mm_device_write_firmware_mbim_qdu(FuDevice *device,
 		return FALSE;
 	filename = xb_node_get_attr(part, "filename");
 	csum = xb_node_get_attr(part, "MD5");
-	data = fu_archive_lookup_by_fn(archive, filename, error);
-	if (data == NULL)
+	data_part = fu_archive_lookup_by_fn(archive, filename, error);
+	if (data_part == NULL)
 		return FALSE;
-	csum_actual = g_compute_checksum_for_bytes(G_CHECKSUM_MD5, data);
+	csum_actual = g_compute_checksum_for_bytes(G_CHECKSUM_MD5, data_part);
 	if (g_strcmp0(csum, csum_actual) != 0) {
 		g_debug("[%s] MD5 not matched", filename);
 		g_set_error(error,
@@ -1251,21 +1253,17 @@ fu_mm_device_write_firmware_mbim_qdu(FuDevice *device,
 			    "[%s] MD5 not matched",
 			    filename);
 		return FALSE;
-	} else {
-		g_debug("[%s] MD5 matched", filename);
 	}
+	g_debug("[%s] MD5 matched", filename);
 
 	/* autosuspend delay updated for a proper firmware update */
-	fu_mm_utils_get_port_info(self->port_mbim, NULL, &device_sysfs_path, NULL, NULL);
-	autosuspend_delay_filename =
-	    g_build_filename(device_sysfs_path, "/power/autosuspend_delay_ms", NULL);
-	if (!fu_mm_device_writeln(autosuspend_delay_filename, "10000", error))
+	if (!fu_mm_device_set_autosuspend_delay(self, 20000, error))
 		return FALSE;
 
 	fu_progress_set_status(progress, FWUPD_STATUS_DEVICE_WRITE);
 	digest = fu_mbim_qdu_updater_write(self->mbim_qdu_updater,
 					   filename,
-					   data,
+					   data_part,
 					   device,
 					   progress,
 					   error);
@@ -1282,9 +1280,6 @@ fu_mm_device_write_firmware_mbim_qdu(FuDevice *device,
 	return TRUE;
 }
 
-#endif /* MM_CHECK_VERSION(1,17,1) && MBIM_CHECK_VERSION(1,25,3) */
-
-#if MM_CHECK_VERSION(1, 17, 2)
 static gboolean
 fu_mm_find_device_file(FuDevice *device, gpointer userdata, GError **error)
 {
@@ -1422,7 +1417,7 @@ fu_mm_device_firehose_write(FuMmDevice *self,
 					 error);
 }
 
-#if MM_CHECK_VERSION(1, 19, 1) && MBIM_CHECK_VERSION(1, 25, 7)
+#if MM_CHECK_VERSION(1, 19, 1)
 static gboolean
 fu_mm_device_sahara_open(FuMmDevice *self, GError **error)
 {
@@ -1438,7 +1433,7 @@ fu_mm_device_sahara_close(FuMmDevice *self, GError **error)
 	loader = g_steal_pointer(&self->sahara_loader);
 	return fu_sahara_loader_close(loader, error);
 }
-#endif // MM_CHECK_VERSION(1, 19, 1) && MBIM_CHECK_VERSION(1, 25, 7)
+#endif // MM_CHECK_VERSION(1, 19, 1)
 
 static gboolean
 fu_mm_setup_firmware_dir(FuMmDevice *self, GError **error)
@@ -1473,11 +1468,19 @@ fu_mm_copy_firehose_prog(FuMmDevice *self, GBytes *prog, GError **error)
 	g_autofree gchar *qcom_fw_dir = NULL;
 	g_autofree gchar *firehose_file_path = NULL;
 
+	if (self->firehose_prog_file == NULL) {
+		g_set_error(error,
+			    FWUPD_ERROR,
+			    FWUPD_ERROR_NOT_FOUND,
+			    "Firehose prog filename is not set for the device");
+		return FALSE;
+	}
+
 	qcom_fw_dir = g_build_filename(self->firmware_path, "qcom", NULL);
 	if (!fu_path_mkdir_parent(qcom_fw_dir, error))
 		return FALSE;
 
-	firehose_file_path = g_build_filename(qcom_fw_dir, "prog_firehose_sdx24.mbn", NULL);
+	firehose_file_path = g_build_filename(qcom_fw_dir, self->firehose_prog_file, NULL);
 
 	if (!fu_bytes_set_contents(firehose_file_path, prog, error))
 		return FALSE;
@@ -1510,11 +1513,11 @@ fu_mm_device_write_firmware_firehose(FuDevice *device,
 {
 	FuMmDevice *self = FU_MM_DEVICE(device);
 	MMModem *modem = mm_object_peek_modem(self->omodem);
-	GBytes *firehose_rawprogram;
-	GBytes *firehose_prog;
 	g_autoptr(FuDeviceLocker) locker = NULL;
 	g_autoptr(FuArchive) archive = NULL;
 	g_autoptr(XbSilo) firehose_rawprogram_silo = NULL;
+	g_autoptr(GBytes) firehose_prog = NULL;
+	g_autoptr(GBytes) firehose_rawprogram = NULL;
 	g_autoptr(GPtrArray) firehose_rawprogram_actions = NULL;
 
 	/* progress */
@@ -1582,7 +1585,7 @@ fu_mm_device_write_firmware_firehose(FuDevice *device,
 
 		g_debug("found edl port: %s", self->port_edl);
 	}
-#if MM_CHECK_VERSION(1, 19, 1) && MBIM_CHECK_VERSION(1, 25, 7)
+#if MM_CHECK_VERSION(1, 19, 1)
 	else if ((FU_MM_DEVICE(self)->update_methods & MM_MODEM_FIRMWARE_UPDATE_METHOD_SAHARA) &&
 		 self->port_mbim != NULL) {
 		/* switch to emergency download (EDL) execution environment */
@@ -1600,7 +1603,7 @@ fu_mm_device_write_firmware_firehose(FuDevice *device,
 		if (!fu_sahara_loader_run(self->sahara_loader, firehose_prog, error))
 			return FALSE;
 	}
-#endif // MM_CHECK_VERSION(1, 19, 1) && MBIM_CHECK_VERSION(1, 25, 7)
+#endif // MM_CHECK_VERSION(1, 19, 1)
 	else {
 		g_set_error(error,
 			    FWUPD_ERROR,
@@ -1624,8 +1627,6 @@ fu_mm_device_write_firmware_firehose(FuDevice *device,
 	return TRUE;
 }
 
-#endif /* MM_CHECK_VERSION(1,17,2) */
-
 static gboolean
 fu_mm_device_write_firmware(FuDevice *device,
 			    FuFirmware *firmware,
@@ -1654,17 +1655,13 @@ fu_mm_device_write_firmware(FuDevice *device,
 							   &self->qmi_pdc_active_id,
 							   error);
 
-#if MM_CHECK_VERSION(1, 17, 1) && MBIM_CHECK_VERSION(1, 25, 3)
 	/* mbim qdu write operation */
 	if (self->update_methods & MM_MODEM_FIRMWARE_UPDATE_METHOD_MBIM_QDU)
 		return fu_mm_device_write_firmware_mbim_qdu(device, fw, progress, error);
 
-#endif /* MM_CHECK_VERSION(1,17,1) && MBIM_CHECK_VERSION(1,25,3) */
-#if MM_CHECK_VERSION(1, 17, 2)
 	/* firehose operation */
 	if (self->update_methods & MM_MODEM_FIRMWARE_UPDATE_METHOD_FIREHOSE)
 		return fu_mm_device_write_firmware_firehose(device, fw, progress, error);
-#endif
 
 	g_set_error(error, FWUPD_ERROR, FWUPD_ERROR_NOT_SUPPORTED, "unsupported update method");
 	return FALSE;
@@ -1681,6 +1678,11 @@ fu_mm_device_set_quirk_kv(FuDevice *device, const gchar *key, const gchar *value
 		return TRUE;
 	}
 
+	if (g_strcmp0(key, "ModemManagerFirehoseProgFile") == 0) {
+		self->firehose_prog_file = g_strdup(value);
+		return TRUE;
+	}
+
 	/* failed */
 	g_set_error_literal(error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED, "quirk key not supported");
 	return FALSE;
@@ -1734,6 +1736,24 @@ fu_mm_device_attach_qmi_pdc_idle(gpointer user_data)
 	return G_SOURCE_REMOVE;
 }
 
+static gboolean
+fu_mm_device_cleanup(FuDevice *device,
+		     FuProgress *progress,
+		     FwupdInstallFlags install_flags,
+		     GError **error)
+{
+	FuMmDevice *self = FU_MM_DEVICE(device);
+
+	/* restore default configuration */
+	if (self->update_methods & MM_MODEM_FIRMWARE_UPDATE_METHOD_MBIM_QDU) {
+		if (!fu_mm_device_set_autosuspend_delay(self, 2000, error))
+			return FALSE;
+	}
+
+	/* success */
+	return TRUE;
+}
+
 static gboolean
 fu_mm_device_attach(FuDevice *device, FuProgress *progress, GError **error)
 {
@@ -1788,9 +1808,9 @@ fu_mm_device_setup_branch_at(FuMmDevice *self, GError **error)
 		return FALSE;
 
 	if (fu_device_get_branch(self) != NULL)
-		g_debug("using firmware branch: %s", fu_device_get_branch(self));
+		g_info("using firmware branch: %s", fu_device_get_branch(self));
 	else
-		g_debug("using firmware branch: default");
+		g_info("using firmware branch: default");
 
 	return TRUE;
 }
@@ -1865,6 +1885,11 @@ fu_mm_device_setup_secboot_status(FuDevice *device)
 	if (fu_device_has_vendor_id(device, "USB:0x2C7C") ||
 	    fu_device_has_vendor_id(device, "PCI:0x1EAC"))
 		fu_mm_device_setup_secboot_status_quectel(self);
+	else if (fu_device_has_vendor_id(device, "USB:0x2CB7")) {
+		fu_device_add_internal_flag(FU_DEVICE(self),
+					    FU_DEVICE_INTERNAL_FLAG_SAVE_INTO_BACKUP_REMOTE);
+		fu_device_add_flag(FU_DEVICE(self), FWUPD_DEVICE_FLAG_UNSIGNED_PAYLOAD);
+	}
 }
 
 static gboolean
@@ -1887,12 +1912,15 @@ fu_mm_device_incorporate(FuDevice *device, FuDevice *donor_device)
 	FuMmDevice *self = FU_MM_DEVICE(device);
 	FuMmDevice *donor = FU_MM_DEVICE(donor_device);
 
+	g_return_if_fail(FU_IS_MM_DEVICE(self));
+	g_return_if_fail(FU_IS_MM_DEVICE(donor));
+
 	self->update_methods = fu_mm_device_get_update_methods(donor);
-	self->detach_fastboot_at = g_strdup(fu_mm_device_get_detach_fastboot_at(donor));
+	self->detach_fastboot_at = g_strdup(donor->detach_fastboot_at);
 	self->inhibition_uid = g_strdup(fu_mm_device_get_inhibition_uid(donor));
-	self->port_at_ifnum = fu_mm_device_get_port_at_ifnum(donor);
-	self->port_qmi_ifnum = fu_mm_device_get_port_qmi_ifnum(donor);
-	self->port_mbim_ifnum = fu_mm_device_get_port_mbim_ifnum(donor);
+	self->port_at_ifnum = donor->port_at_ifnum;
+	self->port_qmi_ifnum = donor->port_qmi_ifnum;
+	self->port_mbim_ifnum = donor->port_mbim_ifnum;
 	g_set_object(&self->manager, donor->manager);
 }
 
@@ -1915,12 +1943,20 @@ fu_mm_device_init(FuMmDevice *self)
 	fu_device_add_flag(FU_DEVICE(self), FWUPD_DEVICE_FLAG_REQUIRE_AC);
 	fu_device_add_internal_flag(FU_DEVICE(self), FU_DEVICE_INTERNAL_FLAG_REPLUG_MATCH_GUID);
 	fu_device_add_internal_flag(FU_DEVICE(self), FU_DEVICE_INTERNAL_FLAG_MD_SET_VERFMT);
+	fu_device_add_internal_flag(FU_DEVICE(self), FU_DEVICE_INTERNAL_FLAG_ADD_INSTANCE_ID_REV);
+	fu_device_add_flag(FU_DEVICE(self), FWUPD_DEVICE_FLAG_UNSIGNED_PAYLOAD);
 	fu_device_set_version_format(FU_DEVICE(self), FWUPD_VERSION_FORMAT_PLAIN);
 	fu_device_set_summary(FU_DEVICE(self), "Mobile broadband device");
 	fu_device_add_icon(FU_DEVICE(self), "modem");
 	fu_device_register_private_flag(FU_DEVICE(self),
 					FU_MM_DEVICE_FLAG_DETACH_AT_FASTBOOT_HAS_NO_RESPONSE,
 					"detach-at-fastboot-has-no-response");
+	fu_device_register_private_flag(FU_DEVICE(self),
+					FU_MM_DEVICE_FLAG_UNINHIBIT_MM_AFTER_FASTBOOT_REBOOT,
+					"uninhibit-modemmanager-after-fastboot-reboot");
+	fu_device_register_private_flag(FU_DEVICE(self),
+					FU_MM_DEVICE_FLAG_USE_BRANCH,
+					"use-branch");
 }
 
 static void
@@ -1946,6 +1982,7 @@ fu_mm_device_finalize(GObject *object)
 	g_free(self->inhibition_uid);
 	g_free(self->firmware_path);
 	g_free(self->restore_firmware_path);
+	g_free(self->firehose_prog_file);
 	G_OBJECT_CLASS(fu_mm_device_parent_class)->finalize(object);
 }
 
@@ -1963,6 +2000,7 @@ fu_mm_device_class_init(FuMmDeviceClass *klass)
 	klass_device->detach = fu_mm_device_detach;
 	klass_device->write_firmware = fu_mm_device_write_firmware;
 	klass_device->attach = fu_mm_device_attach;
+	klass_device->cleanup = fu_mm_device_cleanup;
 	klass_device->set_progress = fu_mm_device_set_progress;
 	klass_device->incorporate = fu_mm_device_incorporate;
 
@@ -2007,13 +2045,6 @@ fu_mm_shadow_device_new(FuMmDevice *device)
 	return shadow_device;
 }
 
-FuUsbDevice *
-fu_mm_device_get_usb_device(FuMmDevice *self)
-{
-	g_return_val_if_fail(FU_IS_MM_DEVICE(self), NULL);
-	return self->usb_device;
-}
-
 void
 fu_mm_device_set_usb_device(FuMmDevice *self, FuUsbDevice *usb_device)
 {
