@@ -7,6 +7,7 @@
 #include "config.h"
 
 #include "fu-mei-common.h"
+#include "fu-mei-struct.h"
 #include "fu-pci-mei-plugin.h"
 
 struct _FuPciMeiPlugin {
@@ -56,6 +57,8 @@ fu_mei_detect_family(FuPlugin *plugin)
 	FuPciMeiPlugin *self = FU_PCI_MEI_PLUGIN(plugin);
 	guint8 ver = self->vers.major;
 
+	if (ver == 0)
+		return FU_MEI_FAMILY_UNKNOWN;
 	if (ver == 1 || ver == 2) {
 		if (self->hfsts1.fields.operation_mode == 0xf)
 			return FU_MEI_FAMILY_SPS;
@@ -65,9 +68,7 @@ fu_mei_detect_family(FuPlugin *plugin)
 		return FU_MEI_FAMILY_TXE;
 	if (ver == 6 || ver == 7 || ver == 8 || ver == 9 || ver == 10)
 		return FU_MEI_FAMILY_ME;
-	if (ver == 11 || ver == 12 || ver == 13 || ver == 14 || ver == 15 || ver == 16)
-		return FU_MEI_FAMILY_CSME;
-	return FU_MEI_FAMILY_UNKNOWN;
+	return FU_MEI_FAMILY_CSME;
 }
 
 static gboolean
@@ -147,18 +148,19 @@ fu_mei_parse_fwvers(FuPlugin *plugin, const gchar *fwvers, GError **error)
 		self->issue = fu_mei_common_is_txe_vulnerable(&self->vers);
 	else if (self->family == FU_MEI_FAMILY_SPS)
 		self->issue = fu_mei_common_is_sps_vulnerable(&self->vers);
-	if (g_getenv("FWUPD_MEI_VERBOSE") != NULL) {
-		g_debug("%s version parsed as %u.%u.%u",
-			fu_mei_common_family_to_string(self->family),
-			self->vers.major,
-			self->vers.minor,
-			self->vers.hotfix);
-	}
+	g_debug("%s version parsed as %u.%u.%u",
+		fu_mei_family_to_string(self->family),
+		self->vers.major,
+		self->vers.minor,
+		self->vers.hotfix);
 	return TRUE;
 }
 
 static gboolean
-fu_pci_mei_plugin_backend_device_added(FuPlugin *plugin, FuDevice *device, GError **error)
+fu_pci_mei_plugin_backend_device_added(FuPlugin *plugin,
+				       FuDevice *device,
+				       FuProgress *progress,
+				       GError **error)
 {
 	FuPciMeiPlugin *self = FU_PCI_MEI_PLUGIN(plugin);
 	const gchar *fwvers;
@@ -231,6 +233,7 @@ fu_plugin_add_security_attrs_manufacturing_mode(FuPlugin *plugin, FuSecurityAttr
 
 	/* create attr */
 	attr = fu_plugin_security_attr_new(plugin, FWUPD_SECURITY_ATTR_ID_MEI_MANUFACTURING_MODE);
+	fwupd_security_attr_set_result_success(attr, FWUPD_SECURITY_ATTR_RESULT_LOCKED);
 	fu_security_attrs_append(attrs, attr);
 
 	/* no device */
@@ -240,9 +243,7 @@ fu_plugin_add_security_attrs_manufacturing_mode(FuPlugin *plugin, FuSecurityAttr
 	}
 
 	/* Manufacturing Mode */
-	fwupd_security_attr_add_metadata(attr,
-					 "kind",
-					 fu_mei_common_family_to_string(self->family));
+	fwupd_security_attr_add_metadata(attr, "kind", fu_mei_family_to_string(self->family));
 	if (self->hfsts1.fields.mfg_mode) {
 		fwupd_security_attr_set_result(attr, FWUPD_SECURITY_ATTR_RESULT_NOT_LOCKED);
 		fwupd_security_attr_add_flag(attr, FWUPD_SECURITY_ATTR_FLAG_ACTION_CONTACT_OEM);
@@ -251,7 +252,6 @@ fu_plugin_add_security_attrs_manufacturing_mode(FuPlugin *plugin, FuSecurityAttr
 
 	/* success */
 	fwupd_security_attr_add_flag(attr, FWUPD_SECURITY_ATTR_FLAG_SUCCESS);
-	fwupd_security_attr_set_result(attr, FWUPD_SECURITY_ATTR_RESULT_LOCKED);
 }
 
 static void
@@ -262,6 +262,7 @@ fu_plugin_add_security_attrs_override_strap(FuPlugin *plugin, FuSecurityAttrs *a
 
 	/* create attr */
 	attr = fu_plugin_security_attr_new(plugin, FWUPD_SECURITY_ATTR_ID_MEI_OVERRIDE_STRAP);
+	fwupd_security_attr_set_result_success(attr, FWUPD_SECURITY_ATTR_RESULT_LOCKED);
 	fu_security_attrs_append(attrs, attr);
 
 	/* no device */
@@ -271,10 +272,8 @@ fu_plugin_add_security_attrs_override_strap(FuPlugin *plugin, FuSecurityAttrs *a
 	}
 
 	/* Flash Descriptor Security Override Strap */
-	fwupd_security_attr_add_metadata(attr,
-					 "kind",
-					 fu_mei_common_family_to_string(self->family));
-	if (self->hfsts1.fields.operation_mode == ME_HFS_MODE_OVER_JMPR) {
+	fwupd_security_attr_add_metadata(attr, "kind", fu_mei_family_to_string(self->family));
+	if (self->hfsts1.fields.operation_mode == FU_ME_HFS_MODE_OVERRIDE_JUMPER) {
 		fwupd_security_attr_set_result(attr, FWUPD_SECURITY_ATTR_RESULT_NOT_LOCKED);
 		fwupd_security_attr_add_flag(attr, FWUPD_SECURITY_ATTR_FLAG_ACTION_CONTACT_OEM);
 		return;
@@ -282,7 +281,6 @@ fu_plugin_add_security_attrs_override_strap(FuPlugin *plugin, FuSecurityAttrs *a
 
 	/* success */
 	fwupd_security_attr_add_flag(attr, FWUPD_SECURITY_ATTR_FLAG_SUCCESS);
-	fwupd_security_attr_set_result(attr, FWUPD_SECURITY_ATTR_RESULT_LOCKED);
 }
 
 static void
@@ -293,6 +291,7 @@ fu_plugin_add_security_attrs_bootguard_enabled(FuPlugin *plugin, FuSecurityAttrs
 
 	/* create attr */
 	attr = fu_plugin_security_attr_new(plugin, FWUPD_SECURITY_ATTR_ID_INTEL_BOOTGUARD_ENABLED);
+	fwupd_security_attr_set_result_success(attr, FWUPD_SECURITY_ATTR_RESULT_ENABLED);
 	fu_security_attrs_append(attrs, attr);
 
 	/* no device */
@@ -316,7 +315,6 @@ fu_plugin_add_security_attrs_bootguard_enabled(FuPlugin *plugin, FuSecurityAttrs
 
 	/* success */
 	fwupd_security_attr_add_flag(attr, FWUPD_SECURITY_ATTR_FLAG_SUCCESS);
-	fwupd_security_attr_set_result(attr, FWUPD_SECURITY_ATTR_RESULT_ENABLED);
 }
 
 static void
@@ -327,6 +325,7 @@ fu_plugin_add_security_attrs_bootguard_verified(FuPlugin *plugin, FuSecurityAttr
 
 	/* create attr */
 	attr = fu_plugin_security_attr_new(plugin, FWUPD_SECURITY_ATTR_ID_INTEL_BOOTGUARD_VERIFIED);
+	fwupd_security_attr_set_result_success(attr, FWUPD_SECURITY_ATTR_RESULT_VALID);
 	fu_security_attrs_append(attrs, attr);
 
 	/* no device */
@@ -356,7 +355,6 @@ fu_plugin_add_security_attrs_bootguard_verified(FuPlugin *plugin, FuSecurityAttr
 
 	/* success */
 	fwupd_security_attr_add_flag(attr, FWUPD_SECURITY_ATTR_FLAG_SUCCESS);
-	fwupd_security_attr_set_result(attr, FWUPD_SECURITY_ATTR_RESULT_VALID);
 }
 
 static void
@@ -367,6 +365,7 @@ fu_plugin_add_security_attrs_bootguard_acm(FuPlugin *plugin, FuSecurityAttrs *at
 
 	/* create attr */
 	attr = fu_plugin_security_attr_new(plugin, FWUPD_SECURITY_ATTR_ID_INTEL_BOOTGUARD_ACM);
+	fwupd_security_attr_set_result_success(attr, FWUPD_SECURITY_ATTR_RESULT_VALID);
 	fu_security_attrs_append(attrs, attr);
 
 	/* no device */
@@ -396,7 +395,6 @@ fu_plugin_add_security_attrs_bootguard_acm(FuPlugin *plugin, FuSecurityAttrs *at
 
 	/* success */
 	fwupd_security_attr_add_flag(attr, FWUPD_SECURITY_ATTR_FLAG_SUCCESS);
-	fwupd_security_attr_set_result(attr, FWUPD_SECURITY_ATTR_RESULT_VALID);
 }
 
 static void
@@ -407,6 +405,7 @@ fu_plugin_add_security_attrs_bootguard_policy(FuPlugin *plugin, FuSecurityAttrs
 
 	/* create attr */
 	attr = fu_plugin_security_attr_new(plugin, FWUPD_SECURITY_ATTR_ID_INTEL_BOOTGUARD_POLICY);
+	fwupd_security_attr_set_result_success(attr, FWUPD_SECURITY_ATTR_RESULT_VALID);
 	fu_security_attrs_append(attrs, attr);
 
 	/* no device */
@@ -427,8 +426,11 @@ fu_plugin_add_security_attrs_bootguard_policy(FuPlugin *plugin, FuSecurityAttrs
 		return;
 	}
 
-	/* policy must be to immediately shutdown */
-	if (self->hfsts6.fields.error_enforce_policy != ME_HFS_ENFORCEMENT_POLICY_SHUTDOWN_NOW) {
+	/* policy must be to immediately shutdown or after 30 mins -- the latter isn't ideal but
+	 * we've been testing for this accidentally for a long time now */
+	if (self->hfsts6.fields.error_enforce_policy != FU_ME_HFS_ENFORCEMENT_POLICY_SHUTDOWN_NOW &&
+	    self->hfsts6.fields.error_enforce_policy !=
+		FU_ME_HFS_ENFORCEMENT_POLICY_SHUTDOWN_30MINS) {
 		fwupd_security_attr_set_result(attr, FWUPD_SECURITY_ATTR_RESULT_NOT_VALID);
 		fwupd_security_attr_add_flag(attr, FWUPD_SECURITY_ATTR_FLAG_ACTION_CONTACT_OEM);
 		return;
@@ -436,7 +438,6 @@ fu_plugin_add_security_attrs_bootguard_policy(FuPlugin *plugin, FuSecurityAttrs
 
 	/* success */
 	fwupd_security_attr_add_flag(attr, FWUPD_SECURITY_ATTR_FLAG_SUCCESS);
-	fwupd_security_attr_set_result(attr, FWUPD_SECURITY_ATTR_RESULT_VALID);
 }
 
 static void
@@ -447,6 +448,7 @@ fu_plugin_add_security_attrs_bootguard_otp(FuPlugin *plugin, FuSecurityAttrs *at
 
 	/* create attr */
 	attr = fu_plugin_security_attr_new(plugin, FWUPD_SECURITY_ATTR_ID_INTEL_BOOTGUARD_OTP);
+	fwupd_security_attr_set_result_success(attr, FWUPD_SECURITY_ATTR_RESULT_VALID);
 	fu_security_attrs_append(attrs, attr);
 
 	/* no device */
@@ -476,7 +478,6 @@ fu_plugin_add_security_attrs_bootguard_otp(FuPlugin *plugin, FuSecurityAttrs *at
 
 	/* success */
 	fwupd_security_attr_add_flag(attr, FWUPD_SECURITY_ATTR_FLAG_SUCCESS);
-	fwupd_security_attr_set_result(attr, FWUPD_SECURITY_ATTR_RESULT_VALID);
 }
 
 static void
@@ -498,6 +499,7 @@ fu_plugin_add_security_attrs_mei_version(FuPlugin *plugin, FuSecurityAttrs *attr
 
 	/* create attr */
 	attr = fu_plugin_security_attr_new(plugin, FWUPD_SECURITY_ATTR_ID_MEI_VERSION);
+	fwupd_security_attr_set_result_success(attr, FWUPD_SECURITY_ATTR_RESULT_VALID);
 	fu_security_attrs_append(attrs, attr);
 
 	/* not enabled */
@@ -518,9 +520,7 @@ fu_plugin_add_security_attrs_mei_version(FuPlugin *plugin, FuSecurityAttrs *attr
 		return;
 	}
 	fwupd_security_attr_add_metadata(attr, "version", version);
-	fwupd_security_attr_add_metadata(attr,
-					 "kind",
-					 fu_mei_common_family_to_string(self->family));
+	fwupd_security_attr_add_metadata(attr, "kind", fu_mei_family_to_string(self->family));
 
 	/* Flash Descriptor Security Override Strap */
 	if (self->issue == FU_MEI_ISSUE_VULNERABLE) {
@@ -530,17 +530,25 @@ fu_plugin_add_security_attrs_mei_version(FuPlugin *plugin, FuSecurityAttrs *attr
 	}
 
 	/* success */
-	fwupd_security_attr_set_result(attr, FWUPD_SECURITY_ATTR_RESULT_VALID);
 	fwupd_security_attr_add_flag(attr, FWUPD_SECURITY_ATTR_FLAG_SUCCESS);
 }
 
 static void
 fu_pci_mei_plugin_add_security_attrs(FuPlugin *plugin, FuSecurityAttrs *attrs)
 {
+	g_autoptr(FwupdSecurityAttr) attr = NULL;
+
 	/* only Intel */
 	if (fu_cpu_get_vendor() != FU_CPU_VENDOR_INTEL)
 		return;
 
+	/* CPU supported */
+	attr = fu_security_attrs_get_by_appstream_id(attrs,
+						     FWUPD_SECURITY_ATTR_ID_SUPPORTED_CPU,
+						     NULL);
+	if (attr != NULL)
+		fwupd_security_attr_add_flag(attr, FWUPD_SECURITY_ATTR_FLAG_SUCCESS);
+
 	fu_plugin_add_security_attrs_manufacturing_mode(plugin, attrs);
 	fu_plugin_add_security_attrs_override_strap(plugin, attrs);
 	fu_plugin_add_security_attrs_bootguard(plugin, attrs);
@@ -574,8 +582,8 @@ fu_pci_mei_plugin_class_init(FuPciMeiPluginClass *klass)
 	FuPluginClass *plugin_class = FU_PLUGIN_CLASS(klass);
 	GObjectClass *object_class = G_OBJECT_CLASS(klass);
 
-	object_class->constructed = fu_pci_mei_plugin_constructed;
 	object_class->finalize = fu_pci_mei_finalize;
+	plugin_class->constructed = fu_pci_mei_plugin_constructed;
 	plugin_class->to_string = fu_pci_mei_plugin_to_string;
 	plugin_class->add_security_attrs = fu_pci_mei_plugin_add_security_attrs;
 	plugin_class->backend_device_added = fu_pci_mei_plugin_backend_device_added;
