@@ -7,9 +7,9 @@
 
 #include "config.h"
 
-#include <fcntl.h>
 #include <glib/gi18n.h>
 
+#include "fu-acpi-uefi.h"
 #include "fu-uefi-backend.h"
 #include "fu-uefi-bgrt.h"
 #include "fu-uefi-bootmgr.h"
@@ -17,10 +17,13 @@
 #include "fu-uefi-cod-device.h"
 #include "fu-uefi-common.h"
 #include "fu-uefi-grub-device.h"
+#include "fu-uefi-struct.h"
+#include "fu-uefi-update-info.h"
 
 struct _FuUefiCapsulePlugin {
 	FuPlugin parent_instance;
 	FuUefiBgrt *bgrt;
+	FuFirmware *acpi_uefi; /* optional */
 	FuVolume *esp;
 	FuBackend *backend;
 	GFile *fwupd_efi_file;
@@ -47,36 +50,32 @@ fu_uefi_capsule_plugin_fwupd_efi_parse(FuUefiCapsulePlugin *self, GError **error
 {
 	FuContext *ctx = fu_plugin_get_context(FU_PLUGIN(self));
 	const guint8 needle[] = "f\0w\0u\0p\0d\0-\0e\0f\0i\0 \0v\0e\0r\0s\0i\0o\0n\0 ";
-	gsize bufsz = 0;
 	gsize offset = 0;
-	guint16 version_tmp[16] = {0x0};
-	g_autofree gchar *buf = NULL;
+	g_autofree gchar *fn = g_file_get_path(self->fwupd_efi_file);
 	g_autofree gchar *version = NULL;
+	g_autoptr(GBytes) buf = NULL;
+	g_autoptr(GBytes) ubuf = NULL;
 
 	/* find the UTF-16 version string */
-	if (!g_file_load_contents(self->fwupd_efi_file, NULL, &buf, &bufsz, NULL, error))
+	buf = fu_bytes_get_contents(fn, error);
+	if (buf == NULL)
 		return FALSE;
-	if (!fu_memmem_safe((const guint8 *)buf, bufsz, needle, sizeof(needle), &offset, error)) {
-		g_autofree gchar *fn = g_file_get_path(self->fwupd_efi_file);
+	if (!fu_memmem_safe(g_bytes_get_data(buf, NULL),
+			    g_bytes_get_size(buf),
+			    needle,
+			    sizeof(needle),
+			    &offset,
+			    error)) {
 		g_prefix_error(error, "searching %s: ", fn);
 		return FALSE;
 	}
-
-	/* align */
-	if (!fu_memcpy_safe((guint8 *)version_tmp,
-			    sizeof(version_tmp),
-			    0x0, /* dst */
-			    (const guint8 *)buf,
-			    bufsz,
-			    offset + sizeof(needle),
-			    sizeof(version_tmp) - sizeof(guint16),
-			    error))
+	ubuf = fu_bytes_new_offset(buf, offset + sizeof(needle), 30, error);
+	if (ubuf == NULL)
 		return FALSE;
 
 	/* convert to UTF-8 */
-	version = g_utf16_to_utf8(version_tmp, -1, NULL, NULL, error);
+	version = fu_utf16_to_utf8_bytes(ubuf, G_LITTLE_ENDIAN, error);
 	if (version == NULL) {
-		g_autofree gchar *fn = g_file_get_path(self->fwupd_efi_file);
 		g_prefix_error(error, "converting %s: ", fn);
 		return FALSE;
 	}
@@ -111,7 +110,7 @@ fu_uefi_capsule_plugin_fwupd_efi_probe(FuUefiCapsulePlugin *self, GError **error
 	g_autofree gchar *fn = NULL;
 
 	/* find the app binary */
-	fn = fu_uefi_get_built_app_path(error);
+	fn = fu_uefi_get_built_app_path("fwupd", error);
 	if (fn == NULL)
 		return FALSE;
 	self->fwupd_efi_file = g_file_new_for_path(fn);
@@ -137,14 +136,37 @@ fu_uefi_capsule_plugin_clear_results(FuPlugin *plugin, FuDevice *device, GError
 	return fu_uefi_device_clear_status(device_uefi, error);
 }
 
+static gchar *
+fu_uefi_capsule_plugin_efivar_attrs_to_string(guint32 attrs)
+{
+	const gchar *data[7] = {0};
+	guint idx = 0;
+	if (attrs & FU_EFIVAR_ATTR_NON_VOLATILE)
+		data[idx++] = "non-volatile";
+	if (attrs & FU_EFIVAR_ATTR_BOOTSERVICE_ACCESS)
+		data[idx++] = "bootservice-access";
+	if (attrs & FU_EFIVAR_ATTR_RUNTIME_ACCESS)
+		data[idx++] = "runtime-access";
+	if (attrs & FU_EFIVAR_ATTR_HARDWARE_ERROR_RECORD)
+		data[idx++] = "hardware-error-record";
+	if (attrs & FU_EFIVAR_ATTR_AUTHENTICATED_WRITE_ACCESS)
+		data[idx++] = "authenticated-write-access";
+	if (attrs & FU_EFIVAR_ATTR_TIME_BASED_AUTHENTICATED_WRITE_ACCESS)
+		data[idx++] = "time-based-authenticated-write-access";
+	if (attrs & FU_EFIVAR_ATTR_APPEND_WRITE)
+		data[idx++] = "append-write";
+	return g_strjoinv(",", (gchar **)data);
+}
+
 static void
-fu_uefi_capsule_plugin_add_security_attrs(FuPlugin *plugin, FuSecurityAttrs *attrs)
+fu_uefi_capsule_plugin_add_security_attrs_secureboot(FuPlugin *plugin, FuSecurityAttrs *attrs)
 {
 	g_autoptr(FwupdSecurityAttr) attr = NULL;
 	g_autoptr(GError) error = NULL;
 
 	/* create attr */
 	attr = fu_plugin_security_attr_new(plugin, FWUPD_SECURITY_ATTR_ID_UEFI_SECUREBOOT);
+	fwupd_security_attr_set_result_success(attr, FWUPD_SECURITY_ATTR_RESULT_ENABLED);
 	fu_security_attrs_append(attrs, attr);
 
 	/* SB not available or disabled */
@@ -162,7 +184,71 @@ fu_uefi_capsule_plugin_add_security_attrs(FuPlugin *plugin, FuSecurityAttrs *att
 
 	/* success */
 	fwupd_security_attr_add_flag(attr, FWUPD_SECURITY_ATTR_FLAG_SUCCESS);
-	fwupd_security_attr_set_result(attr, FWUPD_SECURITY_ATTR_RESULT_ENABLED);
+}
+
+static void
+fu_uefi_capsule_plugin_add_security_attrs_bootservices(FuPlugin *plugin, FuSecurityAttrs *attrs)
+{
+	g_autoptr(FwupdSecurityAttr) attr = NULL;
+	const gchar *guids[] = {FU_EFIVAR_GUID_SECURITY_DATABASE, FU_EFIVAR_GUID_EFI_GLOBAL};
+
+	/* create attr */
+	attr = fu_plugin_security_attr_new(plugin, FWUPD_SECURITY_ATTR_ID_UEFI_BOOTSERVICE_VARS);
+	fwupd_security_attr_set_result_success(attr, FWUPD_SECURITY_ATTR_RESULT_LOCKED);
+	fu_security_attrs_append(attrs, attr);
+	for (guint j = 0; j < G_N_ELEMENTS(guids); j++) {
+		g_autoptr(GPtrArray) names = fu_efivar_get_names(guids[j], NULL);
+
+		/* sanity check */
+		if (names == NULL)
+			continue;
+		for (guint i = 0; i < names->len; i++) {
+			const gchar *name = g_ptr_array_index(names, i);
+			g_autoptr(GError) error_local = NULL;
+			gsize data_sz = 0;
+			guint32 data_attr = 0;
+
+			if (!fu_efivar_get_data(guids[j],
+						name,
+						NULL,
+						&data_sz,
+						&data_attr,
+						&error_local)) {
+				g_warning("failed to read %s-%s: %s",
+					  name,
+					  guids[j],
+					  error_local->message);
+				continue;
+			}
+			if ((data_attr & FU_EFIVAR_ATTR_BOOTSERVICE_ACCESS) > 0 &&
+			    (data_attr & FU_EFIVAR_ATTR_RUNTIME_ACCESS) == 0) {
+				g_debug("%s-%s attr of size 0x%x had flags %s",
+					name,
+					guids[j],
+					(guint)data_sz,
+					fu_uefi_capsule_plugin_efivar_attrs_to_string(data_attr));
+				fwupd_security_attr_add_metadata(attr, "guid", guids[j]);
+				fwupd_security_attr_add_metadata(attr, "name", name);
+				fwupd_security_attr_add_flag(
+				    attr,
+				    FWUPD_SECURITY_ATTR_FLAG_ACTION_CONTACT_OEM);
+				fwupd_security_attr_set_result(
+				    attr,
+				    FWUPD_SECURITY_ATTR_RESULT_NOT_LOCKED);
+				return;
+			}
+		}
+	}
+
+	/* success */
+	fwupd_security_attr_add_flag(attr, FWUPD_SECURITY_ATTR_FLAG_SUCCESS);
+}
+
+static void
+fu_uefi_capsule_plugin_add_security_attrs(FuPlugin *plugin, FuSecurityAttrs *attrs)
+{
+	fu_uefi_capsule_plugin_add_security_attrs_secureboot(plugin, attrs);
+	fu_uefi_capsule_plugin_add_security_attrs_bootservices(plugin, attrs);
 }
 
 static GBytes *
@@ -187,8 +273,8 @@ fu_uefi_capsule_plugin_get_splash_data(guint width, guint height, GError **error
 
 	/* find the closest locale match, falling back to `en` and `C` */
 	for (guint i = 0; langs[i] != NULL; i++) {
-		GBytes *blob_tmp;
 		g_autofree gchar *fn = NULL;
+		g_autoptr(GBytes) blob_tmp = NULL;
 		if (g_str_has_suffix(langs[i], ".UTF-8"))
 			continue;
 		fn = g_strdup_printf("fwupd-%s-%u-%u.bmp", langs[i], width, height);
@@ -222,16 +308,14 @@ fu_uefi_capsule_plugin_write_splash_data(FuUefiCapsulePlugin *self,
 	gssize size;
 	guint32 height, width;
 	guint8 csum = 0;
-	efi_ux_capsule_header_t header = {0};
-	efi_capsule_header_t capsule_header = {.flags =
-						   EFI_CAPSULE_HEADER_FLAGS_PERSIST_ACROSS_RESET,
-					       .guid = {0x0},
-					       .header_size = sizeof(efi_capsule_header_t),
-					       .capsule_image_size = 0};
+	fwupd_guid_t guid = {0x0};
+	g_autofree gchar *capsule_path = NULL;
 	g_autofree gchar *esp_path = NULL;
 	g_autofree gchar *fn = NULL;
-	g_autofree gchar *directory = NULL;
 	g_autofree gchar *basename = NULL;
+	g_autofree gchar *directory = NULL;
+	g_autoptr(GByteArray) st_cap = fu_struct_efi_capsule_header_new();
+	g_autoptr(GByteArray) st_uxh = fu_struct_efi_ux_capsule_header_new();
 	g_autoptr(GFile) ofile = NULL;
 	g_autoptr(GOutputStream) ostream = NULL;
 
@@ -247,11 +331,12 @@ fu_uefi_capsule_plugin_write_splash_data(FuUefiCapsulePlugin *self,
 		return FALSE;
 	}
 
-	/* save to a predicatable filename */
+	/* save to a predictable filename */
 	esp_path = fu_volume_get_mount_point(self->esp);
-	directory = fu_uefi_get_esp_path_for_os(device, esp_path);
+	directory = fu_uefi_get_esp_path_for_os();
 	basename = g_strdup_printf("fwupd-%s.cap", FU_EFIVAR_GUID_UX_CAPSULE);
-	fn = g_build_filename(directory, "fw", basename, NULL);
+	capsule_path = g_build_filename(directory, "fw", basename, NULL);
+	fn = g_build_filename(esp_path, capsule_path, NULL);
 	if (!fu_path_mkdir_parent(fn, error))
 		return FALSE;
 	ofile = g_file_new_for_path(fn);
@@ -260,40 +345,39 @@ fu_uefi_capsule_plugin_write_splash_data(FuUefiCapsulePlugin *self,
 	if (ostream == NULL)
 		return FALSE;
 
+	fu_struct_efi_capsule_header_set_flags(st_cap,
+					       EFI_CAPSULE_HEADER_FLAGS_PERSIST_ACROSS_RESET);
 	if (!fwupd_guid_from_string(FU_EFIVAR_GUID_UX_CAPSULE,
-				    &capsule_header.guid,
+				    &guid,
 				    FWUPD_GUID_FLAG_MIXED_ENDIAN,
 				    error))
 		return FALSE;
-	capsule_header.capsule_image_size =
-	    g_bytes_get_size(blob) + sizeof(efi_capsule_header_t) + sizeof(efi_ux_capsule_header_t);
+	fu_struct_efi_capsule_header_set_guid(st_cap, &guid);
+	fu_struct_efi_capsule_header_set_image_size(st_cap,
+						    g_bytes_get_size(blob) +
+							FU_STRUCT_EFI_CAPSULE_HEADER_SIZE +
+							FU_STRUCT_EFI_UX_CAPSULE_HEADER_SIZE);
 
-	header.version = 1;
-	header.image_type = 0;
-	header.reserved = 0;
-	header.x_offset = (screen_x / 2) - (width / 2);
+	fu_struct_efi_ux_capsule_header_set_x_offset(st_uxh, (screen_x / 2) - (width / 2));
 	if (screen_y == fu_uefi_bgrt_get_height(self->bgrt)) {
-		header.y_offset = (gdouble)screen_y * 0.8f;
+		fu_struct_efi_ux_capsule_header_set_y_offset(st_uxh, (gdouble)screen_y * 0.8f);
 	} else {
-		header.y_offset =
-		    fu_uefi_bgrt_get_yoffset(self->bgrt) + fu_uefi_bgrt_get_height(self->bgrt);
+		fu_struct_efi_ux_capsule_header_set_y_offset(
+		    st_uxh,
+		    fu_uefi_bgrt_get_yoffset(self->bgrt) + fu_uefi_bgrt_get_height(self->bgrt));
 	};
 
 	/* header, payload and image has to add to zero */
-	csum += fu_sum8((guint8 *)&capsule_header, sizeof(capsule_header));
-	csum += fu_sum8((guint8 *)&header, sizeof(header));
+	csum += fu_sum8(st_cap->data, st_cap->len);
+	csum += fu_sum8(st_uxh->data, st_uxh->len);
 	csum += fu_sum8_bytes(blob);
-	header.checksum = 0x100 - csum;
+	fu_struct_efi_ux_capsule_header_set_checksum(st_uxh, 0x100 - csum);
 
 	/* write capsule file */
-	size = g_output_stream_write(ostream,
-				     &capsule_header,
-				     capsule_header.header_size,
-				     NULL,
-				     error);
+	size = g_output_stream_write(ostream, st_cap->data, st_cap->len, NULL, error);
 	if (size < 0)
 		return FALSE;
-	size = g_output_stream_write(ostream, &header, sizeof(header), NULL, error);
+	size = g_output_stream_write(ostream, st_uxh->data, st_uxh->len, NULL, error);
 	if (size < 0)
 		return FALSE;
 	size = g_output_stream_write_bytes(ostream, blob, NULL, error);
@@ -302,7 +386,7 @@ fu_uefi_capsule_plugin_write_splash_data(FuUefiCapsulePlugin *self,
 
 	/* write display capsule location as UPDATE_INFO */
 	return fu_uefi_device_write_update_info(FU_UEFI_DEVICE(device),
-						fn,
+						capsule_path,
 						"fwupd-ux-capsule",
 						FU_EFIVAR_GUID_UX_CAPSULE,
 						error);
@@ -333,7 +417,7 @@ fu_uefi_capsule_plugin_update_splash(FuPlugin *plugin, FuDevice *device, GError
 
 	/* no UX capsule support, so deleting var if it exists */
 	if (fu_device_has_private_flag(device, FU_UEFI_DEVICE_FLAG_NO_UX_CAPSULE)) {
-		g_debug("not providing UX capsule");
+		g_info("not providing UX capsule");
 		return fu_efivar_delete(FU_EFIVAR_GUID_FWUPDATE, "fwupd-ux-capsule", error);
 	}
 
@@ -425,9 +509,8 @@ fu_uefi_capsule_plugin_write_firmware(FuPlugin *plugin,
 
 	/* perform the update */
 	fu_progress_set_status(progress, FWUPD_STATUS_SCHEDULING);
-	if (!fu_uefi_capsule_plugin_update_splash(plugin, device, &error_splash)) {
-		g_debug("failed to upload UEFI UX capsule text: %s", error_splash->message);
-	}
+	if (!fu_uefi_capsule_plugin_update_splash(plugin, device, &error_splash))
+		g_info("failed to upload UEFI UX capsule text: %s", error_splash->message);
 
 	return fu_device_write_firmware(device, blob_fw, progress, flags, error);
 }
@@ -435,51 +518,41 @@ fu_uefi_capsule_plugin_write_firmware(FuPlugin *plugin,
 static void
 fu_uefi_capsule_plugin_load_config(FuPlugin *plugin, FuDevice *device)
 {
-	gboolean disable_shim;
-	gboolean fallback_removable_path;
-	guint64 sz_reqd = FU_UEFI_COMMON_REQUIRED_ESP_FREE_SPACE;
+	guint64 sz_reqd = 0;
 	g_autofree gchar *require_esp_free_space = NULL;
 	g_autoptr(GError) error_local = NULL;
 
 	/* parse free space needed for ESP */
 	require_esp_free_space = fu_plugin_get_config_value(plugin, "RequireESPFreeSpace");
-	if (require_esp_free_space != NULL) {
-		if (!fu_strtoull(require_esp_free_space, &sz_reqd, 0, G_MAXUINT64, &error_local)) {
-			g_warning("invalid ESP free space specified: %s", error_local->message);
-		}
-	}
-	fu_device_set_metadata_integer(device, "RequireESPFreeSpace", sz_reqd);
+	if (!fu_strtoull(require_esp_free_space, &sz_reqd, 0, G_MAXUINT64, &error_local))
+		g_warning("invalid ESP free space specified: %s", error_local->message);
+	fu_uefi_device_set_require_esp_free_space(FU_UEFI_DEVICE(device), sz_reqd);
 
 	/* shim used for SB or not? */
-	disable_shim = fu_plugin_get_config_value_boolean(plugin, "DisableShimForSecureBoot");
-	if (!disable_shim)
+	if (!fu_plugin_get_config_value_boolean(plugin, "DisableShimForSecureBoot"))
 		fu_device_add_private_flag(device, FU_UEFI_DEVICE_FLAG_USE_SHIM_FOR_SB);
 
-	/* check if using UEFI removable path */
-	fallback_removable_path =
-	    fu_plugin_get_config_value_boolean(plugin, "FallbacktoRemovablePath");
-	if (fallback_removable_path)
-		fu_device_add_private_flag(device, FU_UEFI_DEVICE_FLAG_FALLBACK_TO_REMOVABLE_PATH);
+	/* enable the fwupd.efi debug log? */
+	if (fu_plugin_get_config_value_boolean(plugin, "EnableEfiDebugging"))
+		fu_device_add_private_flag(device, FU_UEFI_DEVICE_FLAG_ENABLE_EFI_DEBUGGING);
 }
 
 static gboolean
 fu_uefi_capsule_plugin_is_esp_linux(FuVolume *esp, GError **error)
 {
-	const gchar *basenames_root[] = {"grub", "shim", "systemd-boot", NULL};
-	const gchar *efi_suffix;
-	g_autofree gchar *basenames_str = NULL;
+	const gchar *prefixes[] = {"grub", "shim", "systemd-boot", "zfsbootmenu", NULL};
+	g_autofree gchar *prefixes_str = NULL;
 	g_autofree gchar *mount_point = fu_volume_get_mount_point(esp);
-	g_auto(GStrv) basenames = g_new0(gchar *, G_N_ELEMENTS(basenames_root));
 	g_autoptr(GPtrArray) files = NULL;
 
-	/* build a list of possible files, e.g. grubx64.efi, shimaa64.efi, etc. */
-	efi_suffix = fu_uefi_bootmgr_get_suffix(error);
-	if (efi_suffix == NULL)
-		return FALSE;
-	for (guint i = 0; basenames_root[i] != NULL; i++)
-		basenames[i] = g_strdup_printf("%s%s.efi", basenames_root[i], efi_suffix);
-
 	/* look for any likely basenames */
+	if (mount_point == NULL) {
+		g_set_error_literal(error,
+				    G_IO_ERROR,
+				    G_IO_ERROR_NOT_SUPPORTED,
+				    "no mountpoint for ESP");
+		return FALSE;
+	}
 	files = fu_path_get_files(mount_point, error);
 	if (files == NULL)
 		return FALSE;
@@ -487,71 +560,205 @@ fu_uefi_capsule_plugin_is_esp_linux(FuVolume *esp, GError **error)
 		const gchar *fn = g_ptr_array_index(files, i);
 		g_autofree gchar *basename = g_path_get_basename(fn);
 		g_autofree gchar *basename_lower = g_utf8_strdown(basename, -1);
-		if (g_strv_contains((const gchar *const *)basenames, basename_lower)) {
-			g_debug("found %s which indicates a Linux ESP, using %s", fn, mount_point);
+
+		for (guint j = 0; prefixes[j] != NULL; j++) {
+			if (!g_str_has_prefix(basename_lower, prefixes[j]))
+				continue;
+			if (!g_str_has_suffix(basename_lower, ".efi"))
+				continue;
+			g_info("found %s which indicates a Linux ESP, using %s", fn, mount_point);
 			return TRUE;
 		}
 	}
 
 	/* failed */
-	basenames_str = g_strjoinv("|", (gchar **)basenames);
+	prefixes_str = g_strjoinv("|", (gchar **)prefixes);
 	g_set_error(error,
-		    G_IO_ERROR,
-		    G_IO_ERROR_NOT_FOUND,
-		    "did not find %s in %s",
-		    basenames_str,
+		    FWUPD_ERROR,
+		    FWUPD_ERROR_NOT_FOUND,
+		    "did not any files with prefix %s in %s",
+		    prefixes_str,
 		    mount_point);
 	return FALSE;
 }
 
+static gint
+fu_uefi_capsule_plugin_sort_volume_score_cb(gconstpointer a, gconstpointer b, gpointer user_data)
+{
+	GHashTable *esp_scores = (GHashTable *)user_data;
+	guint esp1_score = GPOINTER_TO_UINT(g_hash_table_lookup(esp_scores, *((FuVolume **)a)));
+	guint esp2_score = GPOINTER_TO_UINT(g_hash_table_lookup(esp_scores, *((FuVolume **)b)));
+	if (esp1_score < esp2_score)
+		return 1;
+	if (esp1_score > esp2_score)
+		return -1;
+	return 0;
+}
+
 static FuVolume *
 fu_uefi_capsule_plugin_get_default_esp(FuPlugin *plugin, GError **error)
 {
 	g_autoptr(GPtrArray) esp_volumes = NULL;
+	const gchar *recovery_partitions[] = {
+	    "DELLRESTORE",
+	    "DELLUTILITY",
+	    "DIAGS",
+	    "HP_RECOVERY",
+	    "IBM_SERVICE",
+	    "INTELRST",
+	    "LENOVO_RECOVERY",
+	    "OS",
+	    "PQSERVICE",
+	    "RECOVERY",
+	    "RECOVERY_PARTITION",
+	    "SERVICEV001",
+	    "SERVICEV002",
+	    "SYSTEM_RESERVED",
+	    "WINRE_DRV",
+	    NULL,
+	}; /* from https://github.com/storaged-project/udisks/blob/master/data/80-udisks2.rules */
+	const gchar *user_esp_location = fu_context_get_esp_location(fu_plugin_get_context(plugin));
 
 	/* show which volumes we're choosing from */
 	esp_volumes = fu_context_get_esp_volumes(fu_plugin_get_context(plugin), error);
 	if (esp_volumes == NULL)
 		return NULL;
-	if (esp_volumes->len > 1) {
-		g_autoptr(GString) str = g_string_new(NULL);
-		for (guint i = 0; i < esp_volumes->len; i++) {
-			FuVolume *vol = g_ptr_array_index(esp_volumes, i);
-			if (str->len > 0)
-				g_string_append_c(str, ',');
-			g_string_append(str, fu_volume_get_id(vol));
-		}
-		g_debug("more than one ESP possible: %s", str->str);
-	}
 
-	/* we found more than one: lets look for something plausible */
+	/* we found more than one: lets look for the best one */
 	if (esp_volumes->len > 1) {
+		g_autoptr(GString) str = g_string_new("more than one ESP possible:");
+		g_autoptr(GHashTable) esp_scores = g_hash_table_new(g_direct_hash, g_direct_equal);
 		for (guint i = 0; i < esp_volumes->len; i++) {
 			FuVolume *esp = g_ptr_array_index(esp_volumes, i);
+			guint score = 0;
+			g_autofree gchar *name = NULL;
+			g_autofree gchar *kind = NULL;
 			g_autoptr(FuDeviceLocker) locker = NULL;
 			g_autoptr(GError) error_local = NULL;
 
+			/* ignore the volume completely if we cannot mount it */
 			locker = fu_volume_locker(esp, &error_local);
 			if (locker == NULL) {
 				g_warning("failed to mount ESP: %s", error_local->message);
 				continue;
 			}
+
+			/* if user specified, make sure that it matches */
+			if (user_esp_location != NULL) {
+				g_autofree gchar *mount = fu_volume_get_mount_point(esp);
+				if (g_strcmp0(mount, user_esp_location) != 0) {
+					g_debug("skipping %s as it's not the user "
+						"specified ESP",
+						mount);
+					continue;
+				}
+			}
+
+			/* ignore a partition that claims to be a recovery partition */
+			name = fu_volume_get_partition_name(esp);
+			if (name != NULL) {
+				g_autoptr(GString) name_safe = g_string_new(name);
+				g_string_replace(name_safe, " ", "_", 0);
+				g_string_replace(name_safe, "\"", "", 0);
+				g_string_ascii_up(name_safe);
+				if (g_strv_contains(recovery_partitions, name_safe->str)) {
+					if (g_strcmp0(name_safe->str, name) == 0) {
+						g_debug("skipping partition '%s'", name);
+					} else {
+						g_debug("skipping partition '%s' -> '%s'",
+							name,
+							name_safe->str);
+					}
+					continue;
+				}
+			}
+
+			/* big partitions are better than small partitions */
+			score += fu_volume_get_size(esp) / (1024 * 1024);
+
+			/* prefer partitions with the ESP flag set over msftdata */
+			kind = fu_volume_get_partition_kind(esp);
+			if (g_strcmp0(kind, FU_VOLUME_KIND_ESP) == 0)
+				score += 0x20000;
+
+			/* prefer linux ESP */
 			if (!fu_uefi_capsule_plugin_is_esp_linux(esp, &error_local)) {
 				g_debug("not a Linux ESP: %s", error_local->message);
-				continue;
+			} else {
+				score += 0x10000;
 			}
-			return g_object_ref(esp);
+			g_hash_table_insert(esp_scores, (gpointer)esp, GUINT_TO_POINTER(score));
 		}
 
-		/* we never found anything plausible */
-		g_warning("more than one ESP possible -- using %s because it is listed first",
-			  fu_volume_get_id(FU_VOLUME(g_ptr_array_index(esp_volumes, 0))));
+		if (g_hash_table_size(esp_scores) == 0) {
+			g_set_error(error,
+				    G_IO_ERROR,
+				    G_IO_ERROR_NOT_SUPPORTED,
+				    "no EFI system partition found");
+			return NULL;
+		}
+
+		g_ptr_array_sort_with_data(esp_volumes,
+					   fu_uefi_capsule_plugin_sort_volume_score_cb,
+					   esp_scores);
+		for (guint i = 0; i < esp_volumes->len; i++) {
+			FuVolume *esp = g_ptr_array_index(esp_volumes, i);
+			guint score = GPOINTER_TO_UINT(g_hash_table_lookup(esp_scores, esp));
+			g_string_append_printf(str, "\n - 0x%x:\t%s", score, fu_volume_get_id(esp));
+		}
+		g_debug("%s", str->str);
+	}
+
+	if (esp_volumes->len == 1) {
+		FuVolume *esp = g_ptr_array_index(esp_volumes, 0);
+		g_autoptr(FuDeviceLocker) locker = NULL;
+
+		/* ensure it can be mounted */
+		locker = fu_volume_locker(esp, error);
+		if (locker == NULL)
+			return NULL;
+
+		/* if user specified, does it match mountpoints ? */
+		if (user_esp_location != NULL) {
+			g_autofree gchar *mount = fu_volume_get_mount_point(esp);
+
+			if (g_strcmp0(mount, user_esp_location) != 0) {
+				g_set_error(error,
+					    G_IO_ERROR,
+					    G_IO_ERROR_NOT_SUPPORTED,
+					    "user specified ESP %s not found",
+					    user_esp_location);
+				return NULL;
+			}
+		}
 	}
 
 	/* "success" */
 	return g_object_ref(g_ptr_array_index(esp_volumes, 0));
 }
 
+static void
+fu_uefi_capsule_plugin_validate_esp(FuUefiCapsulePlugin *self)
+{
+	g_autofree gchar *kind = NULL;
+
+	/* nothing found */
+	if (self->esp == NULL) {
+		fu_plugin_add_flag(FU_PLUGIN(self), FWUPD_PLUGIN_FLAG_ESP_NOT_FOUND);
+		fu_plugin_add_flag(FU_PLUGIN(self), FWUPD_PLUGIN_FLAG_CLEAR_UPDATABLE);
+		fu_plugin_add_flag(FU_PLUGIN(self), FWUPD_PLUGIN_FLAG_USER_WARNING);
+		return;
+	}
+
+	/* found *something* but it wasn't quite an ESP... */
+	kind = fu_volume_get_partition_kind(self->esp);
+	if (kind != NULL &&
+	    g_strcmp0(fu_volume_kind_convert_to_gpt(kind), FU_VOLUME_KIND_ESP) != 0) {
+		fu_plugin_add_flag(FU_PLUGIN(self), FWUPD_PLUGIN_FLAG_ESP_NOT_VALID);
+		fu_plugin_add_flag(FU_PLUGIN(self), FWUPD_PLUGIN_FLAG_USER_WARNING);
+	}
+}
+
 static void
 fu_uefi_capsule_plugin_register_proxy_device(FuPlugin *plugin, FuDevice *device)
 {
@@ -562,8 +769,12 @@ fu_uefi_capsule_plugin_register_proxy_device(FuPlugin *plugin, FuDevice *device)
 	/* load all configuration variables */
 	dev = fu_uefi_backend_device_new_from_dev(FU_UEFI_BACKEND(self->backend), device);
 	fu_uefi_capsule_plugin_load_config(plugin, FU_DEVICE(dev));
-	if (self->esp == NULL)
+	if (self->esp == NULL) {
 		self->esp = fu_uefi_capsule_plugin_get_default_esp(plugin, &error_local);
+		if (self->esp == NULL)
+			g_warning("cannot find default ESP: %s", error_local->message);
+		fu_uefi_capsule_plugin_validate_esp(self);
+	}
 	if (self->esp == NULL) {
 		fu_device_inhibit(device, "no-esp", error_local->message);
 	} else {
@@ -617,6 +828,7 @@ fu_uefi_capsule_plugin_get_name_for_type(FuPlugin *plugin, FuUefiDeviceKind devi
 static gboolean
 fu_uefi_capsule_plugin_coldplug_device(FuPlugin *plugin, FuUefiDevice *dev, GError **error)
 {
+	FuUefiCapsulePlugin *self = FU_UEFI_CAPSULE_PLUGIN(plugin);
 	FuContext *ctx = fu_plugin_get_context(plugin);
 	FuUefiDeviceKind device_kind;
 
@@ -639,6 +851,21 @@ fu_uefi_capsule_plugin_coldplug_device(FuPlugin *plugin, FuUefiDevice *dev, GErr
 		fu_device_add_private_flag(FU_DEVICE(dev), FU_UEFI_DEVICE_FLAG_NO_UX_CAPSULE);
 	if (fu_context_has_hwid_flag(ctx, "no-lid-closed"))
 		fu_device_add_internal_flag(FU_DEVICE(dev), FU_DEVICE_INTERNAL_FLAG_NO_LID_CLOSED);
+	if (fu_context_has_hwid_flag(ctx, "display-required")) {
+		fu_device_add_internal_flag(FU_DEVICE(dev),
+					    FU_DEVICE_INTERNAL_FLAG_DISPLAY_REQUIRED);
+	}
+	if (fu_context_has_hwid_flag(ctx, "modify-bootorder"))
+		fu_device_add_private_flag(FU_DEVICE(dev), FU_UEFI_DEVICE_FLAG_MODIFY_BOOTORDER);
+	if (fu_context_has_hwid_flag(ctx, "cod-dell-recovery"))
+		fu_device_add_private_flag(FU_DEVICE(dev), FU_UEFI_DEVICE_FLAG_COD_DELL_RECOVERY);
+
+	/* detected InsydeH2O */
+	if (self->acpi_uefi != NULL &&
+	    fu_acpi_uefi_cod_indexed_filename(FU_ACPI_UEFI(self->acpi_uefi))) {
+		fu_device_add_private_flag(FU_DEVICE(dev),
+					   FU_UEFI_DEVICE_FLAG_COD_INDEXED_FILENAME);
+	}
 
 	/* set fallback name if nothing else is set */
 	device_kind = fu_uefi_device_get_kind(dev);
@@ -669,6 +896,7 @@ fu_uefi_capsule_plugin_coldplug_device(FuPlugin *plugin, FuUefiDevice *dev, GErr
 			fu_device_add_vendor_id(FU_DEVICE(dev), vendor_id);
 		}
 	}
+	fu_device_add_request_flag(FU_DEVICE(dev), FWUPD_REQUEST_FLAG_ALLOW_GENERIC_MESSAGE);
 
 	/* success */
 	return TRUE;
@@ -683,15 +911,32 @@ fu_uefi_capsule_plugin_test_secure_boot(FuPlugin *plugin)
 	fu_plugin_add_report_metadata(plugin, "SecureBoot", result_str);
 }
 
+static gboolean
+fu_uefi_capsule_plugin_parse_acpi_uefi(FuUefiCapsulePlugin *self, GError **error)
+{
+	g_autofree gchar *fn = NULL;
+	g_autofree gchar *path = NULL;
+	g_autoptr(GBytes) blob = NULL;
+
+	/* if we have a table, parse it and validate it */
+	path = fu_path_from_kind(FU_PATH_KIND_ACPI_TABLES);
+	fn = g_build_filename(path, "UEFI", NULL);
+	blob = fu_bytes_get_contents(fn, error);
+	if (blob == NULL)
+		return FALSE;
+	self->acpi_uefi = fu_acpi_uefi_new();
+	return fu_firmware_parse(self->acpi_uefi, blob, FWUPD_INSTALL_FLAG_NONE, error);
+}
+
 static gboolean
 fu_uefi_capsule_plugin_startup(FuPlugin *plugin, FuProgress *progress, GError **error)
 {
 	FuUefiCapsulePlugin *self = FU_UEFI_CAPSULE_PLUGIN(plugin);
 	FuContext *ctx = fu_plugin_get_context(plugin);
 	guint64 nvram_total;
-	g_autofree gchar *esp_path = NULL;
 	g_autofree gchar *nvram_total_str = NULL;
 	g_autoptr(GError) error_local = NULL;
+	g_autoptr(GError) error_acpi_uefi = NULL;
 
 	/* don't let user's environment influence test suite failures */
 	if (g_getenv("FWUPD_UEFI_TEST") != NULL)
@@ -731,18 +976,9 @@ fu_uefi_capsule_plugin_startup(FuPlugin *plugin, FuProgress *progress, GError **
 	nvram_total_str = g_strdup_printf("%" G_GUINT64_FORMAT, nvram_total);
 	fu_plugin_add_report_metadata(plugin, "EfivarNvramUsed", nvram_total_str);
 
-	/* override the default ESP path */
-	esp_path = fu_plugin_get_config_value(plugin, "OverrideESPMountPoint");
-	if (esp_path != NULL) {
-		self->esp = fu_volume_new_esp_for_path(esp_path, error);
-		if (self->esp == NULL) {
-			g_prefix_error(error,
-				       "invalid OverrideESPMountPoint=%s "
-				       "specified in config: ",
-				       esp_path);
-			return FALSE;
-		}
-	}
+	/* we use this both for quirking the CoD implementation sanity and the CoD filename */
+	if (!fu_uefi_capsule_plugin_parse_acpi_uefi(self, &error_acpi_uefi))
+		g_debug("failed to load ACPI UEFI table: %s", error_acpi_uefi->message);
 
 	/* test for invalid ESP in coldplug, and set the update-error rather
 	 * than showing no output if the plugin had self-disabled here */
@@ -768,7 +1004,7 @@ fu_uefi_capsule_plugin_unlock(FuPlugin *plugin, FuDevice *device, GError **error
 	}
 
 	/* for unlocking TPM1.2 <-> TPM2.0 switching */
-	g_debug("Unlocking upgrades for: %s (%s)",
+	g_debug("unlocking upgrades for: %s (%s)",
 		fu_device_get_name(device),
 		fu_device_get_id(device));
 	device_alt = fu_device_get_alternate(device);
@@ -780,7 +1016,7 @@ fu_uefi_capsule_plugin_unlock(FuPlugin *plugin, FuDevice *device, GError **error
 			    fu_device_get_name(device));
 		return FALSE;
 	}
-	g_debug("Preventing upgrades for: %s (%s)",
+	g_debug("preventing upgrades for: %s (%s)",
 		fu_device_get_name(device_alt),
 		fu_device_get_id(device_alt));
 
@@ -846,14 +1082,11 @@ fu_plugin_uefi_update_state_notify_cb(GObject *object, GParamSpec *pspec, FuPlug
 }
 
 static gboolean
-fu_uefi_capsule_plugin_check_cod_support(FuPlugin *plugin, GError **error)
+fu_uefi_capsule_plugin_check_cod_support(FuUefiCapsulePlugin *self, GError **error)
 {
 	gsize bufsz = 0;
 	guint64 value = 0;
 	g_autofree guint8 *buf = NULL;
-#ifndef __aarch64__
-	FuContext *ctx = fu_plugin_get_context(plugin);
-#endif
 
 	if (!fu_efivar_get_data(FU_EFIVAR_GUID_EFI_GLOBAL,
 				"OsIndicationsSupported",
@@ -874,21 +1107,10 @@ fu_uefi_capsule_plugin_check_cod_support(FuPlugin *plugin, GError **error)
 		return FALSE;
 	}
 
-#ifndef __aarch64__
-	/* several models with Insyde firmware have been released where OsIndications advertises
-	 * support for CoD, but it simply does not work */
-	if (!fu_context_has_hwid_flag(ctx, "uefi-allow-cod")) {
-		g_set_error_literal(error,
-				    G_IO_ERROR,
-				    G_IO_ERROR_NOT_SUPPORTED,
-				    "Capsule-on-Disk is supported, but non-aarch64 platforms "
-				    "require the 'uefi-allow-cod' quirk for it to be used");
-		return FALSE;
-	}
-#endif
-
-	/* success */
-	return TRUE;
+	/* no table, nothing to check */
+	if (self->acpi_uefi == NULL)
+		return TRUE;
+	return fu_acpi_uefi_cod_functional(FU_ACPI_UEFI(self->acpi_uefi), error);
 }
 
 static gboolean
@@ -913,19 +1135,16 @@ fu_uefi_capsule_plugin_coldplug(FuPlugin *plugin, FuProgress *progress, GError *
 
 	if (self->esp == NULL) {
 		self->esp = fu_uefi_capsule_plugin_get_default_esp(plugin, &error_udisks2);
-		if (self->esp == NULL) {
-			fu_plugin_add_flag(plugin, FWUPD_PLUGIN_FLAG_ESP_NOT_FOUND);
-			fu_plugin_add_flag(plugin, FWUPD_PLUGIN_FLAG_CLEAR_UPDATABLE);
-			fu_plugin_add_flag(plugin, FWUPD_PLUGIN_FLAG_USER_WARNING);
+		if (self->esp == NULL)
 			g_warning("cannot find default ESP: %s", error_udisks2->message);
-		}
+		fu_uefi_capsule_plugin_validate_esp(self);
 	}
 	fu_progress_step_done(progress);
 
 	/* firmware may lie */
 	if (!fu_plugin_get_config_value_boolean(plugin, "DisableCapsuleUpdateOnDisk")) {
 		g_autoptr(GError) error_cod = NULL;
-		if (!fu_uefi_capsule_plugin_check_cod_support(plugin, &error_cod)) {
+		if (!fu_uefi_capsule_plugin_check_cod_support(self, &error_cod)) {
 			g_debug("not using CapsuleOnDisk support: %s", error_cod->message);
 		} else {
 			fu_uefi_backend_set_device_gtype(FU_UEFI_BACKEND(self->backend),
@@ -986,17 +1205,132 @@ fu_uefi_capsule_plugin_coldplug(FuPlugin *plugin, FuProgress *progress, GError *
 	if (!fu_uefi_bgrt_setup(self->bgrt, &error_local))
 		g_debug("BGRT setup failed: %s", error_local->message);
 	str = fu_uefi_bgrt_get_supported(self->bgrt) ? "Enabled" : "Disabled";
-	g_debug("UX Capsule support : %s", str);
+	g_info("UX capsule support : %s", str);
 	fu_plugin_add_report_metadata(plugin, "UEFIUXCapsule", str);
 	fu_progress_step_done(progress);
 
 	return TRUE;
 }
 
+static gboolean
+fu_uefi_capsule_plugin_cleanup_esp(FuUefiCapsulePlugin *self, GError **error)
+{
+	g_autofree gchar *esp_os_base = fu_uefi_get_esp_path_for_os();
+	g_autofree gchar *esp_path = NULL;
+	g_autofree gchar *pattern = NULL;
+	g_autoptr(FuDeviceLocker) esp_locker = NULL;
+	g_autoptr(GPtrArray) files = NULL;
+
+	/* delete any files matching the glob in the ESP */
+	esp_locker = fu_volume_locker(self->esp, error);
+	if (esp_locker == NULL)
+		return FALSE;
+	esp_path = fu_volume_get_mount_point(self->esp);
+	if (esp_path == NULL) {
+		g_set_error_literal(error,
+				    G_IO_ERROR,
+				    G_IO_ERROR_NOT_SUPPORTED,
+				    "no mountpoint for ESP");
+		return FALSE;
+	}
+	files = fu_path_get_files(esp_path, error);
+	if (files == NULL)
+		return FALSE;
+	pattern = g_build_filename(esp_path, esp_os_base, "fw", "fwupd*.cap", NULL);
+	for (guint i = 0; i < files->len; i++) {
+		const gchar *fn = g_ptr_array_index(files, i);
+		if (g_pattern_match_simple(pattern, fn)) {
+			g_autoptr(GFile) file = g_file_new_for_path(fn);
+			g_debug("deleting %s", fn);
+			if (!g_file_delete(file, NULL, error))
+				return FALSE;
+		}
+	}
+
+	/* success */
+	return TRUE;
+}
+
+static gboolean
+fu_uefi_capsule_plugin_cleanup_bootnext(FuUefiCapsulePlugin *self, GError **error)
+{
+	guint16 boot_next = 0;
+	g_autofree gchar *boot_xxxxstr = NULL;
+	g_autofree gchar *loadoptstr = NULL;
+	g_autoptr(FuEfiLoadOption) loadopt = fu_efi_load_option_new();
+	g_autoptr(GBytes) boot_nextbuf = NULL;
+	g_autoptr(GBytes) boot_xxxxbuf = NULL;
+
+	/* unset */
+	if (!fu_efivar_exists(FU_EFIVAR_GUID_EFI_GLOBAL, "BootNext"))
+		return TRUE;
+
+	/* get the value of BootNext */
+	boot_nextbuf = fu_efivar_get_data_bytes(FU_EFIVAR_GUID_EFI_GLOBAL, "BootNext", NULL, error);
+	if (boot_nextbuf == NULL)
+		return FALSE;
+	if (!fu_memread_uint16_safe(g_bytes_get_data(boot_nextbuf, NULL),
+				    g_bytes_get_size(boot_nextbuf),
+				    0x0,
+				    &boot_next,
+				    G_LITTLE_ENDIAN,
+				    error))
+		return FALSE;
+
+	/* get the correct BootXXXX key */
+	boot_xxxxstr = g_strdup_printf("Boot%04X", boot_next);
+	boot_xxxxbuf =
+	    fu_efivar_get_data_bytes(FU_EFIVAR_GUID_EFI_GLOBAL, boot_xxxxstr, NULL, error);
+	if (boot_xxxxbuf == NULL)
+		return FALSE;
+	if (!fu_firmware_parse(FU_FIRMWARE(loadopt), boot_xxxxbuf, FWUPD_INSTALL_FLAG_NONE, error))
+		return FALSE;
+
+	/* is this us? */
+	loadoptstr = fu_firmware_to_string(FU_FIRMWARE(loadopt));
+	g_debug("EFI LoadOption: %s", loadoptstr);
+	if (g_strcmp0(fu_firmware_get_id(FU_FIRMWARE(loadopt)), "Linux Firmware Updater") == 0 ||
+	    g_strcmp0(fu_firmware_get_id(FU_FIRMWARE(loadopt)), "Linux-Firmware-Updater") == 0) {
+		g_warning("BootNext was not deleted automatically, so removing: "
+			  "this normally indicates a BIOS bug");
+		if (!fu_efivar_delete(FU_EFIVAR_GUID_EFI_GLOBAL, "BootNext", error))
+			return FALSE;
+	}
+
+	/* success */
+	return TRUE;
+}
+
+static gboolean
+fu_uefi_capsule_plugin_reboot_cleanup(FuPlugin *plugin, FuDevice *device, GError **error)
+{
+	FuUefiCapsulePlugin *self = FU_UEFI_CAPSULE_PLUGIN(plugin);
+
+	/* provide an escape hatch for debugging */
+	if (!fu_plugin_get_config_value_boolean(plugin, "RebootCleanup"))
+		return TRUE;
+
+	/* delete capsules */
+	if (!fu_uefi_capsule_plugin_cleanup_esp(self, error))
+		return FALSE;
+
+	/* delete any old variables */
+	if (!fu_efivar_delete_with_glob(FU_EFIVAR_GUID_FWUPDATE, "fwupd*-*", error))
+		return FALSE;
+
+	/* this should not be required, but, hey -- here were are */
+	if (!fu_uefi_capsule_plugin_cleanup_bootnext(self, error))
+		return FALSE;
+
+	/* success */
+	return TRUE;
+}
+
 static void
 fu_uefi_capsule_plugin_init(FuUefiCapsulePlugin *self)
 {
 	self->bgrt = fu_uefi_bgrt_new();
+	fu_plugin_add_flag(FU_PLUGIN(self), FWUPD_PLUGIN_FLAG_MEASURE_SYSTEM_INTEGRITY);
 }
 
 static void
@@ -1014,6 +1348,17 @@ fu_uefi_capsule_plugin_constructed(GObject *obj)
 	fu_plugin_add_rule(plugin, FU_PLUGIN_RULE_METADATA_SOURCE, "linux_lockdown");
 	fu_plugin_add_rule(plugin, FU_PLUGIN_RULE_METADATA_SOURCE, "acpi_phat");
 	fu_plugin_add_rule(plugin, FU_PLUGIN_RULE_CONFLICTS, "uefi"); /* old name */
+	fu_plugin_add_firmware_gtype(FU_PLUGIN(self), NULL, FU_TYPE_ACPI_UEFI);
+	fu_plugin_add_firmware_gtype(FU_PLUGIN(self), NULL, FU_TYPE_UEFI_UPDATE_INFO);
+
+	/* defaults changed here will also be reflected in the fwupd.conf man page */
+	fu_plugin_set_config_default(plugin, "DisableCapsuleUpdateOnDisk", "false");
+	fu_plugin_set_config_default(plugin, "DisableShimForSecureBoot", "false");
+	fu_plugin_set_config_default(plugin, "EnableEfiDebugging", "false");
+	fu_plugin_set_config_default(plugin, "EnableGrubChainLoad", "false");
+	fu_plugin_set_config_default(plugin, "OverrideESPMountPoint", NULL);
+	fu_plugin_set_config_default(plugin, "RebootCleanup", "true");
+	fu_plugin_set_config_default(plugin, "RequireESPFreeSpace", "0"); /* in MB */
 
 	/* add a requirement on the fwupd-efi version -- which can change  */
 	if (!fu_uefi_capsule_plugin_fwupd_efi_probe(self, &error_local))
@@ -1028,12 +1373,16 @@ fu_uefi_capsule_finalize(GObject *obj)
 		g_object_unref(self->esp);
 	if (self->fwupd_efi_file != NULL)
 		g_object_unref(self->fwupd_efi_file);
+	if (self->acpi_uefi != NULL)
+		g_object_unref(self->acpi_uefi);
 	if (self->fwupd_efi_monitor != NULL) {
 		g_file_monitor_cancel(self->fwupd_efi_monitor);
 		g_object_unref(self->fwupd_efi_monitor);
 	}
-	g_object_unref(self->backend);
-	g_object_unref(self->bgrt);
+	if (self->backend != NULL)
+		g_object_unref(self->backend);
+	if (self->bgrt != NULL)
+		g_object_unref(self->bgrt);
 	G_OBJECT_CLASS(fu_uefi_capsule_plugin_parent_class)->finalize(obj);
 }
 
@@ -1043,8 +1392,8 @@ fu_uefi_capsule_plugin_class_init(FuUefiCapsulePluginClass *klass)
 	FuPluginClass *plugin_class = FU_PLUGIN_CLASS(klass);
 	GObjectClass *object_class = G_OBJECT_CLASS(klass);
 
-	object_class->constructed = fu_uefi_capsule_plugin_constructed;
 	object_class->finalize = fu_uefi_capsule_finalize;
+	plugin_class->constructed = fu_uefi_capsule_plugin_constructed;
 	plugin_class->to_string = fu_uefi_capsule_plugin_to_string;
 	plugin_class->clear_results = fu_uefi_capsule_plugin_clear_results;
 	plugin_class->add_security_attrs = fu_uefi_capsule_plugin_add_security_attrs;
@@ -1053,4 +1402,5 @@ fu_uefi_capsule_plugin_class_init(FuUefiCapsulePluginClass *klass)
 	plugin_class->unlock = fu_uefi_capsule_plugin_unlock;
 	plugin_class->coldplug = fu_uefi_capsule_plugin_coldplug;
 	plugin_class->write_firmware = fu_uefi_capsule_plugin_write_firmware;
+	plugin_class->reboot_cleanup = fu_uefi_capsule_plugin_reboot_cleanup;
 }
